@@ -8,7 +8,7 @@ open Environ
 
 (** {6 Checking constants } *)
 
-let indirect_accessor : (cooking_info Opaqueproof.opaque -> Constr.t * unit Opaqueproof.delayed_universes) ref =
+let indirect_accessor : (Opaqueproof.opaque -> Constr.t * unit Opaqueproof.delayed_universes) ref =
   ref (fun _ -> assert false)
 
 let set_indirect_accessor f = indirect_accessor := f
@@ -55,7 +55,7 @@ let check_constant_declaration env opac kn cb opacify =
       let c, u = !indirect_accessor o in
       let env = match u, cb.const_universes with
         | Opaqueproof.PrivateMonomorphic (), Monomorphic -> env
-        | Opaqueproof.PrivatePolymorphic (_, local), Polymorphic _ ->
+        | Opaqueproof.PrivatePolymorphic local, Polymorphic _ ->
           push_subgraph local env
         | _ -> assert false
       in
@@ -98,7 +98,7 @@ let mk_mtb mp sign delta =
     mod_retroknowledge = ModTypeRK; }
 
 let rec collect_constants_without_body sign mp accu =
-  let collect_sf s lab = function
+  let collect_field s lab = function
   | SFBconst cb ->
      let c = Constant.make2 mp lab in
      if Declareops.constant_has_body cb then s else Cset.add c s
@@ -107,7 +107,7 @@ let rec collect_constants_without_body sign mp accu =
   match sign with
   | MoreFunctor _ -> Cset.empty  (* currently ignored *)
   | NoFunctor struc ->
-     List.fold_left (fun s (lab,mb) -> collect_sf s lab mb) accu struc
+     List.fold_left (fun s (lab,mb) -> collect_field s lab mb) accu struc
 
 let rec check_module env opac mp mb opacify =
   Flags.if_verbose Feedback.msg_notice (str "  checking module: " ++ str (ModPath.to_string mp));
@@ -116,22 +116,22 @@ let rec check_module env opac mp mb opacify =
     check_signature env opac mb.mod_type mb.mod_mp mb.mod_delta opacify
   in
   let optsign, opac = match mb.mod_expr with
-    |Struct sign_struct ->
+    | Struct sign_struct ->
       let opacify = collect_constants_without_body sign mb.mod_mp opacify in
       let sign, opac = check_signature env opac sign_struct mb.mod_mp mb.mod_delta opacify in
       Some (sign, mb.mod_delta), opac
-    |Algebraic me -> Some (check_mexpression env opac me mb.mod_mp mb.mod_delta), opac
-    |Abstract|FullStruct -> None, opac
+    | Algebraic me -> Some (check_mexpression env opac me mb.mod_mp mb.mod_delta), opac
+    | Abstract|FullStruct -> None, opac
   in
   let () = match optsign with
-  |None -> ()
-  |Some (sign,delta) ->
+  | None -> ()
+  | Some (sign,delta) ->
     let mtb1 = mk_mtb mp sign delta
     and mtb2 = mk_mtb mp mb.mod_type mb.mod_delta in
     let env = Modops.add_module_type mp mtb1 env in
-    let cu = Subtyping.check_subtypes env mtb1 mtb2 in
-    if not (Environ.check_constraints cu env) then
-      CErrors.user_err Pp.(str "Incorrect universe constraints for module subtyping");
+    let state = (Environ.universes env, Reduction.checked_universes) in
+    let _ : UGraph.t = Subtyping.check_subtypes state env mtb1 mtb2 in
+    ()
   in
   opac
 
@@ -159,15 +159,14 @@ and check_structure_field env opac mp lab res opacify = function
 and check_mexpr env opac mse mp_mse res = match mse with
   | MEident mp ->
       let mb = lookup_module mp env in
-      let mb = Modops.strengthen_and_subst_mb mb mp_mse false in
+      let mb = Modops.strengthen_and_subst_module_body mb mp_mse false in
       mb.mod_type, mb.mod_delta
   | MEapply (f,mp) ->
       let sign, delta = check_mexpr env opac f mp_mse res in
       let farg_id, farg_b, fbody_b = Modops.destr_functor sign in
       let mtb = Modops.module_type_of_module (lookup_module mp env) in
-      let cu = Subtyping.check_subtypes env mtb farg_b in
-      if not (Environ.check_constraints cu env) then
-        CErrors.user_err Pp.(str "Incorrect universe constraints for module subtyping");
+      let state = (Environ.universes env, Reduction.checked_universes) in
+      let _ : UGraph.t = Subtyping.check_subtypes state env mtb farg_b in
       let subst = Mod_subst.map_mbid farg_id mp Mod_subst.empty_delta_resolver in
       Modops.subst_signature subst fbody_b, Mod_subst.subst_codom_delta_resolver subst delta
   | MEwith _ -> CErrors.user_err Pp.(str "Unsupported 'with' constraint in module implementation")
