@@ -81,6 +81,7 @@ sig
   exception IndexOutOfRange
   val goto : int -> 'a list -> 'a list * 'a list
   val split_when : ('a -> bool) -> 'a list -> 'a list * 'a list
+  val sep_first : 'a list -> 'a * 'a list
   val sep_last : 'a list -> 'a * 'a list
   val drop_last : 'a list -> 'a list
   val last : 'a list -> 'a
@@ -127,6 +128,8 @@ sig
   module Smart :
   sig
     val map : ('a -> 'a) -> 'a list -> 'a list
+    val fold_left_map : ('a -> 'b -> 'a * 'b) -> 'a -> 'b list -> 'a * 'b list
+    val fold_right_map : ('b -> 'a -> 'b * 'a) -> 'b list -> 'a -> 'b list * 'a
   end
 
   module type MonoS = sig
@@ -745,6 +748,10 @@ let firstn n l =
   in
   aux [] n l
 
+let sep_first = function
+  | [] -> failwith "sep_first"
+  | hd :: tl -> (hd, tl)
+
 let rec sep_last = function
   | [] -> failwith "sep_last"
   | hd :: [] -> (hd,[])
@@ -1045,6 +1052,22 @@ struct
     let tl' = map f tl in
     if h' == h && tl' == tl then l else h' :: tl'
 
+  let rec fold_left_map f e l =
+    match l with
+    | [] -> e, []
+    | h :: tl ->
+      let e', h' = f e h in
+      let e'', tl' = fold_left_map f e' tl in
+      e'', (if h' == h && tl' == tl then l else h' :: tl')
+
+  let rec fold_right_map f l e =
+    match l with
+    | [] -> [], e
+    | h :: tl ->
+      let tl', e' = fold_right_map f tl e in
+      let h', e'' = f h e' in
+      (if h' == h && tl' == tl then l else h' :: tl'), e''
+
 end
 
 module type MonoS = sig
