@@ -558,11 +558,8 @@ let plugin_regex = Str.regexp "^plugins/\\([a-zA-Z0-9_]+\\)/"
 let level_regex = Str.regexp "[a-zA-Z0-9_]*$"
 
 let get_plugin_name file =
-  if file = "user-contrib/Ltac2/g_ltac2.mlg" then
-    "Ltac2"
-  (* todo: would be better if g_search.mlg has an "ssr" prefix *)
-  else if List.mem file ["plugins/ssr/ssrparser.mlg"; "plugins/ssr/ssrvernac.mlg";
-      "plugins/ssrmatching/g_ssrmatching.mlg"; "plugins/ssrsearch/g_search.mlg"] then
+  if List.mem file ["plugins/ssr/ssrparser.mlg"; "plugins/ssr/ssrvernac.mlg";
+      "plugins/ssrmatching/g_ssrmatching.mlg"] then
     "SSR"
   else if Str.string_match plugin_regex file 0 then
     Str.matched_group 1 file
@@ -1560,8 +1557,10 @@ let finish_with_file old_file args =
   if !exit_code <> 0 then
     Sys.remove temp_file
   else if args.verify then begin
-    if not (files_eq old_file temp_file) then
+    if not (files_eq old_file temp_file) then begin
       error "%s is not current\n" old_file;
+      ignore (CUnix.sys_command "diff" [ old_file ; old_file ^ ".new"])
+    end;
     Sys.remove temp_file
   end else if args.update then
     Sys.rename temp_file old_file
@@ -1739,7 +1738,6 @@ let process_rst g file args seen tac_prods cmd_prods =
   let cmd_exclude_files = [
     "doc/sphinx/proof-engine/ssreflect-proof-language.rst";
     "doc/sphinx/proof-engine/tactics.rst";
-    "doc/sphinx/proofs/writing-proofs/reasoning-inductives.rst";
   ]
   in
 
@@ -1805,13 +1803,13 @@ let process_rst g file args seen tac_prods cmd_prods =
           if rhs = "coq" && !show_warn then
             warn "%s line %d: Missing 'insertprodn' before 'prodn:: coq'\n" file !linenum;
           fprintf new_rst "%s\n" line;
-        | "tacn::" when args.check_tacs ->
+        | "tacn::" ->
           seen := { !seen with tacs = save_n_get_more "tacn" pfx rhs !seen.tacs tac_prods }
-        | "tacv::" when args.check_tacs ->
+        | "tacv::" ->
           seen := { !seen with tacvs = save_n_get_more "tacv" pfx rhs !seen.tacvs StringSet.empty }
-        | "cmd::" when args.check_cmds ->
+        | "cmd::" ->
           seen := { !seen with cmds = save_n_get_more "cmd" pfx rhs !seen.cmds cmd_prods }
-        | "cmdv::" when args.check_cmds ->
+        | "cmdv::" ->
           seen := { !seen with cmdvs = save_n_get_more "cmdv" pfx rhs !seen.cmdvs StringSet.empty }
         | "insertprodn" ->
           process_insertprodn line rhs
@@ -1825,38 +1823,25 @@ let process_rst g file args seen tac_prods cmd_prods =
   finish_with_file file args
 
 let report_omitted_prods g seen label split =
-  let maybe_warn first last n =
-    if first <> "" && !show_warn then begin
-      if first <> last then
-        warn "%ss '%s' to %s'%s' not included in .rst files (%d)\n" label first split last n
-      else
-        warn "%s %s not included in .rst files\n" label first;
-    end
-  in
-
-  let included = try
-      List.map (fun prod ->
-        match prod with
-        | Snterm nt :: tl -> nt
-        | _ -> "") (NTMap.find "NOTINRSTS" !g.map)
-    with Not_found -> []
-  in
-  let first, last, n, total = List.fold_left (fun missing nt ->
-      let first, last, n, total = missing in
-      if NTMap.mem nt seen || (List.mem nt included) then begin
-        maybe_warn first last n;
-        "", "", 0, total
-      end else
-        (if first = "" then nt else first), nt, n + 1, total + 1)
-    ("", "", 0, 0) !g.order in
-  maybe_warn first last n;
-  Printf.printf "\n\n";
-  NTMap.iter (fun nt _ ->
-      if !show_warn && not (NTMap.mem nt seen || (List.mem nt included)) then
+  if !show_warn then begin
+    let included = try
+        List.map (fun prod ->
+          match prod with
+          | Snterm nt :: tl -> nt
+          | _ -> "") (NTMap.find "NOTINRSTS" !g.map)
+      with Not_found -> []
+    in
+    Printf.printf "\n\n";
+    let missing = NTMap.filter (fun nt _ ->
+        not (NTMap.mem nt seen || (List.mem nt included)))
+      !g.map in
+    NTMap.iter (fun nt _ ->
         warn "%s %s not included in .rst files\n" "Nonterminal" nt)
-    !g.map;
-  if total <> 0 then
-    Printf.eprintf "TOTAL %ss not included = %d\n" label total
+      missing;
+    let total = NTMap.cardinal missing in
+    if total <> 0 then
+      Printf.eprintf "TOTAL %ss not included = %d\n" label total
+  end
 
 let process_grammar args =
   let symdef_map = ref StringMap.empty in
