@@ -232,7 +232,7 @@ let decompose_lam_assum sigma c =
     match kind sigma c with
     | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) c
     | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) c
-    | Cast (c,_,_)      -> lamdec_rec l c
+    | Cast (c,_,_)    -> lamdec_rec l c
     | _               -> l,c
   in
   lamdec_rec Context.Rel.empty c
@@ -240,30 +240,30 @@ let decompose_lam_assum sigma c =
 let decompose_lam_n_assum sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    user_err Pp.(str "decompose_lam_n_assum: integer parameter must be positive");
+    anomaly Pp.(str "decompose_lam_n_assum: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
       match kind sigma c with
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) n c
-      | Cast (c,_,_)      -> lamdec_rec l n c
-      | c -> user_err Pp.(str "decompose_lam_n_assum: not enough abstractions")
+      | Cast (c,_,_)    -> lamdec_rec l n c
+      | c -> anomaly Pp.(str "decompose_lam_n_assum: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n c
 
 let decompose_lam_n_decls sigma n =
   let open Rel.Declaration in
   if n < 0 then
-    user_err Pp.(str "decompose_lam_n_decls: integer parameter must be positive");
+    anomaly Pp.(str "decompose_lam_n_decls: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
       match kind sigma c with
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
-      | Cast (c,_,_)      -> lamdec_rec l n c
-      | c -> user_err Pp.(str "decompose_lam_n_decls: not enough abstractions")
+      | Cast (c,_,_)    -> lamdec_rec l n c
+      | c -> anomaly Pp.(str "decompose_lam_n_decls: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n
 
@@ -284,13 +284,13 @@ let rec to_lambda sigma n prod =
     match kind sigma prod with
       | Prod (na,ty,bd) -> mkLambda (na,ty,to_lambda sigma (n-1) bd)
       | Cast (c,_,_) -> to_lambda sigma n c
-      | _   -> user_err ~hdr:"to_lambda" (Pp.mt ())
+      | _   -> anomaly Pp.(str "Not enough products.")
 
 let decompose_prod sigma c =
   let rec proddec_rec l c = match kind sigma c with
     | Prod (x,t,c) -> proddec_rec ((x,t)::l) c
-    | Cast (c,_,_)     -> proddec_rec l c
-    | _                -> l,c
+    | Cast (c,_,_) -> proddec_rec l c
+    | _            -> l,c
   in
   proddec_rec [] c
 
@@ -298,9 +298,9 @@ let decompose_prod_assum sigma c =
   let open Rel.Declaration in
   let rec proddec_rec l c =
     match kind sigma c with
-    | Prod (x,t,c)  -> proddec_rec (Context.Rel.add (LocalAssum (x,t)) l) c
+    | Prod (x,t,c)    -> proddec_rec (Context.Rel.add (LocalAssum (x,t)) l) c
     | LetIn (x,b,t,c) -> proddec_rec (Context.Rel.add (LocalDef (x,b,t)) l) c
-    | Cast (c,_,_)      -> proddec_rec l c
+    | Cast (c,_,_)    -> proddec_rec l c
     | _               -> l,c
   in
   proddec_rec Context.Rel.empty c
@@ -308,15 +308,15 @@ let decompose_prod_assum sigma c =
 let decompose_prod_n_assum sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    user_err Pp.(str "decompose_prod_n_assum: integer parameter must be positive");
+    anomaly Pp.(str "decompose_prod_n_assum: integer parameter must be positive.");
   let rec prodec_rec l n c =
     if Int.equal n 0 then l,c
     else
       match kind sigma c with
       | Prod (x,t,c)    -> prodec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> prodec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
-      | Cast (c,_,_)      -> prodec_rec l n c
-      | c -> user_err Pp.(str "decompose_prod_n_assum: not enough assumptions")
+      | Cast (c,_,_)    -> prodec_rec l n c
+      | c -> anomaly Pp.(str "decompose_prod_n_assum: not enough declarations.")
   in
   prodec_rec Context.Rel.empty n c
 
@@ -391,11 +391,10 @@ let expand_branch env _sigma u pms (ind, i) (nas, _br) =
   let pms = unsafe_to_constr_array pms in
   let (mib, mip) = Inductive.lookup_mind_specif env ind in
   let paramdecl = Vars.subst_instance_context u mib.mind_params_ctxt in
-  let paramsubst = Vars.subst_of_rel_context_instance paramdecl (Array.to_list pms) in
-  let subst = paramsubst @ Inductive.ind_subst (fst ind) mib u in
+  let paramsubst = Vars.subst_of_rel_context_instance paramdecl pms in
   let (ctx, _) = mip.mind_nf_lc.(i - 1) in
   let (ctx, _) = List.chop mip.mind_consnrealdecls.(i - 1) ctx in
-  let ans = Inductive.instantiate_context u subst nas ctx in
+  let ans = Inductive.instantiate_context u paramsubst nas ctx in
   let ans : rel_context = match Evd.MiniEConstr.unsafe_eq with Refl -> ans in
   ans
 
@@ -447,6 +446,10 @@ let fold sigma f acc c =
   let f acc c = f acc (of_constr c) in
   Constr.fold f acc (unsafe_to_constr (whd_evar sigma c))
 
+let fold_with_binders sigma g f acc e c =
+  let f e acc c = f e acc (of_constr c) in
+  Constr.fold_constr_with_binders g f acc e (unsafe_to_constr (whd_evar sigma c))
+
 let compare_gen k eq_inst eq_sort eq_constr nargs c1 c2 =
   (c1 == c2) || Constr.compare_head_gen_with k k eq_inst eq_sort eq_constr nargs c1 c2
 
@@ -477,21 +480,18 @@ let compare_cumulative_instances cv_pb nargs_ok variances u u' cstrs =
   let open UnivProblem in
   if not nargs_ok then enforce_eq_instances_univs false u u' cstrs
   else
+    let make u = Sorts.sort_of_univ @@ Univ.Universe.make u in
     CArray.fold_left3
       (fun cstrs v u u' ->
          let open Univ.Variance in
          match v with
          | Irrelevant -> Set.add (UWeak (u,u')) cstrs
          | Covariant ->
-           let u = Univ.Universe.make u in
-           let u' = Univ.Universe.make u' in
            (match cv_pb with
-            | Reduction.CONV -> Set.add (UEq (u,u')) cstrs
-            | Reduction.CUMUL -> Set.add (ULe (u,u')) cstrs)
+            | Reduction.CONV -> Set.add (UEq (make u, make u')) cstrs
+            | Reduction.CUMUL -> Set.add (ULe (make u, make u')) cstrs)
          | Invariant ->
-           let u = Univ.Universe.make u in
-           let u' = Univ.Universe.make u' in
-           Set.add (UEq (u,u')) cstrs)
+           Set.add (UEq (make u, make u')) cstrs)
       cstrs variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
 
 let cmp_inductives cv_pb (mind,ind as spec) nargs u1 u2 cstrs =
@@ -537,7 +537,7 @@ let eq_universes env sigma cstrs cv_pb refargs l l' =
       cstrs := cmp_constructors (mind,ind,ctor) nargs l l' !cstrs;
       true
 
-let test_constr_universes env sigma leq m n =
+let test_constr_universes env sigma leq ?(nargs=0) m n =
   let open UnivProblem in
   let kind c = kind sigma c in
   if m == n then Some Set.empty
@@ -551,7 +551,7 @@ let test_constr_universes env sigma leq m n =
       let s2 = ESorts.kind sigma s2 in
       if Sorts.equal s1 s2 then true
       else (cstrs := Set.add
-              (UEq (Sorts.univ_of_sort s1,Sorts.univ_of_sort s2)) !cstrs;
+              (UEq (s1, s2)) !cstrs;
             true)
     in
     let leq_sorts s1 s2 =
@@ -560,7 +560,7 @@ let test_constr_universes env sigma leq m n =
       if Sorts.equal s1 s2 then true
       else
         (cstrs := Set.add
-           (ULe (Sorts.univ_of_sort s1,Sorts.univ_of_sort s2)) !cstrs;
+           (ULe (s1, s2)) !cstrs;
          true)
     in
     let rec eq_constr' nargs m n = compare_gen kind eq_universes eq_sorts eq_constr' nargs m n in
@@ -570,16 +570,16 @@ let test_constr_universes env sigma leq m n =
           Constr.compare_head_gen_leq_with kind kind leq_universes leq_sorts
             eq_constr' leq_constr' nargs m n
         and leq_constr' nargs m n = m == n || compare_leq nargs m n in
-        compare_leq 0 m n
+        compare_leq nargs m n
       else
-        Constr.compare_head_gen_with kind kind eq_universes eq_sorts eq_constr' 0 m n
+        Constr.compare_head_gen_with kind kind eq_universes eq_sorts eq_constr' nargs m n
     in
     if res then Some !cstrs else None
 
-let eq_constr_universes env sigma m n =
-  test_constr_universes env sigma false m n
-let leq_constr_universes env sigma m n =
-  test_constr_universes env sigma true m n
+let eq_constr_universes env sigma ?nargs m n =
+  test_constr_universes env sigma false ?nargs m n
+let leq_constr_universes env sigma ?nargs m n =
+  test_constr_universes env sigma true ?nargs m n
 
 let compare_head_gen_proj env sigma equ eqs eqc' nargs m n =
   let kind c = kind sigma c in
@@ -607,7 +607,7 @@ let eq_constr_universes_proj env sigma m n =
       if Sorts.equal s1 s2 then true
       else
         (cstrs := Set.add
-           (UEq (Sorts.univ_of_sort s1, Sorts.univ_of_sort s2)) !cstrs;
+           (UEq (s1, s2)) !cstrs;
          true)
     in
     let rec eq_constr' nargs m n =
@@ -621,26 +621,25 @@ let universes_of_constr sigma c =
   let rec aux s c =
     match kind sigma c with
     | Const (c, u) ->
-      LSet.fold LSet.add (Instance.levels (EInstance.kind sigma u)) s
+      Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s
     | Ind ((mind,_), u) | Construct (((mind,_),_), u) ->
-      LSet.fold LSet.add (Instance.levels (EInstance.kind sigma u)) s
+      Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s
     | Sort u ->
       let sort = ESorts.kind sigma u in
       if Sorts.is_small sort then s
       else
-        let u = Sorts.univ_of_sort sort in
-        LSet.fold LSet.add (Universe.levels u) s
+        Level.Set.fold Level.Set.add (Sorts.levels sort) s
     | Evar (k, args) ->
       let concl = Evd.evar_concl (Evd.find sigma k) in
       fold sigma aux (aux s concl) c
     | Array (u,_,_,_) ->
-      let s = LSet.fold LSet.add (Instance.levels (EInstance.kind sigma u)) s in
+      let s = Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s in
       fold sigma aux s c
     | Case (_,u,_,_,_,_,_) ->
-      let s = LSet.fold LSet.add (Instance.levels (EInstance.kind sigma u)) s in
+      let s = Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s in
       fold sigma aux s c
     | _ -> fold sigma aux s c
-  in aux LSet.empty c
+  in aux Level.Set.empty c
 
 open Context
 open Environ
@@ -651,6 +650,9 @@ let cast_list : type a b. (a,b) eq -> a list -> b list =
 let cast_list_snd : type a b. (a,b) eq -> ('c * a) list -> ('c * b) list =
   fun Refl x -> x
 
+let cast_vect : type a b. (a,b) eq -> a array -> b array =
+  fun Refl x -> x
+
 let cast_rel_decl :
   type a b. (a,b) eq -> (a, a) Rel.Declaration.pt -> (b, b) Rel.Declaration.pt =
   fun Refl x -> x
@@ -678,6 +680,8 @@ exception LocalOccur
 let to_constr = unsafe_to_constr
 let to_rel_decl = unsafe_to_rel_decl
 
+type instance = t array
+type instance_list = t list
 type substl = t list
 
 (** Operations that commute with evar-normalization *)
@@ -701,6 +705,12 @@ let subst_var subst c = of_constr (Vars.subst_var subst (to_constr c))
 let subst_univs_level_constr subst c =
   of_constr (Vars.subst_univs_level_constr subst (to_constr c))
 
+let subst_instance_context subst ctx =
+  cast_rel_context (sym unsafe_eq) (Vars.subst_instance_context subst (cast_rel_context unsafe_eq ctx))
+
+let subst_instance_constr subst c =
+  of_constr (Vars.subst_instance_constr subst (to_constr c))
+
 (** Operations that dot NOT commute with evar-normalization *)
 let noccurn sigma n term =
   let rec occur_rec n c = match kind sigma c with
@@ -727,7 +737,39 @@ let closed0 sigma c = closedn sigma 0 c
 
 let subst_of_rel_context_instance ctx subst =
   cast_list (sym unsafe_eq)
-    (Vars.subst_of_rel_context_instance (cast_rel_context unsafe_eq ctx) (cast_list unsafe_eq subst))
+    (Vars.subst_of_rel_context_instance (cast_rel_context unsafe_eq ctx) (cast_vect unsafe_eq subst))
+let subst_of_rel_context_instance_list ctx subst =
+  cast_list (sym unsafe_eq)
+    (Vars.subst_of_rel_context_instance_list (cast_rel_context unsafe_eq ctx) (cast_list unsafe_eq subst))
+
+let liftn_rel_context n k ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.liftn_rel_context n k (cast_rel_context unsafe_eq ctx))
+
+let lift_rel_context n ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.lift_rel_context n (cast_rel_context unsafe_eq ctx))
+
+let substnl_rel_context subst n ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.substnl_rel_context (cast_list unsafe_eq subst) n (cast_rel_context unsafe_eq ctx))
+
+let substl_rel_context subst ctx =
+  cast_rel_context (sym unsafe_eq)
+    (Vars.substl_rel_context (cast_list unsafe_eq subst) (cast_rel_context unsafe_eq ctx))
+
+let smash_rel_context ctx =
+  cast_rel_context (sym unsafe_eq)
+      (Vars.smash_rel_context (cast_rel_context unsafe_eq ctx))
+
+let esubst : (int -> 'a -> t) -> 'a Esubst.subs -> t -> t =
+match unsafe_eq with
+| Refl -> Vars.esubst
+
+type substituend = Vars.substituend
+let make_substituend c = Vars.make_substituend (unsafe_to_constr c)
+let lift_substituend n s = of_constr (Vars.lift_substituend n s)
+
 
 end
 
@@ -861,6 +903,11 @@ let to_named_context =
     = fun Refl x -> x
   in
   gen unsafe_eq
+let to_rel_context =
+  let gen : type a b. (a, b) eq -> (a,a) Context.Rel.pt -> (b,b) Context.Rel.pt
+    = fun Refl x -> x
+  in
+  gen unsafe_eq
 let to_case_invert = unsafe_to_case_invert
 
 let eq = unsafe_eq
