@@ -34,7 +34,7 @@ let set_print_constr f = term_printer := f
 
 module EvMap = Evar.Map
 
-let evar_suggested_name evk sigma =
+let evar_suggested_name env sigma evk =
   let open Evd in
   let base_id evk' evi =
   match evar_ident evk' sigma with
@@ -45,7 +45,7 @@ let evar_suggested_name evk sigma =
   | _,Evar_kinds.QuestionMark {Evar_kinds.qm_name = Name id} -> id
   | _,Evar_kinds.GoalEvar -> Id.of_string "Goal"
   | _ ->
-      let env = reset_with_named_context evi.evar_hyps (Global.env()) in
+      let env = reset_with_named_context evi.evar_hyps env in
       Namegen.id_of_name_using_hdchar env sigma evi.evar_concl Anonymous
   in
   let names = EvMap.mapi base_id (undefined_map sigma) in
@@ -59,11 +59,11 @@ let evar_suggested_name evk sigma =
   let (_, n) = EvMap.fold fold names (false, 0) in
   if n = 0 then id else Nameops.add_suffix id (string_of_int (pred n))
 
-let pr_existential_key sigma evk =
+let pr_existential_key env sigma evk =
 let open Evd in
 match evar_ident evk sigma with
 | None ->
-  str "?" ++ Id.print (evar_suggested_name evk sigma)
+  str "?" ++ Id.print (evar_suggested_name env sigma evk)
 | Some id ->
   str "?" ++ Id.print id
 
@@ -127,7 +127,7 @@ let pr_evar_source env sigma = function
       let pp = match ido with
         | Some id -> str "?" ++ Id.print id
         | None ->
-          try pr_existential_key sigma evk
+          try pr_existential_key env sigma evk
           with (* defined *) Not_found -> str "an internal placeholder" in
      str "type of " ++ pp
   | Evar_kinds.ImplicitArg (c,(n,ido),b) ->
@@ -228,7 +228,7 @@ let evar_dependency_closure n sigma =
   EvMap.bindings map
 
 let has_no_evar sigma =
-  try let () = Evd.fold (fun _ _ () -> raise Exit) sigma () in true
+  try let () = Evd.fold (fun _ _ () -> raise_notrace Exit) sigma () in true
   with Exit -> false
 
 let pr_evd_level sigma = UState.pr_uctx_level (Evd.evar_universe_context sigma)
@@ -239,12 +239,12 @@ let pr_evar_universe_context ctx =
   let prl = pr_uctx_level ctx in
   if UState.is_empty ctx then mt ()
   else
-    (str"UNIVERSES:"++brk(0,1)++
+    v 0 (str"UNIVERSES:"++brk(0,1)++
        h (Univ.pr_universe_context_set prl (UState.context_set ctx)) ++ fnl () ++
      str"ALGEBRAIC UNIVERSES:"++brk(0,1)++
-     h (Univ.LSet.pr prl (UState.algebraics ctx)) ++ fnl() ++
+     h (Univ.Level.Set.pr prl (UState.algebraics ctx)) ++ fnl() ++
      str"UNDEFINED UNIVERSES:"++brk(0,1)++
-     h (UnivSubst.pr_universe_opt_subst (UState.subst ctx)) ++ fnl() ++
+     h (UState.pr_universe_opt_subst (UState.subst ctx)) ++ fnl() ++
      str "WEAK CONSTRAINTS:"++brk(0,1)++
      h (UState.pr_weak prl ctx) ++ fnl ())
 
@@ -327,7 +327,7 @@ let pr_evar_list env sigma l =
       str "==" ++ pr_evar_info env sigma evi ++
       pr_alias ev ++
       (if evi.evar_body == Evar_empty
-       then str " {" ++ pr_existential_key sigma ev ++ str "}"
+       then str " {" ++ pr_existential_key env sigma ev ++ str "}"
        else mt ()))
   in
   hv 0 (prlist_with_sep fnl pr l)
@@ -885,6 +885,34 @@ let free_rels sigma m =
   in
   frec 1 Int.Set.empty m
 
+let free_rels_and_unqualified_refs sigma t =
+  let rec aux k (gseen, vseen, ids as accu) t =
+    match EConstr.kind sigma t with
+    | Const _ | Ind _ | Construct _ | Var _ ->
+      let g, _ = EConstr.destRef sigma t in
+      if not (GlobRef.Set_env.mem g gseen) then begin
+        try
+          let gseen = GlobRef.Set_env.add g gseen in
+          let short = Nametab.shortest_qualid_of_global Id.Set.empty g in
+          let dir, id = Libnames.repr_qualid short in
+          let ids = if DirPath.is_empty dir then Id.Set.add id ids else ids in
+          (gseen, vseen, ids)
+        with Not_found when !Flags.in_debugger || !Flags.in_toplevel ->
+          accu
+      end else
+        accu
+    | Rel p ->
+      if p > k && not (Int.Set.mem (p - k) vseen) then
+        let vseen = Int.Set.add (p - k) vseen in
+        (gseen, vseen, ids)
+      else
+        accu
+    | _ ->
+      EConstr.fold_with_binders sigma succ aux k accu t in
+  let accu = (GlobRef.Set_env.empty, Int.Set.empty, Id.Set.empty) in
+  let (_, rels, ids) = aux 0 accu t in
+  rels, ids
+
 (* collects all metavar occurrences, in left-to-right order, preserving
  * repetitions and all. *)
 
@@ -977,25 +1005,11 @@ let rec strip_outer_cast sigma c = match EConstr.kind sigma c with
   | Cast (c,_,_) -> strip_outer_cast sigma c
   | _ -> c
 
-(* flattens application lists throwing casts in-between *)
-let collapse_appl sigma c = match EConstr.kind sigma c with
-  | App (f,cl) ->
-    if EConstr.isCast sigma f then
-      let rec collapse_rec f cl2 =
-        match EConstr.kind sigma (strip_outer_cast sigma f) with
-        | App (g,cl1) -> collapse_rec g (Array.append cl1 cl2)
-        | _ -> EConstr.mkApp (f,cl2)
-      in
-      collapse_rec f cl
-    else c
-  | _ -> c
-
 (* First utilities for avoiding telescope computation for subst_term *)
 
 let prefix_application sigma eq_fun k l1 t =
   let open EConstr in
-  let t' = collapse_appl sigma t in
-  if 0 < l1 then match EConstr.kind sigma t' with
+  if 0 < l1 then match EConstr.kind sigma t with
     | App (f2,cl2) ->
         let l2 = Array.length cl2 in
         if l1 <= l2
@@ -1032,7 +1046,6 @@ let replace_term_gen sigma eq_fun ar by_c in_t =
 
 let replace_term sigma c byc t =
   let cache = ref Int.Map.empty in
-  let c = collapse_appl sigma c in
   let ar = Array.length (snd (decompose_app_vect sigma c)) in
   let eq sigma k t = eq_upto_lift cache c sigma k t in
   replace_term_gen sigma eq ar byc t
@@ -1084,8 +1097,8 @@ let ids_of_context env =
 let names_of_rel_context env =
   List.map RelDecl.get_name (rel_context env)
 
-let is_section_variable id =
-  try let _ = Global.lookup_named id in true
+let is_section_variable env id =
+  try let _ = Environ.lookup_named id env in true
   with Not_found -> false
 
 let global_of_constr sigma c =
@@ -1305,27 +1318,10 @@ let map_rel_context_in_env f env sign =
   in
   aux env [] (List.rev sign)
 
-let map_rel_context_with_binders f sign =
-  let rec aux k = function
-    | d::sign -> RelDecl.map_constr (f k) d :: aux (k-1) sign
-    | [] -> []
-  in
-  aux (Context.Rel.length sign) sign
-
-let substl_rel_context l =
-  map_rel_context_with_binders (fun k -> substnl l (k-1))
-
-let lift_rel_context n =
-  map_rel_context_with_binders (liftn n)
-
-let smash_rel_context sign =
-  let rec aux acc = function
-  | [] -> acc
-  | (RelDecl.LocalAssum _ as d) :: l -> aux (d::acc) l
-  | RelDecl.LocalDef (_,b,_) :: l ->
-      (* Quadratic in the number of let but there are probably a few of them *)
-      aux (List.rev (substl_rel_context [b] (List.rev acc))) l
-  in List.rev (aux [] sign)
+let map_rel_context_with_binders = Context.Rel.map_with_binders
+let substl_rel_context = Vars.substl_rel_context
+let lift_rel_context = Vars.lift_rel_context
+let smash_rel_context = Vars.smash_rel_context
 
 let fold_named_context_both_sides f l ~init = List.fold_right_and_left f l init
 
@@ -1377,8 +1373,6 @@ let global_vars_set env sigma constr =
   in
   filtrec Id.Set.empty constr
 
-let global_vars env sigma ids = Id.Set.elements (global_vars_set env sigma ids)
-
 let global_vars_set_of_decl env sigma = function
   | NamedDecl.LocalAssum (_,t) -> global_vars_set env sigma t
   | NamedDecl.LocalDef (_,c,t) ->
