@@ -20,11 +20,13 @@ open Constrexpr
 open Notation_term
 open Glob_term
 open Glob_ops
-open Context.Named.Declaration
 open NumTok
 
 (*i*)
 
+let notation_cat = Libobject.create_category "notations"
+
+
 (*s A scope is a set of notations; it includes
 
   - a set of ML interpreters/parsers for positive (e.g. 0, 1, 15, ...) and
@@ -166,7 +168,7 @@ let declare_scope scope =
     scope_map := String.Map.add scope empty_scope !scope_map
 
 let error_unknown_scope ~info sc =
-  user_err ~hdr:"Notation" ~info
+  user_err ~info
     (str "Scope " ++ str sc ++ str " is not declared.")
 
 let find_scope ?(tolerant=false) scope =
@@ -227,7 +229,7 @@ let scope_is_open sc = scope_is_open_in_scopes sc (!scope_stack)
 (* TODO: push nat_scope, z_scope, ... in scopes summary *)
 
 (* Exportation of scopes *)
-let open_scope i (_,(local,op,sc)) =
+let open_scope i (local,op,sc) =
   if Int.equal i 1 then
     scope_stack :=
       if op then sc :: !scope_stack
@@ -240,22 +242,22 @@ let subst_scope (subst,sc) = sc
 
 open Libobject
 
-let discharge_scope (_,(local,_,_ as o)) =
+let discharge_scope (local,_,_ as o) =
   if local then None else Some o
 
-let classify_scope (local,_,_ as o) =
-  if local then Dispose else Substitute o
+let classify_scope (local,_,_) =
+  if local then Dispose else Substitute
 
 let inScope : bool * bool * scope_item -> obj =
   declare_object {(default_object "SCOPE") with
       cache_function = cache_scope;
-      open_function = simple_open open_scope;
+      open_function = simple_open ~cat:notation_cat open_scope;
       subst_function = subst_scope;
       discharge_function = discharge_scope;
       classify_function = classify_scope }
 
 let open_close_scope (local,opening,sc) =
-  Lib.add_anonymous_leaf (inScope (local,opening,OpenScopeItem (normalize_scope sc)))
+  Lib.add_leaf (inScope (local,opening,OpenScopeItem (normalize_scope sc)))
 
 let empty_scope_stack = []
 
@@ -308,15 +310,11 @@ let remove_delimiters scope =
 let find_delimiters_scope ?loc key =
   try String.Map.find key !delimiters_map
   with Not_found ->
-    user_err ?loc ~hdr:"find_delimiters"
+    user_err ?loc
       (str "Unknown scope delimiting key " ++ str key ++ str ".")
 
 (* Uninterpretation tables *)
 
-type interp_rule =
-  | NotationRule of specific_notation
-  | SynDefRule of KerName.t
-
 (* We define keys for glob_constr and aconstr to split the syntax entries
    according to the key of the pattern (adapted from Chet Murthy by HH) *)
 
@@ -380,6 +378,7 @@ let glob_prim_constr_key c = match DAst.get c with
     | GRef (ref, _) -> Some (canonical_gr ref)
     | _ -> None
     end
+  | GProj ((cst,_), _, _) -> Some (canonical_gr (GlobRef.ConstRef cst))
   | _ -> None
 
 let glob_constr_keys c = match DAst.get c with
@@ -388,6 +387,7 @@ let glob_constr_keys c = match DAst.get c with
     | GRef (ref, _) -> [RefKey (canonical_gr ref); Oth]
     | _ -> [Oth]
     end
+  | GProj ((cst,_), _, _) -> [RefKey (canonical_gr (GlobRef.ConstRef cst))]
   | GRef (ref,_) -> [RefKey (canonical_gr ref)]
   | _ -> [Oth]
 
@@ -397,6 +397,7 @@ let cases_pattern_key c = match DAst.get c with
 
 let notation_constr_key = function (* Rem: NApp(NRef ref,[]) stands for @ref *)
   | NApp (NRef (ref,_),args) -> RefKey(canonical_gr ref), AppBoundedNotation (List.length args)
+  | NProj ((cst,_),args,_) -> RefKey(canonical_gr (GlobRef.ConstRef cst)), AppBoundedNotation (List.length args + 1)
   | NList (_,_,NApp (NRef (ref,_),args),_,_)
   | NBinderList (_,_,NApp (NRef (ref,_),args),_,_) ->
       RefKey (canonical_gr ref), AppBoundedNotation (List.length args)
@@ -552,6 +553,7 @@ type target_kind =
   | UInt of int_ty (* Coq.Init.Number.uint *)
   | Z of z_pos_ty (* Coq.Numbers.BinNums.Z and positive *)
   | Int63 of pos_neg_int63_ty (* Coq.Numbers.Cyclic.Int63.PrimInt63.pos_neg_int63 *)
+  | Float64 (* Coq.Floats.PrimFloat.float *)
   | Number of number_ty (* Coq.Init.Number.number + uint + int *)
 
 type string_target_kind =
@@ -576,7 +578,7 @@ type 'target conversion_kind = 'target * option_kind
    [ToPostCheck r] behaves as [ToPostCopy] except in the reverse
    translation which fails if the copied term is not [r].
    When [n] is null, no translation is performed. *)
-type to_post_arg = ToPostCopy | ToPostAs of int | ToPostHole | ToPostCheck of GlobRef.t
+type to_post_arg = ToPostCopy | ToPostAs of int | ToPostHole | ToPostCheck of Constr.t
 type ('target, 'warning) prim_token_notation_obj =
   { to_kind : 'target conversion_kind;
     to_ty : GlobRef.t;
@@ -611,6 +613,7 @@ module PrimTokenNotation = struct
 
 let eval_constr env sigma (c : Constr.t) =
   let c = EConstr.of_constr c in
+  let sigma, _ = Typing.type_of env sigma c in
   let c' = Tacred.compute env sigma c in
   EConstr.Unsafe.to_constr c'
 
@@ -629,7 +632,7 @@ exception NotAValidPrimToken
     what it means for a term to be ground / to be able to be
     considered for parsing. *)
 
-let constr_of_globref allow_constant env sigma = function
+let constr_of_globref ?(allow_constant=true) env sigma = function
   | GlobRef.ConstructRef c ->
      let sigma,c = Evd.fresh_constructor_instance env sigma c in
      sigma,mkConstructU c
@@ -641,17 +644,49 @@ let constr_of_globref allow_constant env sigma = function
      sigma,mkConstU c
   | _ -> raise NotAValidPrimToken
 
-let rec constr_of_glob allow_constant to_post post env sigma g = match DAst.get g with
+(** [check_glob g c] checks that glob [g] is equal to constr [c]
+    and returns [g] as a constr (with fresh universe instances)
+    or raises [NotAValidPrimToken]. *)
+let rec check_glob env sigma g c = match DAst.get g, Constr.kind c with
+  | Glob_term.GRef (GlobRef.ConstructRef c as g, _), Constr.Construct (c', _)
+       when Construct.CanOrd.equal c c' -> constr_of_globref env sigma g
+  | Glob_term.GRef (GlobRef.IndRef c as g, _), Constr.Ind (c', _)
+       when Ind.CanOrd.equal c c' -> constr_of_globref env sigma g
+  | Glob_term.GRef (GlobRef.ConstRef c as g, _), Constr.Const (c', _)
+       when Constant.CanOrd.equal c c' -> constr_of_globref env sigma g
+  | Glob_term.GApp (gc, gcl), Constr.App (gc', gc'a) ->
+     let sigma,c = check_glob env sigma gc gc' in
+     let sigma,cl =
+       try List.fold_left2_map (check_glob env) sigma gcl (Array.to_list gc'a)
+       with Invalid_argument _ -> raise NotAValidPrimToken in
+     sigma, mkApp (c, Array.of_list cl)
+  | Glob_term.GInt i, Constr.Int i' when Uint63.equal i i' -> sigma, mkInt i
+  | Glob_term.GFloat f, Constr.Float f' when Float64.equal f f' -> sigma, mkFloat f
+  | Glob_term.GArray (_,t,def,ty), Constr.Array (_,t',def',ty') ->
+     let sigma,u = Evd.fresh_array_instance env sigma in
+     let sigma,def = check_glob env sigma def def' in
+     let sigma,t =
+       try Array.fold_left2_map (check_glob env) sigma t t'
+       with Invalid_argument _ -> raise NotAValidPrimToken in
+     let sigma,ty = check_glob env sigma ty ty' in
+     sigma, mkArray (u,t,def,ty)
+  | Glob_term.GSort s, Constr.Sort s' ->
+     let sigma,s = Evd.fresh_sort_in_family sigma (Glob_ops.glob_sort_family s) in
+     if not (Sorts.equal s s') then raise NotAValidPrimToken;
+     sigma,mkSort s
+  | _ -> raise NotAValidPrimToken
+
+let rec constr_of_glob to_post post env sigma g = match DAst.get g with
   | Glob_term.GRef (r, _) ->
       let o = List.find_opt (fun (_,r',_) -> GlobRef.equal r r') post in
       begin match o with
-      | None -> constr_of_globref allow_constant env sigma r
+      | None -> constr_of_globref ~allow_constant:false env sigma r
       | Some (r, _, a) ->
          (* [g] is not a GApp so check that [post]
             does not expect any actual argument
             (i.e., [a] contains only ToPostHole since they mean "ignore arg") *)
          if List.exists ((<>) ToPostHole) a then raise NotAValidPrimToken;
-         constr_of_globref true env sigma r
+         constr_of_globref env sigma r
       end
   | Glob_term.GApp (gc, gcl) ->
       let o = match DAst.get gc with
@@ -659,26 +694,23 @@ let rec constr_of_glob allow_constant to_post post env sigma g = match DAst.get
         | _ -> None in
       begin match o with
       | None ->
-         let sigma,c = constr_of_glob allow_constant to_post post env sigma gc in
-         let sigma,cl = List.fold_left_map (constr_of_glob allow_constant to_post post env) sigma gcl in
+         let sigma,c = constr_of_glob to_post post env sigma gc in
+         let sigma,cl = List.fold_left_map (constr_of_glob to_post post env) sigma gcl in
          sigma,mkApp (c, Array.of_list cl)
       | Some (r, _, a) ->
-         let sigma,c = constr_of_globref true env sigma r in
+         let sigma,c = constr_of_globref env sigma r in
          let rec aux sigma a gcl = match a, gcl with
            | [], [] -> sigma,[]
            | ToPostCopy :: a, gc :: gcl ->
-              let sigma,c = constr_of_glob allow_constant [||] [] env sigma gc in
+              let sigma,c = constr_of_glob [||] [] env sigma gc in
               let sigma,cl = aux sigma a gcl in
               sigma, c :: cl
            | ToPostCheck r :: a, gc :: gcl ->
-              let () = match DAst.get gc with
-                | Glob_term.GRef (r', _) when GlobRef.equal r r' -> ()
-                | _ -> raise NotAValidPrimToken in
-              let sigma,c = constr_of_glob true [||] [] env sigma gc in
+              let sigma,c = check_glob env sigma gc r in
               let sigma,cl = aux sigma a gcl in
               sigma, c :: cl
            | ToPostAs i :: a, gc :: gcl ->
-              let sigma,c = constr_of_glob allow_constant to_post to_post.(i) env sigma gc in
+              let sigma,c = constr_of_glob to_post to_post.(i) env sigma gc in
               let sigma,cl = aux sigma a gcl in
               sigma, c :: cl
            | ToPostHole :: post, _ :: gcl -> aux sigma post gcl
@@ -691,9 +723,9 @@ let rec constr_of_glob allow_constant to_post post env sigma g = match DAst.get
   | Glob_term.GFloat f -> sigma, mkFloat f
   | Glob_term.GArray (_,t,def,ty) ->
       let sigma, u' = Evd.fresh_array_instance env sigma in
-      let sigma, def' = constr_of_glob allow_constant to_post post env sigma def in
-      let sigma, t' = Array.fold_left_map (constr_of_glob allow_constant to_post post env) sigma t in
-      let sigma, ty' = constr_of_glob allow_constant to_post post env sigma ty in
+      let sigma, def' = constr_of_glob to_post post env sigma def in
+      let sigma, t' = Array.fold_left_map (constr_of_glob to_post post env) sigma t in
+      let sigma, ty' = constr_of_glob to_post post env sigma ty in
        sigma, mkArray (u',t',def',ty')
   | Glob_term.GSort gs ->
       let sigma,c = Evd.fresh_sort_in_family sigma (Glob_ops.glob_sort_family gs) in
@@ -703,7 +735,7 @@ let rec constr_of_glob allow_constant to_post post env sigma g = match DAst.get
 
 let constr_of_glob to_post env sigma (Glob_term.AnyGlobConstr g) =
   let post = match to_post with [||] -> [] | _ -> to_post.(0) in
-  constr_of_glob false to_post post env sigma g
+  constr_of_glob to_post post env sigma g
 
 let rec glob_of_constr token_kind ?loc env sigma c = match Constr.kind c with
   | App (c, ca) ->
@@ -1006,7 +1038,8 @@ let int63_of_pos_bigint ?loc n =
   mkInt i
 
 let error_overflow ?loc n =
-  CErrors.user_err ?loc ~hdr:"interp_int63" Pp.(str "overflow in int63 literal: " ++ str (Z.to_string n))
+  CErrors.user_err ?loc Pp.(str "Overflow in int63 literal: " ++ str (Z.to_string n)
+    ++ str ".")
 
 let coqpos_neg_int63_of_bigint ?loc ind (sign,n) =
   let uint = int63_of_pos_bigint ?loc n in
@@ -1020,6 +1053,58 @@ let interp_int63 ?loc ind n =
   then coqpos_neg_int63_of_bigint ?loc ind (sign,an)
   else error_overflow ?loc n
 
+let warn_inexact_float =
+  CWarnings.create ~name:"inexact-float" ~category:"parsing"
+    (fun (sn, f) ->
+      Pp.strbrk
+        (Printf.sprintf
+           "The constant %s is not a binary64 floating-point value. \
+            A closest value %s will be used and unambiguously printed %s."
+           sn (Float64.to_hex_string f) (Float64.to_string f)))
+
+let interp_float64 ?loc n =
+  let sn = NumTok.Signed.to_string n in
+  let f = Float64.of_string sn in
+  (* return true when f is not exactly equal to n,
+     this is only used to decide whether or not to display a warning
+     and does not play any actual role in the parsing *)
+  let inexact () = match Float64.classify f with
+    | Float64.(PInf | NInf | NaN) -> true
+    | Float64.(PZero | NZero) -> not (NumTok.Signed.is_zero n)
+    | Float64.(PNormal | NNormal | PSubn | NSubn) ->
+       let m, e =
+         let (_, i), f, e = NumTok.Signed.to_int_frac_and_exponent n in
+         let i = NumTok.UnsignedNat.to_string i in
+         let f = match f with
+           | None -> "" | Some f -> NumTok.UnsignedNat.to_string f in
+         let e = match e with
+           | None -> "0" | Some e -> NumTok.SignedNat.to_string e in
+         Z.of_string (i ^ f),
+         (try int_of_string e with Failure _ -> 0) - String.length f in
+       let m', e' =
+         let m', e' = Float64.frshiftexp f in
+         let m' = Float64.normfr_mantissa m' in
+         let e' = Uint63.to_int_min e' 4096 - Float64.eshift - 53 in
+         Z.of_string (Uint63.to_string m'),
+         e' in
+       let c2, c5 = Z.(of_int 2, of_int 5) in
+       (* check m*5^e <> m'*2^e' *)
+       let check m e m' e' =
+         not (Z.(equal (mul m (pow c5 e)) (mul m' (pow c2 e')))) in
+       (* check m*5^e*2^e' <> m' *)
+       let check' m e e' m' =
+         not (Z.(equal (mul (mul m (pow c5 e)) (pow c2 e')) m')) in
+       (* we now have to check m*10^e <> m'*2^e' *)
+       if e >= 0 then
+         if e <= e' then check m e m' (e' - e)
+         else check' m e (e - e') m'
+       else  (* e < 0 *)
+         if e' <= e then check m' (-e) m (e - e')
+         else check' m' (-e) (e' - e) m in
+  if NumTok.(Signed.classify n = CDec) && inexact () then
+    warn_inexact_float ?loc (sn, f);
+  mkFloat f
+
 let bigint_of_int63 c =
   match Constr.kind c with
   | Int i -> Z.of_int64 (Uint63.to_int64 i)
@@ -1034,6 +1119,18 @@ let bigint_of_coqpos_neg_int63 c =
       | _ -> raise NotAValidPrimToken)
   | _ -> raise NotAValidPrimToken
 
+let get_printing_float = Goptions.declare_bool_option_and_ref
+    ~depr:false
+    ~key:["Printing";"Float"]
+    ~value:true
+
+let uninterp_float64 c =
+  match Constr.kind c with
+  | Float f when not (Float64.is_infinity f || Float64.is_neg_infinity f
+                      || Float64.is_nan f) && get_printing_float () ->
+     NumTok.Signed.of_string (Float64.to_string f)
+  | _ -> raise NotAValidPrimToken
+
 let interp o ?loc n =
   begin match o.warning, n with
   | Warning threshold, n when NumTok.Signed.is_bigger_int_than n threshold ->
@@ -1051,6 +1148,7 @@ let interp o ?loc n =
        interp_int63 ?loc pos_neg_int63_ty.pos_neg_int63_ty (NumTok.SignedNat.to_bigint n)
     | (Int _ | UInt _ | Z _ | Int63 _), _ ->
        no_such_prim_token "number" ?loc o.ty_name
+    | Float64, _ -> interp_float64 ?loc n
     | Number number_ty, _ -> coqnumber_of_rawnum number_ty n
   in
   let env = Global.env () in
@@ -1075,6 +1173,7 @@ let uninterp o n =
       | (UInt _, c) -> NumTok.Signed.of_nat (rawnum_of_coquint c)
       | (Z _, c) -> NumTok.Signed.of_bigint CDec (bigint_of_z c)
       | (Int63 _, c) -> NumTok.Signed.of_bigint CDec (bigint_of_coqpos_neg_int63 c)
+      | (Float64, c) -> uninterp_float64 c
       | (Number _, c) -> rawnum_of_coqnumber c
     end o n
 end
@@ -1112,7 +1211,7 @@ let coqbyte_of_string ?loc byte s =
         if Int.equal (String.length s) 3 && is_digit s.[0] && is_digit s.[1] && is_digit s.[2]
         then int_of_string s else 256 in
       if n < 256 then n else
-       user_err ?loc ~hdr:"coqbyte_of_string"
+       user_err ?loc
          (str "Expects a single character or a three-digit ASCII code.") in
   coqbyte_of_char_code byte p
 
@@ -1217,7 +1316,7 @@ let hashtbl_check_and_set allow_overwrite uid f h eq =
    | _ when allow_overwrite -> Hashtbl.add h uid f
    | g when eq f g -> ()
    | _ ->
-      user_err ~hdr:"prim_token_interpreter"
+      user_err
        (str "Unique identifier " ++ str uid ++
         str " already used to register a number or string (un)interpreter.")
 
@@ -1239,7 +1338,7 @@ let register_string_interpretation ?(allow_overwrite=false) uid (interp, uninter
   register_gen_interpretation allow_overwrite uid
     (InnerPrimToken.StringInterp interp, InnerPrimToken.StringUninterp uninterp)
 
-let cache_prim_token_interpretation (_,infos) =
+let cache_prim_token_interpretation infos =
   let ptii = infos.pt_interp_info in
   let sc = infos.pt_scope in
   check_scope ~tolerant:true sc;
@@ -1257,20 +1356,20 @@ let subst_prim_token_interpretation (subs,infos) =
     pt_refs = List.map (subst_global_reference subs) infos.pt_refs }
 
 let classify_prim_token_interpretation infos =
-    if infos.pt_local then Dispose else Substitute infos
+    if infos.pt_local then Dispose else Substitute
 
 let open_prim_token_interpretation i o =
   if Int.equal i 1 then cache_prim_token_interpretation o
 
 let inPrimTokenInterp : prim_token_infos -> obj =
   declare_object {(default_object "PRIM-TOKEN-INTERP") with
-     open_function  = simple_open open_prim_token_interpretation;
+     open_function  = simple_open ~cat:notation_cat open_prim_token_interpretation;
      cache_function = cache_prim_token_interpretation;
      subst_function = subst_prim_token_interpretation;
      classify_function = classify_prim_token_interpretation}
 
 let enable_prim_token_interpretation infos =
-  Lib.add_anonymous_leaf (inPrimTokenInterp infos)
+  Lib.add_leaf (inPrimTokenInterp infos)
 
 (** Compatibility.
     Avoid the next two functions, they will now store unnecessary
@@ -1311,11 +1410,11 @@ let check_required_module ?loc sc (sp,d) =
     let _, info = Exninfo.capture exn in
     match d with
     | [] ->
-      user_err ?loc ~info ~hdr:"prim_token_interpreter"
+      user_err ?loc ~info
         (str "Cannot interpret in " ++ str sc ++ str " because " ++ pr_path sp ++
          str " could not be found in the current environment.")
     | _ ->
-      user_err ?loc ~info ~hdr:"prim_token_interpreter"
+      user_err ?loc ~info
         (str "Cannot interpret in " ++ str sc ++ str " without requiring first module " ++
          str (List.last d) ++ str ".")
 
@@ -1509,7 +1608,7 @@ let interp_prim_token_gen ?loc g p local_scopes =
     pat, (loc,sc)
   with Not_found as exn ->
     let _, info = Exninfo.capture exn in
-    user_err ?loc ~info ~hdr:"interp_prim_token"
+    user_err ?loc ~info
     ((match p with
       | Number _ ->
          str "No interpretation for number " ++ pr_notation (notation_of_prim_token p)
@@ -1586,8 +1685,8 @@ let is_printing_inactive_rule rule pat =
   match rule with
   | NotationRule (scope,ntn) ->
     not (is_printing_active_in_scope (scope,ntn) pat)
-  | SynDefRule kn ->
-    try let _ = Nametab.path_of_syndef kn in false with Not_found -> true
+  | AbbrevRule kn ->
+    try let _ = Nametab.path_of_abbreviation kn in false with Not_found -> true
 
 let availability_of_notation (ntn_scope,ntn) scopes =
   find_without_delimiters (has_active_parsing_rule_in_scope ntn) (ntn_scope,Some ntn) (make_current_scopes scopes)
@@ -1730,19 +1829,13 @@ let declare_notation (scopt,ntn) pat df ~use ~also_in_cases_pattern coe deprecat
   let notation_update,printing_update = update_notation_data (scopt,ntn) use notdata sc.notations in
   let sc = { sc with notations = notation_update } in
   scope_map := String.Map.add scope sc !scope_map;
-  (* Update the uninterpretation cache *)
-  begin match printing_update with
-  | Some pat -> remove_uninterpretation (NotationRule (scopt,ntn)) pat
-  | None -> ()
-  end;
-  if use <> OnlyParsing then declare_uninterpretation ~also_in_cases_pattern (NotationRule (scopt,ntn)) pat;
   (* Register visibility of lonely notations *)
   begin match scopt with
   | LastLonelyNotation -> scope_stack := LonelyNotationItem ntn :: !scope_stack
   | NotationInScope _ -> ()
   end;
   (* Declare a possible coercion *)
-  begin match coe with
+  if use <> OnlyParsing then begin match coe with
    | Some (IsEntryCoercion entry) ->
      let (_,level,_) = level_of_notation ntn in
      let level = match fst ntn with
@@ -1752,7 +1845,13 @@ let declare_notation (scopt,ntn) pat df ~use ~also_in_cases_pattern coe deprecat
      declare_entry_coercion (scopt,ntn) level entry
    | Some (IsEntryGlobal (entry,n)) -> declare_custom_entry_has_global entry n
    | Some (IsEntryIdent (entry,n)) -> declare_custom_entry_has_ident entry n
-   | None -> ()
+   | None ->
+     (* Update the uninterpretation cache *)
+     begin match printing_update with
+     | Some pat -> remove_uninterpretation (NotationRule (scopt,ntn)) pat
+     | None -> ()
+     end;
+     declare_uninterpretation ~also_in_cases_pattern (NotationRule (scopt,ntn)) pat
   end
 
 let availability_of_prim_token n printer_scope local_scopes =
@@ -1877,8 +1976,8 @@ let find_scope_class_opt = function
 let rec compute_arguments_classes env sigma t =
   match EConstr.kind sigma (Reductionops.whd_betaiotazeta env sigma t) with
     | Prod (na, t, u) ->
-        let env = EConstr.push_rel (Context.Rel.Declaration.LocalAssum (na, t)) env in
         let cl = try Some (compute_scope_class env sigma t) with Not_found -> None in
+        let env = EConstr.push_rel (Context.Rel.Declaration.LocalAssum (na, t)) env in
         cl :: compute_arguments_classes env sigma u
     | _ -> []
 
@@ -1920,7 +2019,7 @@ type arguments_scope_discharge_request =
   | ArgsScopeManual
   | ArgsScopeNoDischarge
 
-let load_arguments_scope _ (_,(_,r,n,scl,cls)) =
+let load_arguments_scope _ (_,r,n,scl,cls) =
   List.iter (Option.iter check_scope) scl;
   let initial_stamp = ScopeClassMap.empty in
   arguments_scope := GlobRef.Map.add r (scl,cls,initial_stamp) !arguments_scope
@@ -1943,19 +2042,18 @@ let subst_arguments_scope (subst,(req,r,n,scl,cls)) =
   let cls' = List.Smart.map subst_cl cls in
   (ArgsScopeNoDischarge,r',n,scl,cls')
 
-let discharge_arguments_scope (_,(req,r,n,l,_)) =
+let discharge_arguments_scope (req,r,n,l,_) =
   if req == ArgsScopeNoDischarge || (isVarRef r && Lib.is_in_section r) then None
   else
     let n =
       try
-        let vars = Lib.variable_section_segment_of_reference r in
-        vars |> List.filter is_local_assum |> List.length
+        Array.length (Lib.section_instance r)
       with
         Not_found (* Not a ref defined in this section *) -> 0 in
     Some (req,r,n,l,[])
 
-let classify_arguments_scope (req,_,_,_,_ as obj) =
-  if req == ArgsScopeNoDischarge then Dispose else Substitute obj
+let classify_arguments_scope (req,_,_,_,_) =
+  if req == ArgsScopeNoDischarge then Dispose else Substitute
 
 let rebuild_arguments_scope sigma (req,r,n,l,_) =
   match req with
@@ -1995,7 +2093,7 @@ let inArgumentsScope : arguments_scope_obj -> obj =
 let is_local local ref = local || isVarRef ref && Lib.is_in_section ref
 
 let declare_arguments_scope_gen req r n (scl,cls) =
-  Lib.add_anonymous_leaf (inArgumentsScope (req,r,n,scl,cls))
+  Lib.add_leaf (inArgumentsScope (req,r,n,scl,cls))
 
 let declare_arguments_scope local r scl =
   let req = if is_local local r then ArgsScopeNoDischarge else ArgsScopeManual in
@@ -2208,6 +2306,12 @@ let rec raw_analyze_anonymous_notation_tokens = function
 
 (* Interpret notations with a recursive component *)
 
+type notation_symbols = {
+  recvars : (Id.t * Id.t) list; (* pairs (x,y) as in [ x ; .. ; y ] *)
+  mainvars : Id.t list; (* variables non involved in a recursive pattern *)
+  symbols : symbol list; (* the decomposition of the notation into terminals and nonterminals *)
+}
+
 let out_nt = function NonTerminal x -> x | _ -> assert false
 
 let msg_expected_form_of_recursive_notation =
@@ -2223,7 +2327,7 @@ let rec find_pattern nt xl = function
   | _, Break s :: _ | Break s :: _, _ ->
       user_err Pp.(str ("A break occurs on one side of \"..\" but not on the other side."))
   | _, Terminal s :: _ | Terminal s :: _, _ ->
-      user_err ~hdr:"Metasyntax.find_pattern"
+      user_err
         (str "The token \"" ++ str s ++ str "\" occurs on one side of \"..\" but not on the other side.")
   | _, [] ->
       user_err Pp.(str msg_expected_form_of_recursive_notation)
@@ -2256,10 +2360,10 @@ let get_notation_vars l =
 
 let decompose_raw_notation ntn =
   let l = split_notation_string ntn in
-  let l = raw_analyze_notation_tokens l in
-  let recvars,l = interp_list_parser [] l in
-  let vars = get_notation_vars l in
-  recvars, vars, l
+  let symbols = raw_analyze_notation_tokens l in
+  let recvars, symbols = interp_list_parser [] symbols in
+  let mainvars = get_notation_vars symbols in
+  {recvars; mainvars; symbols}
 
 let interpret_notation_string ntn =
   (* We collect the possible interpretations of a notation string depending on whether it is
