@@ -34,9 +34,7 @@ type existential_key = Evar.t
 type metavariable = int
 
 (* This defines the strategy to use for verifiying a Cast *)
-(* Warning: REVERTcast is not exported to vo-files; as of r14492, it has to *)
-(* come after the vo-exported cast_kind so as to be compatible with coqchk *)
-type cast_kind = VMcast | NATIVEcast | DEFAULTcast | REVERTcast
+type cast_kind = VMcast | NATIVEcast | DEFAULTcast
 
 (* This defines Cases annotations *)
 type case_style = LetStyle | IfStyle | LetPatternStyle | MatchStyle | RegularStyle
@@ -696,7 +694,7 @@ let map f c = match kind c with
 let fold_map_invert f acc = function
   | NoInvert -> acc, NoInvert
   | CaseInvert {indices;} as orig ->
-    let acc, indices' = Array.fold_left_map f acc indices in
+    let acc, indices' = Array.Smart.fold_left_map f acc indices in
     if indices==indices' then acc, orig
     else acc, CaseInvert {indices=indices';}
 
@@ -746,8 +744,7 @@ let fold_map f accu c = match kind c with
       if t' == t then accu, c
       else accu, mkProj (p, t')
   | Evar (e,l) ->
-    (* Doesn't matter, we should not hashcons evars anyways *)
-      let accu, l' = List.fold_left_map f accu l in
+      let accu, l' = List.Smart.fold_left_map f accu l in
       if l'==l then accu, c
       else accu, mkEvar (e, l')
   | Case (ci,u,pms,p,iv,b,bl) ->
@@ -852,7 +849,9 @@ let map_with_binders g f l c0 = match kind c0 with
 let rec exliftn el c =
   let open Esubst in
   match kind c with
-  | Rel i -> mkRel(reloc_rel i el)
+  | Rel i ->
+    let j = reloc_rel i el in
+    if Int.equal i j then c else mkRel j
   | _ -> map_with_binders el_lift exliftn el c
 
 (* Lifting the binding depth across k bindings *)
@@ -976,7 +975,7 @@ let eq_constr_univs univs m n =
   if m == n then true
   else
     let eq_universes _ = UGraph.check_eq_instances univs in
-    let eq_sorts s1 s2 = s1 == s2 || UGraph.check_eq univs (Sorts.univ_of_sort s1) (Sorts.univ_of_sort s2) in
+    let eq_sorts s1 s2 = s1 == s2 || UGraph.check_eq_sort univs s1 s2 in
     let rec eq_constr' nargs m n =
       m == n ||	compare_head_gen eq_universes eq_sorts eq_constr' nargs m n
     in compare_head_gen eq_universes eq_sorts eq_constr' 0 m n
@@ -986,9 +985,9 @@ let leq_constr_univs univs m n =
   else
     let eq_universes _ = UGraph.check_eq_instances univs in
     let eq_sorts s1 s2 = s1 == s2 ||
-      UGraph.check_eq univs (Sorts.univ_of_sort s1) (Sorts.univ_of_sort s2) in
+      UGraph.check_eq_sort univs s1 s2 in
     let leq_sorts s1 s2 = s1 == s2 ||
-      UGraph.check_leq univs (Sorts.univ_of_sort s1) (Sorts.univ_of_sort s2) in
+      UGraph.check_leq_sort univs s1 s2 in
     let rec eq_constr' nargs m n =
       m == n || compare_head_gen eq_universes eq_sorts eq_constr' nargs m n
     in
@@ -997,59 +996,6 @@ let leq_constr_univs univs m n =
     and leq_constr' nargs m n = m == n || compare_leq nargs m n in
     compare_leq 0 m n
 
-let eq_constr_univs_infer univs m n =
-  if m == n then true, Constraint.empty
-  else
-    let cstrs = ref Constraint.empty in
-    let eq_universes _ = UGraph.check_eq_instances univs in
-    let eq_sorts s1 s2 =
-      if Sorts.equal s1 s2 then true
-      else
-        let u1 = Sorts.univ_of_sort s1 and u2 = Sorts.univ_of_sort s2 in
-        if UGraph.check_eq univs u1 u2 then true
-        else
-          (cstrs := Univ.enforce_eq u1 u2 !cstrs;
-           true)
-    in
-    let rec eq_constr' nargs m n =
-      m == n || compare_head_gen eq_universes eq_sorts eq_constr' nargs m n
-    in
-    let res = compare_head_gen eq_universes eq_sorts eq_constr' 0 m n in
-    res, !cstrs
-
-let leq_constr_univs_infer univs m n =
-  if m == n then true, Constraint.empty
-  else
-    let cstrs = ref Constraint.empty in
-    let eq_universes _ l l' = UGraph.check_eq_instances univs l l' in
-    let eq_sorts s1 s2 =
-      if Sorts.equal s1 s2 then true
-      else
-        let u1 = Sorts.univ_of_sort s1 and u2 = Sorts.univ_of_sort s2 in
-        if UGraph.check_eq univs u1 u2 then true
-        else (cstrs := Univ.enforce_eq u1 u2 !cstrs;
-              true)
-    in
-    let leq_sorts s1 s2 =
-      if Sorts.equal s1 s2 then true
-      else
-        let u1 = Sorts.univ_of_sort s1 and u2 = Sorts.univ_of_sort s2 in
-        if UGraph.check_leq univs u1 u2 then true
-        else
-          (try let c, _ = UGraph.enforce_leq_alg u1 u2 univs in
-            cstrs := Univ.Constraint.union c !cstrs;
-            true
-          with Univ.UniverseInconsistency _ -> false)
-    in
-    let rec eq_constr' nargs m n =
-      m == n || compare_head_gen eq_universes eq_sorts eq_constr' nargs m n
-    in
-    let rec compare_leq nargs m n =
-      compare_head_gen_leq eq_universes leq_sorts eq_constr' leq_constr' nargs m n
-    and leq_constr' nargs m n = m == n || compare_leq nargs m n in
-    let res = compare_leq 0 m n in
-    res, !cstrs
-
 let rec eq_constr_nounivs m n =
   (m == n) || compare_head_gen (fun _ _ _ -> true) (fun _ _ -> true) (fun _ -> eq_constr_nounivs) 0 m n
 
@@ -1256,7 +1202,6 @@ let hash_cast_kind = function
 | VMcast -> 0
 | NATIVEcast -> 1
 | DEFAULTcast -> 2
-| REVERTcast -> 3
 
 let sh_instance = Univ.Instance.share
 
