@@ -184,7 +184,7 @@ struct
 
   module Hdir = Hashcons.Hlist(Id)
 
-  let hcons = Hashcons.recursive_hcons Hdir.generate Hdir.hcons Id.hcons
+  let hcons = Hashcons.simple_hcons Hdir.generate Hdir.hcons Id.hcons
 
 end
 
@@ -499,7 +499,6 @@ module KerPair = struct
 
   let make1 = same
   let make2 mp l = same (KerName.make mp l)
-  let repr2 kp = KerName.repr (user kp)
   let label kp = KerName.label (user kp)
   let modpath kp = KerName.modpath (user kp)
 
@@ -587,6 +586,8 @@ module KerPair = struct
 
   module HashKP = Hashcons.Make(Self_Hashcons)
 
+  let hcons = Hashcons.simple_hcons HashKP.generate HashKP.hcons KerName.hcons
+
 end
 
 (** {6 Constant Names} *)
@@ -775,8 +776,8 @@ module Hconstruct = Hashcons.Make(
     let hash = constructor_hash
   end)
 
-let hcons_con = Hashcons.simple_hcons Constant.HashKP.generate Constant.HashKP.hcons KerName.hcons
-let hcons_mind = Hashcons.simple_hcons MutInd.HashKP.generate MutInd.HashKP.hcons KerName.hcons
+let hcons_con = Constant.hcons
+let hcons_mind = MutInd.hcons
 let hcons_ind = Hashcons.simple_hcons Hind.generate Hind.hcons hcons_mind
 let hcons_construct = Hashcons.simple_hcons Hconstruct.generate Hconstruct.hcons hcons_ind
 
@@ -821,12 +822,13 @@ struct
   module Repr = struct
     type t =
       { proj_ind : inductive;
+        proj_relevant : bool;
         proj_npars : int;
         proj_arg : int;
         proj_name : Label.t; }
 
-    let make proj_ind ~proj_npars ~proj_arg proj_name =
-      {proj_ind;proj_npars;proj_arg;proj_name}
+    let make proj_ind ~proj_npars ~proj_arg ~proj_relevant proj_name =
+      {proj_ind;proj_npars;proj_arg;proj_relevant;proj_name}
 
     let inductive c = c.proj_ind
 
@@ -840,22 +842,36 @@ struct
 
     let arg c = c.proj_arg
 
-    let equal a b =
-      eq_ind a.proj_ind b.proj_ind && Int.equal a.proj_arg b.proj_arg
+    let relevant c = c.proj_relevant
 
     let hash p =
       Hashset.Combine.combinesmall p.proj_arg (ind_hash p.proj_ind)
 
+    let bcomp b1 b2 = match b1, b2 with
+    | true, true | false, false -> 0
+    | true, false -> 1
+    | false, true -> -1
+
+    let compare_gen a b =
+      let c = Int.compare a.proj_arg b.proj_arg in
+      if c <> 0 then c
+      else
+        let c = Int.compare a.proj_npars b.proj_npars in
+        if c <> 0 then c
+        else
+          let c = Label.compare a.proj_name b.proj_name in
+          if c <> 0 then c
+          else bcomp a.proj_relevant b.proj_relevant
+
     module SyntacticOrd = struct
       type nonrec t = t
 
       let compare a b =
         let c = ind_syntactic_ord a.proj_ind b.proj_ind in
-        if c == 0 then Int.compare a.proj_arg b.proj_arg
-        else c
+        if c <> 0 then c
+        else compare_gen a b
 
-      let equal a b =
-        a.proj_arg == b.proj_arg && eq_syntactic_ind a.proj_ind b.proj_ind
+      let equal a b = compare a b == 0
 
       let hash p =
         Hashset.Combine.combinesmall p.proj_arg (ind_hash p.proj_ind)
@@ -865,11 +881,10 @@ struct
 
       let compare a b =
         let c = ind_ord a.proj_ind b.proj_ind in
-        if c == 0 then Int.compare a.proj_arg b.proj_arg
-        else c
+        if c <> 0 then c
+        else compare_gen a b
 
-      let equal a b =
-        a.proj_arg == b.proj_arg && eq_ind a.proj_ind b.proj_ind
+      let equal a b = compare a b == 0
 
       let hash p =
         Hashset.Combine.combinesmall p.proj_arg (ind_hash p.proj_ind)
@@ -879,44 +894,45 @@ struct
 
       let compare a b =
         let c = ind_user_ord a.proj_ind b.proj_ind in
-        if c == 0 then Int.compare a.proj_arg b.proj_arg
-        else c
+        if c <> 0 then c
+        else compare_gen a b
 
-      let equal a b =
-        a.proj_arg == b.proj_arg && eq_user_ind a.proj_ind b.proj_ind
+      let equal a b = compare a b == 0
 
       let hash p =
         Hashset.Combine.combinesmall p.proj_arg (ind_user_hash p.proj_ind)
     end
 
-    let compare a b =
-      let c = ind_ord a.proj_ind b.proj_ind in
-      if c == 0 then Int.compare a.proj_arg b.proj_arg
-      else c
+    let equal = CanOrd.equal
+    let compare = CanOrd.compare
 
     module Self_Hashcons = struct
       type nonrec t = t
       type u = (inductive -> inductive) * (Id.t -> Id.t)
       let hashcons (hind,hid) p =
         { proj_ind = hind p.proj_ind;
+          proj_relevant = p.proj_relevant;
           proj_npars = p.proj_npars;
           proj_arg = p.proj_arg;
           proj_name = hid p.proj_name }
       let eq p p' =
-        p == p' || (p.proj_ind == p'.proj_ind && p.proj_npars == p'.proj_npars && p.proj_arg == p'.proj_arg && p.proj_name == p'.proj_name)
+        p == p' || (p.proj_relevant == p'.proj_relevant && p.proj_ind == p'.proj_ind && p.proj_npars == p'.proj_npars && p.proj_arg == p'.proj_arg && p.proj_name == p'.proj_name)
       let hash = hash
     end
     module HashRepr = Hashcons.Make(Self_Hashcons)
     let hcons = Hashcons.simple_hcons HashRepr.generate HashRepr.hcons (hcons_ind,Id.hcons)
 
     let map_npars f p =
-      let ind = fst p.proj_ind in
       let npars = p.proj_npars in
-      let ind', npars' = f ind npars in
-      if ind == ind' && npars == npars' then p
-      else {p with proj_ind = (ind',snd p.proj_ind); proj_npars = npars'}
+      let npars' = f npars in
+      if npars == npars' then p
+      else {p with proj_npars = npars'}
 
-    let map f p = map_npars (fun mind n -> f mind, n) p
+    let map f p =
+      let ind = fst p.proj_ind in
+      let ind' = f ind in
+      if ind == ind' then p
+      else {p with proj_ind = (ind',snd p.proj_ind)}
 
     let to_string p = Constant.to_string (constant p)
     let print p = Constant.print (constant p)
