@@ -49,9 +49,9 @@
      E[Delta,Gamma] |-_G
      ------------------------
      E,c:T[Delta,Gamma] |-_G'
-
    add_mind(Ind(Ind[Gamma_p](Gamma_I:=Gamma_C))):
 
+
      E[Delta,Gamma] |-_G
      ------------------------
      E,Ind[Gamma_p](Gamma_I:=Gamma_C)[Delta,Gamma] |-_G'
@@ -130,6 +130,7 @@ type section_data = {
   rev_univ : Univ.ContextSet.t;
   rev_objlabels : Label.Set.t;
   rev_reimport : reimport list;
+  rev_revstruct : structure_body;
 }
 
 module HandleMap = Opaqueproof.HandleMap
@@ -505,23 +506,23 @@ let safe_push_named d env =
 
 let push_named_def (id,de) senv =
   let sections = get_section senv.sections in
-  let sections = Section.push_local sections in
   let c, r, typ = Term_typing.translate_local_def senv.env id de in
-  let x = Context.make_annot id r in
-  let env'' = safe_push_named (LocalDef (x, c, typ)) senv.env in
+  let d = LocalDef (Context.make_annot id r, c, typ) in
+  let env'' = safe_push_named d senv.env in
+  let sections = Section.push_local d sections in
   { senv with sections=Some sections; env = env'' }
 
 let push_named_assum (x,t) senv =
   let sections = get_section senv.sections in
-  let sections = Section.push_local sections in
   let t, r = Term_typing.translate_local_assum senv.env t in
-  let x = Context.make_annot x r in
-  let env'' = safe_push_named (LocalAssum (x,t)) senv.env in
+  let d = LocalAssum (Context.make_annot x r, t) in
+  let sections = Section.push_local d sections in
+  let env'' = safe_push_named d senv.env in
   { senv with sections=Some sections; env = env'' }
 
 let push_section_context uctx senv =
   let sections = get_section senv.sections in
-  let sections = Section.push_context uctx sections in
+  let sections = Section.push_local_universe_context uctx sections in
   let senv = { senv with sections=Some sections } in
   let ctx = Univ.ContextSet.of_context uctx in
   (* We check that the universes are fresh. FIXME: This should be done
@@ -583,10 +584,10 @@ let add_field ((l,sfb) as field) gn senv =
       match sfb, gn with
       | SFBconst cb, C con ->
         let poly = Declareops.constant_is_polymorphic cb in
-        Some (Section.push_constant ~poly con sections)
+        Some Section.(push_global ~poly env' (SecDefinition con) sections)
       | SFBmind mib, I mind ->
         let poly = Declareops.inductive_is_polymorphic mib in
-        Some (Section.push_inductive ~poly mind sections)
+        Some Section.(push_global ~poly env' (SecInductive mind) sections)
       | _, (M | MT) -> Some sections
       | _ -> assert false
   in
@@ -617,7 +618,7 @@ type exported_private_constant = Constant.t * exported_opaque option
 let repr_exported_opaque o =
   let priv = match o .exp_univs with
   | None -> Opaqueproof.PrivateMonomorphic ()
-  | Some n -> Opaqueproof.PrivatePolymorphic (n, Univ.ContextSet.empty)
+  | Some _ -> Opaqueproof.PrivatePolymorphic Univ.ContextSet.empty
   in
   (o.exp_handle, (o.exp_body, priv))
 
@@ -768,12 +769,12 @@ let export_eff eff =
 
 let is_empty_private = function
 | Opaqueproof.PrivateMonomorphic ctx -> Univ.ContextSet.is_empty ctx
-| Opaqueproof.PrivatePolymorphic (_, ctx) -> Univ.ContextSet.is_empty ctx
+| Opaqueproof.PrivatePolymorphic ctx -> Univ.ContextSet.is_empty ctx
 
 (* Special function to call when the body of an opaque definition is provided.
   It performs the type-checking of the body immediately. *)
-let translate_direct_opaque env kn ce =
-  let cb, ctx = Term_typing.translate_opaque env kn ce in
+let translate_direct_opaque ~sec_univs env kn ce =
+  let cb, ctx = Term_typing.translate_opaque ~sec_univs env kn ce in
   let body = ce.Entries.opaque_entry_body, Univ.ContextSet.empty in
   let handle _env c () = (c, Univ.ContextSet.empty, 0) in
   let (c, u) = Term_typing.check_delayed handle ctx (body, ()) in
@@ -782,23 +783,24 @@ let translate_direct_opaque env kn ce =
   { cb with const_body = OpaqueDef c }
 
 let export_side_effects senv eff =
+  let sec_univs = Option.map Section.all_poly_univs senv.sections in
   let env = senv.env in
-      let not_exists e = not (Environ.mem_constant e.seff_constant env) in
-      let aux (acc,sl) e =
-        if not (not_exists e) then acc, sl
-        else e :: acc, e.seff_certif :: sl in
-      let seff, signatures = List.fold_left aux ([],[]) (SideEffects.repr eff) in
-      let trusted = check_signatures senv signatures in
-      let push_seff env eff =
-        let { seff_constant = kn; seff_body = cb ; _ } = eff in
-        let env = Environ.add_constant kn (lift_constant cb) env in
-        env
-      in
-    match trusted with
-    | Some univs ->
-      univs, List.map export_eff seff
-    | None ->
-      let rec recheck_seff seff univs acc env = match seff with
+  let not_exists e = not (Environ.mem_constant e.seff_constant env) in
+  let aux (acc,sl) e =
+    if not (not_exists e) then acc, sl
+    else e :: acc, e.seff_certif :: sl in
+  let seff, signatures = List.fold_left aux ([],[]) (SideEffects.repr eff) in
+  let trusted = check_signatures senv signatures in
+  let push_seff env eff =
+    let { seff_constant = kn; seff_body = cb ; _ } = eff in
+    let env = Environ.add_constant kn (lift_constant cb) env in
+    env
+  in
+  match trusted with
+  | Some univs ->
+    univs, List.map export_eff seff
+  | None ->
+    let rec recheck_seff seff univs acc env = match seff with
       | [] -> univs, List.rev acc
       | eff :: rest ->
         let uctx = eff.seff_univs in
@@ -809,17 +811,17 @@ let export_side_effects senv eff =
           let ce = constant_entry_of_side_effect eff in
           let open Entries in
           let cb = match ce with
-          | DefinitionEff ce ->
-            Term_typing.translate_constant env kn (DefinitionEntry ce)
-          | OpaqueEff ce ->
-            translate_direct_opaque env kn ce
+            | DefinitionEff ce ->
+              Term_typing.translate_constant ~sec_univs env kn (DefinitionEntry ce)
+            | OpaqueEff ce ->
+              translate_direct_opaque ~sec_univs env kn ce
           in
-            let eff = { eff with seff_body = cb } in
-            (push_seff env eff, export_eff eff)
+          let eff = { eff with seff_body = cb } in
+          (push_seff env eff, export_eff eff)
         in
         recheck_seff rest univs (cb :: acc) env
-      in
-      recheck_seff seff Univ.ContextSet.empty [] env
+    in
+    recheck_seff seff Univ.ContextSet.empty [] env
 
 let push_opaque_proof senv =
   let o, otab = Opaqueproof.create (library_dp_of_senv senv) (Environ.opaque_tables senv.env) in
@@ -854,10 +856,11 @@ let export_private_constants eff senv =
 let add_constant l decl senv =
   let kn = Constant.make2 senv.modpath l in
   let senv, cb =
+    let sec_univs = Option.map Section.all_poly_univs senv.sections in
       match decl with
       | OpaqueEntry ce ->
         let senv, o = push_opaque_proof senv in
-        let cb, ctx = Term_typing.translate_opaque senv.env kn ce in
+        let cb, ctx = Term_typing.translate_opaque ~sec_univs senv.env kn ce in
         (* Push the delayed data in the environment *)
         let (_, _, _, i) = Opaqueproof.repr o in
         let nonce = Nonce.create () in
@@ -865,7 +868,7 @@ let add_constant l decl senv =
         let senv = { senv with future_cst } in
         senv, { cb with const_body = OpaqueDef o }
       | ConstantEntry ce ->
-        senv, Term_typing.translate_constant senv.env kn ce
+        senv, Term_typing.translate_constant ~sec_univs senv.env kn ce
   in
   let senv = add_constant_aux senv (kn, cb) in
 
@@ -908,8 +911,8 @@ let check_opaque senv (i : Opaqueproof.opaque_handle) pf =
   let ctx = match ctx with
   | Opaqueproof.PrivateMonomorphic u ->
     Opaqueproof.PrivateMonomorphic (Univ.hcons_universe_context_set u)
-  | Opaqueproof.PrivatePolymorphic (n, u) ->
-    Opaqueproof.PrivatePolymorphic (n, Univ.hcons_universe_context_set u)
+  | Opaqueproof.PrivatePolymorphic u ->
+    Opaqueproof.PrivatePolymorphic (Univ.hcons_universe_context_set u)
   in
   { opq_body = c; opq_univs = ctx; opq_handle = i; opq_nonce = nonce }
 
@@ -946,13 +949,14 @@ let add_private_constant l uctx decl senv : (Constant.t * private_constants) * s
   let kn = Constant.make2 senv.modpath l in
   let senv = push_context_set ~strict:true uctx senv in
     let cb =
+      let sec_univs = Option.map Section.all_poly_univs senv.sections in
       match decl with
       | OpaqueEff ce ->
         let () = assert (check_constraints uctx ce.Entries.opaque_entry_universes) in
-        translate_direct_opaque senv.env kn ce
+        translate_direct_opaque ~sec_univs senv.env kn ce
       | DefinitionEff ce ->
         let () = assert (check_constraints uctx ce.Entries.const_entry_universes) in
-        Term_typing.translate_constant senv.env kn (Entries.DefinitionEntry ce)
+        Term_typing.translate_constant ~sec_univs senv.env kn (Entries.DefinitionEntry ce)
     in
   let dcb = match cb.const_body with
   | Def _ as const_body -> { cb with const_body }
@@ -996,8 +1000,7 @@ let add_checked_mind kn mib senv =
 let add_mind l mie senv =
   let () = check_mind mie l in
   let kn = MutInd.make2 senv.modpath l in
-  let sec_univs = Option.map Section.all_poly_univs  senv.sections
-  in
+  let sec_univs = Option.map Section.all_poly_univs senv.sections in
   let mib = Indtypes.check_inductive senv.env ~sec_univs kn mie in
   (* We still have to add the template monomorphic constraints, and only those
      ones. In all other cases, they are already part of the environment at this
@@ -1013,10 +1016,13 @@ let add_mind ?typing_flags l mie senv =
 
 (** Insertion of module types *)
 
+let check_state senv =
+  (Environ.universes senv.env, Reduction.checked_universes)
+
 let add_modtype l params_mte inl senv =
   let mp = MPdot(senv.modpath, l) in
-  let mtb, cst = Mod_typing.translate_modtype senv.env mp inl params_mte  in
-  let senv = push_context_set ~strict:true (Univ.Level.Set.empty,cst) senv in
+  let state = check_state senv in
+  let mtb, _ = Mod_typing.translate_modtype state senv.env mp inl params_mte  in
   let mtb = Declareops.hcons_module_type mtb in
   let senv = add_field (l,SFBmodtype mtb) MT senv in
   mp, senv
@@ -1035,8 +1041,8 @@ let full_add_module_type mp mt senv =
 
 let add_module l me inl senv =
   let mp = MPdot(senv.modpath, l) in
-  let mb, cst = Mod_typing.translate_module senv.env mp inl me in
-  let senv = push_context_set ~strict:true (Univ.Level.Set.empty,cst) senv in
+  let state = check_state senv in
+  let mb, _ = Mod_typing.translate_module state senv.env mp inl me in
   let mb = Declareops.hcons_module_body mb in
   let senv = add_field (l,SFBmodule mb) M senv in
   let senv =
@@ -1083,8 +1089,8 @@ let start_modtype l senv =
 let add_module_parameter mbid mte inl senv =
   let () = check_empty_struct senv in
   let mp = MPbound mbid in
-  let mtb, cst = Mod_typing.translate_modtype senv.env mp inl ([],mte) in
-  let senv = push_context_set ~strict:true (Univ.Level.Set.empty,cst) senv in
+  let state = check_state senv in
+  let mtb, _ = Mod_typing.translate_modtype state senv.env mp inl ([],mte) in
   let senv = full_add_module_type mp mtb senv in
   let new_variant = match senv.modvariant with
     | STRUCT (params,oldenv) -> STRUCT ((mbid,mtb) :: params, oldenv)
@@ -1127,19 +1133,20 @@ let propagate_loads senv =
 let functorize_module params mb =
   let f x = functorize params x in
   { mb with
-    mod_expr = Modops.implem_smartmap f f mb.mod_expr;
+    mod_expr = Modops.implem_smart_map f f mb.mod_expr;
     mod_type = f mb.mod_type;
     mod_type_alg = Option.map f mb.mod_type_alg }
 
 let build_module_body params restype senv =
   let struc = NoFunctor (List.rev senv.revstruct) in
   let restype' = Option.map (fun (ty,inl) -> (([],ty),inl)) restype in
-  let mb, cst =
-    Mod_typing.finalize_module senv.env senv.modpath
-      (struc,None,senv.modresolver,Univ.Constraints.empty) restype'
+  let state = check_state senv in
+  let mb, _ =
+    Mod_typing.finalize_module state senv.env senv.modpath
+      (struc,None,senv.modresolver) restype'
   in
   let mb' = functorize_module params mb in
-  { mb' with mod_retroknowledge = ModBodyRK senv.local_retroknowledge }, cst
+  { mb' with mod_retroknowledge = ModBodyRK senv.local_retroknowledge }
 
 (** Returning back to the old pre-interactive-module environment,
     with one extra component and some updated fields
@@ -1172,8 +1179,7 @@ let end_module l restype senv =
   let () = check_current_label l mp in
   let () = check_empty_context senv in
   let mbids = List.rev_map fst params in
-  let mb, cst = build_module_body params restype senv in
-  let senv = push_context_set ~strict:true (Univ.Level.Set.empty,cst) senv in
+  let mb = build_module_body params restype senv in
   let newenv = Environ.set_opaque_tables oldsenv.env (Environ.opaque_tables senv.env) in
   let newenv = Environ.set_universes (Environ.universes senv.env) newenv in
   let senv' = propagate_loads { senv with env = newenv } in
@@ -1214,34 +1220,28 @@ let end_modtype l senv =
 let add_include me is_module inl senv =
   let open Mod_typing in
   let mp_sup = senv.modpath in
-  let sign,(),resolver,cst =
-    translate_mse_incl is_module senv.env mp_sup inl me
+  let state = check_state senv in
+  let sign,(),resolver, _ =
+    translate_mse_include is_module state senv.env mp_sup inl me
   in
-  let senv = push_context_set ~strict:true (Univ.Level.Set.empty,cst) senv in
   (* Include Self support  *)
-  let rec compute_sign sign mb resolver senv =
+  let struc = NoFunctor (List.rev senv.revstruct) in
+  let mb = build_mtb mp_sup struc senv.modresolver in
+  let rec compute_sign sign resolver =
     match sign with
     | MoreFunctor(mbid,mtb,str) ->
-      let cst_sub = Subtyping.check_subtypes senv.env mb mtb in
-      let senv =
-        add_constraints
-          (Univ.ContextSet.add_constraints cst_sub Univ.ContextSet.empty)
-          senv in
+      let state = check_state senv in
+      let (_ : UGraph.t) = Subtyping.check_subtypes state senv.env mb mtb in
       let mpsup_delta =
-        Modops.inline_delta_resolver senv.env inl mp_sup mbid mtb mb.mod_delta
+        Modops.inline_delta_resolver senv.env inl mp_sup mbid mtb senv.modresolver
       in
       let subst = Mod_subst.map_mbid mbid mp_sup mpsup_delta in
       let resolver = Mod_subst.subst_codom_delta_resolver subst resolver in
-      compute_sign (Modops.subst_signature subst str) mb resolver senv
-    | NoFunctor str -> resolver,str,senv
-  in
-  let resolver,str,senv =
-    let struc = NoFunctor (List.rev senv.revstruct) in
-    let mtb = build_mtb mp_sup struc senv.modresolver in
-    compute_sign sign mtb resolver senv
-  in
-  let senv = update_resolver (Mod_subst.add_delta_resolver resolver) senv
+      compute_sign (Modops.subst_signature subst str) resolver
+    | NoFunctor str -> resolver, str
   in
+  let resolver, str = compute_sign sign resolver in
+  let senv = update_resolver (Mod_subst.add_delta_resolver resolver) senv in
   let add senv ((l,elem) as field) =
     let new_name = match elem with
       | SFBconst _ ->
@@ -1344,6 +1344,7 @@ let open_section senv =
     rev_univ = senv.univ;
     rev_objlabels = senv.objlabels;
     rev_reimport = [];
+    rev_revstruct = senv.revstruct;
   } in
   let sections = Section.open_section ~custom senv.sections in
   { senv with sections=Some sections }
@@ -1354,34 +1355,15 @@ let close_section senv =
   let env0 = senv.env in
   (* First phase: revert the declarations added in the section *)
   let sections, entries, cstrs, revert = Section.close_section sections0 in
-  let rec pop_revstruct accu entries revstruct = match entries, revstruct with
-  | [], revstruct -> accu, revstruct
-  | _ :: _, [] ->
-    CErrors.anomaly (Pp.str "Unmatched section data")
-  | entry :: entries, (lbl, leaf) :: revstruct ->
-    let data = match entry, leaf with
-    | SecDefinition kn, SFBconst cb ->
-      let () = assert (Label.equal lbl (Constant.label kn)) in
-      `Definition (kn, cb)
-    | SecInductive ind, SFBmind mib ->
-      let () = assert (Label.equal lbl (MutInd.label ind)) in
-      `Inductive (ind, mib)
-    | (SecDefinition _ | SecInductive _), (SFBconst _ | SFBmind _) ->
-      CErrors.anomaly (Pp.str "Section content mismatch")
-    | (SecDefinition _ | SecInductive _), (SFBmodule _ | SFBmodtype _) ->
-      CErrors.anomaly (Pp.str "Module inside a section")
-    in
-    pop_revstruct (data :: accu) entries revstruct
-  in
-  let redo, revstruct = pop_revstruct [] entries senv.revstruct in
-  (* Don't revert the delayed constraints. If some delayed constraints were
-     forced inside the section, they have been turned into global monomorphic
+  (* Don't revert the delayed constraints (future_cst). If some delayed constraints
+     were forced inside the section, they have been turned into global monomorphic
      that are going to be replayed. Those that are not forced are not readded
      by {!add_constant_aux}. *)
-  let { rev_env = env; rev_univ = univ; rev_objlabels = objlabels; rev_reimport } = revert in
+  let { rev_env = env; rev_univ = univ; rev_objlabels = objlabels;
+        rev_reimport; rev_revstruct = revstruct } = revert in
   (* Do not revert the opaque table, the discharged opaque constants are
      referring to it. *)
-  let env = Environ.set_opaque_tables env (Environ.opaque_tables senv.env) in
+  let env = Environ.set_opaque_tables env (Environ.opaque_tables env0) in
   let senv = { senv with env; revstruct; sections; univ; objlabels; } in
   (* Second phase: replay Requires *)
   let senv = List.fold_left (fun senv (lib,cst,vodigest) -> snd (import lib cst vodigest senv))
@@ -1389,21 +1371,21 @@ let close_section senv =
   in
   (* Third phase: replay the discharged section contents *)
   let senv = push_context_set ~strict:true cstrs senv in
-  let modlist = Section.replacement_context env0 sections0 in
-  let cooking_info abstract = { Declarations.modlist; abstract; } in
-  let fold senv = function
-  | `Definition (kn, cb) ->
-    let info = cooking_info (Section.segment_of_constant env0 kn sections0) in
-    let r = { Cooking.from = cb; info } in
-    let cb = Term_typing.translate_recipe senv.env kn r in
+  let fold entry senv =
+    match entry with
+  | SecDefinition kn ->
+    let cb = Environ.lookup_constant kn env0 in
+    let info = Section.segment_of_constant kn sections0 in
+    let cb = Discharge.cook_constant senv.env info cb in
     (* Delayed constants are already in the global environment *)
     add_constant_aux senv (kn, cb)
-  | `Inductive (ind, mib) ->
-    let info = cooking_info (Section.segment_of_inductive env0 ind sections0) in
-    let mib = Cooking.cook_inductive info mib in
+  | SecInductive ind ->
+    let mib = Environ.lookup_mind ind env0 in
+    let info = Section.segment_of_inductive ind sections0 in
+    let mib = Discharge.cook_inductive info mib in
     add_checked_mind ind mib senv
   in
-  List.fold_left fold senv redo
+  List.fold_right fold entries senv
 
 (** {6 Safe typing } *)
 
