@@ -279,11 +279,12 @@ let app_global_with_holes f args n =
     let env = Proofview.Goal.env gl in
     let concl = Proofview.Goal.concl gl in
     Refine.refine ~typecheck:false begin fun sigma ->
-      let t = Tacmach.pf_get_type_of gl fc in
-      let t = Termops.prod_applist sigma t (Array.to_list args) in
-      let ans = mkApp (fc, args) in
-      let (sigma, holes) = gen_holes env sigma t n [] in
-      let ans = applist (ans, holes) in
+      let open Environ in
+      let fj = Retyping.get_judgment_of env sigma fc in
+      let argsj = Array.map (fun c -> Retyping.get_judgment_of env sigma c) args in
+      let sigma, fj = Typing.judge_of_apply env sigma fj argsj in
+      let (sigma, holes) = gen_holes env sigma fj.uj_type n [] in
+      let ans = applist (fj.uj_val, holes) in
       let sigma = Typing.check env sigma ans concl in
       (sigma, ans)
     end
@@ -451,7 +452,7 @@ let cc_tactic depth additional_terms b =
     let _ = debug_congruence (fun () -> Pp.str "Computation completed.") in
     let uf=forest state in
     match sol with
-      None -> Tacticals.tclFAIL 0 (str (if b then "simple congruence failed" else "congruence failed"))
+      None -> Tacticals.tclFAIL (str (if b then "simple congruence failed" else "congruence failed"))
     | Some reason ->
       debug_congruence (fun () -> Pp.str "Goal solved, generating proof ...");
       match reason with
@@ -482,7 +483,7 @@ let cc_tactic depth additional_terms b =
                         end ++
                       fnl() ++ str "  replacing metavariables by arbitrary terms")
         in
-        Tacticals.tclFAIL 0 msg
+        Tacticals.tclFAIL msg
       | Contradiction dis ->
         let env = Proofview.Goal.env gl in
         let p=build_proof env sigma uf (`Prove (dis.lhs,dis.rhs)) in
