@@ -64,6 +64,8 @@ let change_vars =
           GVar new_id | GEvar _ as x -> x | GPatVar _ as x -> x
         | GApp (rt', rtl) ->
           GApp (change_vars mapping rt', List.map (change_vars mapping) rtl)
+        | GProj (f, rtl, rt) ->
+          GProj (f, List.map (change_vars mapping) rtl, change_vars mapping rt)
         | GLambda (name, k, t, b) ->
           GLambda
             ( name
@@ -106,10 +108,8 @@ let change_vars =
         | GRec _ -> user_err ?loc Pp.(str "Local (co)fixes are not supported")
         | GSort _ as x -> x | GHole _ as x -> x | GInt _ as x -> x
         | GFloat _ as x -> x
-        | GCast (b, c) ->
-          GCast
-            ( change_vars mapping b
-            , Glob_ops.map_cast_type (change_vars mapping) c )
+        | GCast (b, k, c) ->
+          GCast (change_vars mapping b, k, change_vars mapping c)
         | GArray (u, t, def, ty) ->
           GArray
             ( u
@@ -284,10 +284,12 @@ let rec alpha_rt excluded rt =
         , alpha_rt excluded rhs )
     | GRec _ -> user_err Pp.(str "Not handled GRec")
     | (GSort _ | GInt _ | GFloat _ | GHole _) as rt -> rt
-    | GCast (b, c) ->
-      GCast (alpha_rt excluded b, Glob_ops.map_cast_type (alpha_rt excluded) c)
+    | GCast (b, k, c) ->
+      GCast (alpha_rt excluded b, k, alpha_rt excluded c)
     | GApp (f, args) ->
       GApp (alpha_rt excluded f, List.map (alpha_rt excluded) args)
+    | GProj (f, args, c) ->
+      GProj (f, List.map (alpha_rt excluded) args, alpha_rt excluded c)
     | GArray (u, t, def, ty) ->
       GArray
         ( u
@@ -313,7 +315,7 @@ let is_free_in id =
     DAst.with_loc_val
       (fun ?loc -> function GRef _ -> false | GVar id' -> Id.compare id' id == 0
         | GEvar _ -> false | GPatVar _ -> false
-        | GApp (rt, rtl) -> List.exists is_free_in (rt :: rtl)
+        | GApp (rt, rtl) | GProj (_, rtl, rt) -> List.exists is_free_in (rt :: rtl)
         | GLambda (n, _, t, b) | GProd (n, _, t, b) ->
           let check_in_b =
             match n with Name id' -> not (Id.equal id' id) | _ -> true
@@ -341,7 +343,7 @@ let is_free_in id =
           is_free_in cond || is_free_in br1 || is_free_in br2
         | GRec _ -> user_err Pp.(str "Not handled GRec") | GSort _ -> false
         | GHole _ -> false
-        | GCast (b, (CastConv t | CastVM t | CastNative t)) ->
+        | GCast (b, _, t) ->
           is_free_in b || is_free_in t
         | GInt _ | GFloat _ -> false
         | GArray (_u, t, def, ty) ->
@@ -378,6 +380,8 @@ let replace_var_by_term x_id term =
         | (GRef _ | GVar _ | GEvar _ | GPatVar _) as rt -> rt
         | GApp (rt', rtl) ->
           GApp (replace_var_by_pattern rt', List.map replace_var_by_pattern rtl)
+        | GProj (f, rtl, rt) ->
+          GProj (f, List.map replace_var_by_pattern rtl, replace_var_by_pattern rt)
         | GLambda (Name id, _, _, _) as rt when Id.compare id x_id == 0 -> rt
         | GLambda (name, k, t, b) ->
           GLambda (name, k, replace_var_by_pattern t, replace_var_by_pattern b)
@@ -424,10 +428,8 @@ let replace_var_by_term x_id term =
             , Array.map replace_var_by_pattern t
             , replace_var_by_pattern def
             , replace_var_by_pattern ty )
-        | GCast (b, c) ->
-          GCast
-            ( replace_var_by_pattern b
-            , Glob_ops.map_cast_type replace_var_by_pattern c ))
+        | GCast (b, k, c) ->
+          GCast (replace_var_by_pattern b, k, replace_var_by_pattern c))
       x
   and replace_var_by_pattern_br ({CAst.loc; v = idl, patl, res} as br) =
     if List.exists (fun id -> Id.compare id x_id == 0) idl then br
@@ -506,6 +508,7 @@ let expand_as =
       | GVar id as rt -> (
         try DAst.get (Id.Map.find id map) with Not_found -> rt )
       | GApp (f, args) -> GApp (expand_as map f, List.map (expand_as map) args)
+      | GProj (f, args, c) -> GProj (f, List.map (expand_as map) args, expand_as map c)
       | GLambda (na, k, t, b) ->
         GLambda (na, k, expand_as map t, expand_as map b)
       | GProd (na, k, t, b) -> GProd (na, k, expand_as map t, expand_as map b)
@@ -525,8 +528,8 @@ let expand_as =
           , expand_as map br1
           , expand_as map br2 )
       | GRec _ -> user_err Pp.(str "Not handled GRec")
-      | GCast (b, c) ->
-        GCast (expand_as map b, Glob_ops.map_cast_type (expand_as map) c)
+      | GCast (b, k, c) ->
+        GCast (expand_as map b, k, expand_as map c)
       | GCases (sty, po, el, brl) ->
         GCases
           ( sty
