@@ -77,12 +77,12 @@ let def_of_const t =
         ++ str "." ) )
   | _ -> assert false
 
-let type_of_const sigma t =
+let type_of_const env sigma t =
   match EConstr.kind sigma t with
   | Const (sp, u) ->
     let u = EInstance.kind sigma u in
     (* FIXME discarding universe constraints *)
-    Typeops.type_of_constant_in (Global.env ()) (sp, u)
+    Typeops.type_of_constant_in env (sp, u)
   | _ -> assert false
 
 let constant sl s = UnivGen.constr_of_monomorphic_global (Global.env ()) (find_reference sl s)
@@ -570,7 +570,7 @@ let rec prove_lt hyple =
           ; New.observe_tac
               (fun _ _ ->
                 str "assumption: "
-                ++ Printer.pr_goal Evd.{it = Proofview.Goal.goal g; sigma})
+                ++ Printer.Debug.pr_goal g)
               assumption ])
 
 let rec destruct_bounds_aux infos (bound, hyple, rechyps) lbounds =
@@ -946,7 +946,7 @@ let rec prove_le () =
                 ; New.observe_tac
                     (fun _ _ -> str "prove_le (rec)")
                     (prove_le ()) ]
-            with Not_found -> Tacticals.tclFAIL 0 (mt ())
+            with Not_found -> Tacticals.tclFAIL (mt ())
           end ])
 
 let rec make_rewrite_list expr_info max = function
@@ -1487,7 +1487,7 @@ let open_new_goal ~lemma build_proof sigma using_lemmas ref_ goal_name
                      (eapply_with_bindings
                         (mkVar (List.nth !lid !h_num), NoBindings))
                      e_assumption
-                 ; Eauto.eauto_with_bases (true, 5)
+                 ; Eauto.eauto_with_bases ~depth:5
                      [(fun _ sigma -> (sigma, Lazy.force refl_equal))]
                      [Hints.Hint_db.empty TransparentState.empty false] ]))
     in
@@ -1572,7 +1572,7 @@ let start_equation (f : GlobRef.t) (term_f : GlobRef.t)
       let terminate_constr = constr_of_monomorphic_global (Global.env ()) term_f in
       let terminate_constr = EConstr.of_constr terminate_constr in
       let nargs =
-        nb_prod sigma (EConstr.of_constr (type_of_const sigma terminate_constr))
+        nb_prod sigma (EConstr.of_constr (type_of_const (Global.env ()) sigma terminate_constr))
       in
       let x = n_x_id ids nargs in
       New.observe_tac
