@@ -621,6 +621,7 @@ let interp_gen kind ist pattern_mode flags env sigma c =
     Exninfo.iraise reraise
 
 let constr_flags () = {
+  use_coercions = true;
   use_typeclasses = UseTC;
   solve_unification_constraints = true;
   fail_evar = true;
@@ -639,6 +640,7 @@ let interp_constr = interp_constr_gen WithoutTypeConstraint
 let interp_type = interp_constr_gen IsType
 
 let open_constr_use_classes_flags () = {
+  use_coercions = true;
   use_typeclasses = UseTC;
   solve_unification_constraints = true;
   fail_evar = false;
@@ -648,6 +650,7 @@ let open_constr_use_classes_flags () = {
 }
 
 let open_constr_no_classes_flags () = {
+  use_coercions = true;
   use_typeclasses = NoUseTC;
   solve_unification_constraints = true;
   fail_evar = false;
@@ -657,6 +660,7 @@ let open_constr_no_classes_flags () = {
 }
 
 let pure_open_constr_flags = {
+  use_coercions = true;
   use_typeclasses = NoUseTC;
   solve_unification_constraints = true;
   fail_evar = false;
@@ -680,7 +684,7 @@ let interp_typed_pattern ist env sigma (_,c,_) =
     interp_gen WithoutTypeConstraint ist true pure_open_constr_flags env sigma c in
   (* FIXME: it is necessary to be unsafe here because of the way we handle
      evars in the pretyper. Sometimes they get solved eagerly. *)
-  pattern_of_constr env sigma (EConstr.Unsafe.to_constr c)
+  legacy_bad_pattern_of_constr env sigma c
 
 (* Interprets a constr expression *)
 let interp_constr_in_compound_list inj_fun dest_fun interp_fun ist env sigma l =
@@ -725,7 +729,7 @@ let interp_closed_typed_pattern_with_occurrences ist env sigma (occs, a) =
       try Inl (coerce_to_evaluable_ref env sigma x)
       with CannotCoerceTo _ ->
         let c = coerce_to_closed_constr env x in
-        Inr (pattern_of_constr env sigma (EConstr.Unsafe.to_constr c)) in
+        Inr (legacy_bad_pattern_of_constr env sigma c) in
     (try try_interp_ltac_var coerce_eval_ref_or_constr ist (Some (env,sigma)) (CAst.make ?loc id)
      with Not_found as exn ->
        let _, info = Exninfo.capture exn in
@@ -774,17 +778,14 @@ let interp_may_eval f ist env sigma = function
       let (redfun, _) = Redexpr.reduction_of_red_expr env redexp in
       redfun env sigma c_interp
   | ConstrContext ({loc;v=s},c) ->
-      (try
-        let (sigma,ic) = f ist env sigma c in
-        let ctxt = try_interp_ltac_var coerce_to_constr_context ist (Some (env, sigma)) (CAst.make ?loc s) in
-        let ctxt = EConstr.Unsafe.to_constr ctxt in
-        let ic = EConstr.Unsafe.to_constr ic in
-        let c = subst_meta [Constr_matching.special_meta,ic] ctxt in
-        Typing.solve_evars env sigma (EConstr.of_constr c)
-      with
-        | Not_found ->
-            user_err ?loc
-            (str "Unbound context identifier" ++ Id.print s ++ str"."))
+    let (sigma,ic) = f ist env sigma c in
+    let ctxt =
+      try try_interp_ltac_var coerce_to_constr_context ist (Some (env, sigma)) (CAst.make ?loc s)
+      with Not_found ->
+        user_err ?loc (str "Unbound context identifier" ++ Id.print s ++ str".")
+    in
+    let c = Constr_matching.instantiate_context ctxt ic in
+    Typing.solve_evars env sigma c
   | ConstrTypeOf c ->
       let (sigma,c_interp) = f ist env sigma c in
       let (sigma, t) = Typing.type_of ~refresh:true env sigma c_interp in
@@ -1038,14 +1039,8 @@ let head_with_value (lvar,lval) =
     {!Matching.matching_result}) into a context value of Ltac.  *)
 let interp_context ctxt = in_gen (topwit wit_constr_context) ctxt
 
-(* Reads a pattern by substituting vars of lfun *)
-let use_types = false
-
-let eval_pattern lfun ist env sigma (bvars,(glob,_),pat as c) =
-  if use_types then
-    (bvars,interp_typed_pattern ist env sigma c)
-  else
-    (bvars,instantiate_pattern env sigma lfun pat)
+let eval_pattern lfun ist env sigma (bvars, _, pat) =
+  (bvars,Constr_matching.instantiate_pattern env sigma lfun pat)
 
 let read_pattern lfun ist env sigma = function
   | Subterm (ido,c) -> Subterm (ido,eval_pattern lfun ist env sigma c)
@@ -1151,7 +1146,7 @@ and eval_tactic_ist ist tac : unit Proofview.tactic =
       end
   | TacFail (g,n,s) ->
       let msg = interp_message ist s in
-      let tac ~info l = Tacticals.tclFAIL ~info (interp_int_or_var ist n) l in
+      let tac ~info l = Tacticals.tclFAILn ~info (interp_int_or_var ist n) l in
       let tac =
         match g with
         | TacLocal ->
