@@ -54,26 +54,22 @@ let debug_ho_unification = CDebug.create ~name:"ho-unification" ()
 (*******************************************)
 (* Functions to deal with impossible cases *)
 (*******************************************)
-let impossible_default_case env =
-  let type_of_id = Coqlib.lib_ref "core.IDProp.type" in
-  let c, ctx = UnivGen.fresh_global_instance env (Coqlib.(lib_ref "core.IDProp.idProp")) in
-  let (_, u) = Constr.destRef c in
-  Some (c, Constr.mkRef (type_of_id, u), ctx)
 
-let coq_unit_judge =
-  let open Environ in
-  let make_judge c t = make_judge (EConstr.of_constr c) (EConstr.of_constr t) in
+(* In case the constants id/ID are not defined *)
+let unit_judge_fallback =
   let na1 = make_annot (Name (Id.of_string "A")) Sorts.Relevant in
   let na2 = make_annot (Name (Id.of_string "H")) Sorts.Relevant in
-  fun env ->
-    match impossible_default_case env with
-    | Some (id, type_of_id, ctx) ->
-      make_judge id type_of_id, ctx
-    | None ->
-      (* In case the constants id/ID are not defined *)
-      Environ.make_judge (mkLambda (na1,mkProp,mkLambda(na2,mkRel 1,mkRel 1)))
-        (mkProd (na1,mkProp,mkArrow (mkRel 1) Sorts.Relevant (mkRel 2))),
-      Univ.ContextSet.empty
+  make_judge
+    (mkLambda (na1,mkProp,mkLambda(na2,mkRel 1,mkRel 1)))
+    (mkProd (na1,mkProp,mkArrow (mkRel 1) Sorts.Relevant (mkRel 2)))
+
+let coq_unit_judge env sigma =
+  match Coqlib.lib_ref "core.IDProp.idProp" with
+  | c ->
+    let sigma, c = Evd.fresh_global env sigma c in
+    let t = Retyping.get_type_of env sigma c in
+    sigma, make_judge c t
+  | exception _ -> sigma, unit_judge_fallback
 
 let unfold_projection env evd ts p c =
   let cst = Projection.constant p in
@@ -234,7 +230,7 @@ let occur_rigidly flags env evd (evk,_) t =
 let check_conv_record env sigma (t1,sk1) (t2,sk2) =
   let open ValuePattern in
   let (proji, u), arg = Termops.global_app_of_constr sigma t1 in
-  let t2, sk2' = decompose_app_vect sigma (shrink_eta env t2) in
+  let t2, sk2' = decompose_app_vect sigma (shrink_eta sigma t2) in
   let sk2 = Stack.append_app sk2' sk2 in
   let (sigma, solution), sk2_effective =
     let t2 =
@@ -362,6 +358,69 @@ let rec ise_app_rev_stack2 env f evd revsk1 revsk2 =
      end
   | _, _ -> (revsk1,revsk2), Success evd
 
+(* Add equality constraints for covariant/invariant positions. For
+   irrelevant positions, unify universes when flexible. *)
+let compare_cumulative_instances evd variances u u' =
+  match Evarutil.compare_cumulative_instances CONV variances u u' evd with
+  | Inl evd ->
+    Success evd
+  | Inr p -> UnifFailure (evd, UnifUnivInconsistency p)
+
+type application = FullyApplied | NumArgs of int
+
+let is_applied o n = match o with FullyApplied -> true | NumArgs m -> Int.equal m n
+
+let compare_heads env evd ~nargs term term' =
+  let check_strict evd u u' =
+    let cstrs = Univ.enforce_eq_instances u u' Univ.Constraints.empty in
+    try Success (Evd.add_constraints evd cstrs)
+    with UGraph.UniverseInconsistency p -> UnifFailure (evd, UnifUnivInconsistency p)
+  in
+  match EConstr.kind evd term, EConstr.kind evd term' with
+  | Const (c, u), Const (c', u') when QConstant.equal env c c' ->
+    if is_applied nargs 1 && Environ.is_array_type env c
+    then
+      let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
+      compare_cumulative_instances evd [|Univ.Variance.Irrelevant|] u u'
+    else
+      let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
+      check_strict evd u u'
+  | Const _, Const _ -> UnifFailure (evd, NotSameHead)
+  | Ind ((mi,i) as ind , u), Ind (ind', u') when Names.Ind.CanOrd.equal ind ind' ->
+    if EInstance.is_empty u && EInstance.is_empty u' then Success evd
+    else
+      let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
+      let mind = Environ.lookup_mind mi env in
+      let open Declarations in
+      begin match mind.mind_variance with
+        | None -> check_strict evd u u'
+        | Some variances ->
+          let needed = Reduction.inductive_cumulativity_arguments (mind,i) in
+          if not (is_applied nargs needed)
+          then check_strict evd u u'
+          else
+            compare_cumulative_instances evd variances u u'
+      end
+  | Ind _, Ind _ -> UnifFailure (evd, NotSameHead)
+  | Construct (((mi,ind),ctor as cons), u), Construct (cons', u')
+    when Names.Construct.CanOrd.equal cons cons' ->
+    if EInstance.is_empty u && EInstance.is_empty u' then Success evd
+    else
+      let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
+      let mind = Environ.lookup_mind mi env in
+      let open Declarations in
+      begin match mind.mind_variance with
+        | None -> check_strict evd u u'
+        | Some variances ->
+          let needed = Reduction.constructor_cumulativity_arguments (mind,ind,ctor) in
+          if not (is_applied nargs needed)
+          then check_strict evd u u'
+          else
+            Success (compare_constructor_instances evd u u')
+      end
+  | Construct _, Construct _ -> UnifFailure (evd, NotSameHead)
+  | _, _ -> anomaly (Pp.str "")
+
 (* This function tries to unify 2 stacks element by element. It works
    from the end to the beginning. If it unifies a non empty suffix of
    stacks but not the entire stacks, the first part of the answer is
@@ -384,12 +443,18 @@ let rec ise_stack2 no_app env evd f sk1 sk2 =
     match revsk1, revsk2 with
     | [], [] -> None, Success i
     | Stack.Case cse1 :: q1, Stack.Case cse2 :: q2 ->
-      let (t1, c1) = Stack.expand_case env evd cse1 in
-      let (t2, c2) = Stack.expand_case env evd cse2 in
+      let (ci1, u1, pms1, t1, br1) = Stack.expand_case env evd cse1 in
+      let (ci2, u2, pms2, t2, br2) = Stack.expand_case env evd cse2 in
+      let hd1 = mkIndU (ci1.ci_ind, u1) in
+      let hd2 = mkIndU (ci2.ci_ind, u2) in
+      let fctx i (ctx1, t1) (_ctx2, t2) = f (push_rel_context ctx1 env) i CONV t1 t2 in
       begin
         match ise_and i [
-          (fun i -> f env i CONV t1 t2);
-          (fun i -> ise_array2 i (fun ii -> f env ii CONV) c1 c2)]
+          (fun i -> compare_heads env i ~nargs:FullyApplied hd1 hd2);
+          (fun i -> ise_array2 i (fun ii -> f env ii CONV) pms1 pms2);
+          (fun i -> fctx i t1 t2);
+          (fun i -> ise_array2 i fctx br1 br2);
+        ]
         with
         | Success i' -> ise_rev_stack2 true i' q1 q2
         | UnifFailure _ as x -> fail x
@@ -423,12 +488,18 @@ let rec exact_ise_stack2 env evd f sk1 sk2 =
     match revsk1, revsk2 with
     | [], [] -> Success i
     | Stack.Case cse1 :: q1, Stack.Case cse2 :: q2 ->
-      let (t1, c1) = Stack.expand_case env evd cse1 in
-      let (t2, c2) = Stack.expand_case env evd cse2 in
+      let (ci1, u1, pms1, t1, br1) = Stack.expand_case env evd cse1 in
+      let (ci2, u2, pms2, t2, br2) = Stack.expand_case env evd cse2 in
+      let hd1 = mkIndU (ci1.ci_ind, u1) in
+      let hd2 = mkIndU (ci2.ci_ind, u2) in
+      let fctx i (ctx1, t1) (_ctx2, t2) = f (push_rel_context ctx1 env) i CONV t1 t2 in
       ise_and i [
-      (fun i -> ise_rev_stack2 i q1 q2);
-      (fun i -> ise_array2 i (fun ii -> f env ii CONV) c1 c2);
-      (fun i -> f env i CONV t1 t2)]
+        (fun i -> ise_rev_stack2 i q1 q2);
+        (fun i -> compare_heads env i ~nargs:FullyApplied hd1 hd2);
+        (fun i -> ise_array2 i (fun ii -> f env ii CONV) pms1 pms2);
+        (fun i -> fctx i t1 t2);
+        (fun i -> ise_array2 i fctx br1 br2);
+      ]
     | Stack.Fix (((li1, i1),(_,tys1,bds1 as recdef1)),a1)::q1,
       Stack.Fix (((li2, i2),(_,tys2,bds2)),a2)::q2 ->
       if Int.equal i1 i2 && Array.equal Int.equal li1 li2 then
@@ -453,65 +524,8 @@ let rec exact_ise_stack2 env evd f sk1 sk2 =
     ise_rev_stack2 evd (List.rev sk1) (List.rev sk2)
   else UnifFailure (evd, (* Dummy *) NotSameHead)
 
-(* Add equality constraints for covariant/invariant positions. For
-   irrelevant positions, unify universes when flexible. *)
-let compare_cumulative_instances evd variances u u' =
-  match Evarutil.compare_cumulative_instances CONV variances u u' evd with
-  | Inl evd ->
-    Success evd
-  | Inr p -> UnifFailure (evd, UnifUnivInconsistency p)
-
 let compare_heads env evd ~nargs term term' =
-    let check_strict evd u u' =
-      let cstrs = Univ.enforce_eq_instances u u' Univ.Constraint.empty in
-      try Success (Evd.add_constraints evd cstrs)
-      with Univ.UniverseInconsistency p -> UnifFailure (evd, UnifUnivInconsistency p)
-    in
-      match EConstr.kind evd term, EConstr.kind evd term' with
-      | Const (c, u), Const (c', u') when QConstant.equal env c c' ->
-        if Int.equal nargs 1 && Environ.is_array_type env c
-        then
-          let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
-          compare_cumulative_instances evd [|Univ.Variance.Irrelevant|] u u'
-        else
-          let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
-          check_strict evd u u'
-      | Const _, Const _ -> UnifFailure (evd, NotSameHead)
-      | Ind ((mi,i) as ind , u), Ind (ind', u') when Names.Ind.CanOrd.equal ind ind' ->
-        if EInstance.is_empty u && EInstance.is_empty u' then Success evd
-        else
-          let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
-          let mind = Environ.lookup_mind mi env in
-          let open Declarations in
-          begin match mind.mind_variance with
-            | None -> check_strict evd u u'
-            | Some variances ->
-              let needed = Reduction.inductive_cumulativity_arguments (mind,i) in
-              if not (Int.equal nargs needed)
-              then check_strict evd u u'
-              else
-                compare_cumulative_instances evd variances u u'
-          end
-      | Ind _, Ind _ -> UnifFailure (evd, NotSameHead)
-      | Construct (((mi,ind),ctor as cons), u), Construct (cons', u')
-        when Names.Construct.CanOrd.equal cons cons' ->
-        if EInstance.is_empty u && EInstance.is_empty u' then Success evd
-        else
-          let u = EInstance.kind evd u and u' = EInstance.kind evd u' in
-          let mind = Environ.lookup_mind mi env in
-          let open Declarations in
-          begin match mind.mind_variance with
-            | None -> check_strict evd u u'
-            | Some variances ->
-              let needed = Reduction.constructor_cumulativity_arguments (mind,ind,ctor) in
-              if not (Int.equal nargs needed)
-              then check_strict evd u u'
-              else
-                Success (compare_constructor_instances evd u u')
-          end
-      | Construct _, Construct _ -> UnifFailure (evd, NotSameHead)
-      | _, _ -> anomaly (Pp.str "")
-
+  compare_heads env evd ~nargs:(NumArgs nargs) term term'
 
 let conv_fun f flags on_types =
   let typefn env evd pbty term1 term2 =
@@ -527,6 +541,22 @@ let conv_fun f flags on_types =
     | TypeUnification -> typefn
     | TermUnification -> termfn
 
+let infer_conv_noticing_evars ~pb ~ts env sigma t1 t2 =
+  let has_evar = ref false in
+  let conv pb ~l2r sigma = Reduction.generic_conv pb ~l2r (fun ev ->
+      let v = existential_opt_value0 sigma ev in
+      if Option.is_empty v then has_evar := true;
+      v)
+  in
+  match infer_conv_gen conv ~catch_incon:false ~pb ~ts env sigma t1 t2 with
+  | Some sigma -> Some (Success sigma)
+  | None ->
+    if !has_evar then None
+    else Some (UnifFailure (sigma, ConversionFailed (env,t1,t2)))
+  | exception UGraph.UniverseInconsistency e ->
+    if !has_evar then None
+    else Some (UnifFailure (sigma, UnifUnivInconsistency e))
+
 let rec evar_conv_x flags env evd pbty term1 term2 =
   let term1 = whd_head_evar evd term1 in
   let term2 = whd_head_evar evd term2 in
@@ -534,16 +564,8 @@ let rec evar_conv_x flags env evd pbty term1 term2 =
      could have found, we do it only if the terms are free of evar.
      Note: incomplete heuristic... *)
   let ground_test =
-    if is_ground_term evd term1 && is_ground_term evd term2 then (
-      let e =
-          match infer_conv ~catch_incon:false ~pb:pbty ~ts:flags.closed_ts env evd term1 term2 with
-          | Some evd -> Success evd
-          | None -> UnifFailure (evd, ConversionFailed (env,term1,term2))
-          | exception Univ.UniverseInconsistency e -> UnifFailure (evd, UnifUnivInconsistency e)
-      in
-        match e with
-        | UnifFailure (evd, e) when not (is_ground_env evd env) -> None
-        | _ -> Some e)
+    if is_ground_term evd term1 && is_ground_term evd term2 then
+      infer_conv_noticing_evars ~pb:pbty ~ts:flags.closed_ts env evd term1 term2
     else None
   in
   match ground_test with
@@ -639,7 +661,7 @@ and evar_eqappr_x ?(rhs_is_already_stuck = false) flags env evd pbty
     else
       ise_and evd [(fun i ->
           try compare_heads env i ~nargs term term'
-          with Univ.UniverseInconsistency p -> UnifFailure (i, UnifUnivInconsistency p));
+          with UGraph.UniverseInconsistency p -> UnifFailure (i, UnifUnivInconsistency p));
          (fun i -> exact_ise_stack2 env i (evar_conv_x flags) sk sk')]
   in
   let consume l2r (_, skF as apprF) (_,skM as apprM) i =
@@ -935,7 +957,7 @@ and evar_eqappr_x ?(rhs_is_already_stuck = false) flags env evd pbty
               ise_and i [(fun i ->
                 try Success (Evd.add_universe_constraints i univs)
                 with UniversesDiffer -> UnifFailure (i,NotSameHead)
-                | Univ.UniverseInconsistency p -> UnifFailure (i, UnifUnivInconsistency p));
+                | UGraph.UniverseInconsistency p -> UnifFailure (i, UnifUnivInconsistency p));
                          (fun i -> exact_ise_stack2 env i (evar_conv_x flags) sk1 sk2)]
           | None ->
             UnifFailure (i,NotSameHead)
@@ -1047,7 +1069,7 @@ and evar_eqappr_x ?(rhs_is_already_stuck = false) flags env evd pbty
                  then Evd.set_eq_sort env evd s1 s2
                  else Evd.set_leq_sort env evd s1 s2
                in Success evd'
-             with Univ.UniverseInconsistency p ->
+             with UGraph.UniverseInconsistency p ->
                UnifFailure (evd,UnifUnivInconsistency p)
              | e when CErrors.noncritical e -> UnifFailure (evd,NotSameHead))
 
@@ -1676,7 +1698,7 @@ let apply_conversion_problem_heuristic flags env evd with_ho pbty t1 t2 =
                  (evar_define evar_unify flags ~choose:true)
                  evar_unify flags env evd
                  (position_problem true pbty) ev1 ev2)
-      with IllTypedInstance (env,t,u) ->
+      with IllTypedInstance (env,evd,t,u) ->
             UnifFailure (evd,InstanceNotSameType (evk1,env,t,u)))
   | Evar ev1,_ when is_evar_allowed flags (fst ev1) && Array.length l1 <= Array.length l2 ->
       (* On "?n t1 .. tn = u u1 .. u(n+p)", try first-order unification *)
@@ -1756,8 +1778,7 @@ let solve_unconstrained_impossible_cases env evd =
   Evd.fold_undefined (fun evk ev_info evd' ->
     match ev_info.evar_source with
     | loc,Evar_kinds.ImpossibleCase ->
-      let j, ctx = coq_unit_judge env in
-      let evd' = Evd.merge_context_set Evd.univ_flexible_alg ?loc evd' ctx in
+      let evd', j = coq_unit_judge env evd' in
       let ty = j_type j in
       let flags = default_flags env in
       instantiate_evar evar_unify flags env evd' evk ty (* should we protect from raising IllTypedInstance? *)
