@@ -103,14 +103,11 @@ let mis_is_recursive (ind,mib,mip) =
   mis_is_recursive_subset (List.interval 0 (mib.mind_ntypes - 1))
     mip.mind_recargs
 
-let mis_nf_constructor_type ((ind,u),mib,mip) j =
-  let specif = mip.mind_nf_lc
-  and ntypes = mib.mind_ntypes
-  and nconstr = Array.length mip.mind_consnames in
-  let make_Ik k = mkIndU (((fst ind),ntypes-k-1),u) in
+let mis_nf_constructor_type ((_,j),u) (mib,mip) =
+  let nconstr = Array.length mip.mind_consnames in
   if j > nconstr then user_err Pp.(str "Not enough constructors in the type.");
-  let (ctx, cty) = specif.(j - 1) in
-  substl (List.init ntypes make_Ik) (subst_instance_constr u (Term.it_mkProd_or_LetIn cty ctx))
+  let (ctx, cty) = mip.mind_nf_lc.(j - 1) in
+  subst_instance_constr u (Term.it_mkProd_or_LetIn cty ctx)
 
 (* Number of constructors *)
 
@@ -308,7 +305,7 @@ let lift_constructor n cs = {
   cs_cstr = cs.cs_cstr;
   cs_params = List.map (lift n) cs.cs_params;
   cs_nargs = cs.cs_nargs;
-  cs_args = lift_rel_context n cs.cs_args;
+  cs_args = Vars.lift_rel_context n cs.cs_args;
   cs_concl_realargs = Array.map (liftn n (cs.cs_nargs+1)) cs.cs_concl_realargs
 }
 
@@ -316,16 +313,19 @@ let lift_constructor n cs = {
 let instantiate_params t params sign =
   let nnonrecpar = Context.Rel.nhyps sign - List.length params in
   (* Adjust the signature if recursively non-uniform parameters are not here *)
-  let _,sign = context_chop nnonrecpar sign in
+  let _,sign = Context.Rel.chop_nhyps nnonrecpar sign in
   let _,t = decompose_prod_n_assum (Context.Rel.length sign) t in
-  let subst = subst_of_rel_context_instance sign params in
+  let subst = subst_of_rel_context_instance_list sign params in
   substl subst t
 
-let get_constructor ((ind,u as indu),mib,mip,params) j =
-  assert (j <= Array.length mip.mind_consnames);
-  let typi = mis_nf_constructor_type (indu,mib,mip) j in
+let instantiate_constructor_params (_,u as cstru) (mib,_ as mind_specif) params =
+  let typi = mis_nf_constructor_type cstru mind_specif in
   let ctx = Vars.subst_instance_context u mib.mind_params_ctxt in
-  let typi = instantiate_params typi params ctx in
+  instantiate_params typi params ctx
+
+let get_constructor ((ind,u),mib,mip,params) j =
+  assert (j <= Array.length mip.mind_consnames);
+  let typi = instantiate_constructor_params ((ind,j),u) (mib,mip) params in
   let (args,ccl) = decompose_prod_assum typi in
   let (_,allargs) = decompose_app ccl in
   let vargs = List.skipn (List.length params) allargs in
@@ -355,9 +355,9 @@ let make_project env sigma ind pred c branches ps =
     let mib, _ = Inductive.lookup_mind_specif env ind in
     if (* dependent *) not (Vars.noccurn sigma 1 t) &&
          not (has_dependent_elim mib) then
-      user_err ~hdr:"make_case_or_project"
+      user_err
         Pp.(str"Dependent case analysis not allowed" ++
-              str" on inductive type " ++ Termops.Internal.print_constr_env env sigma (mkInd ind))
+              str" on inductive type " ++ Termops.Internal.print_constr_env env sigma (mkInd ind) ++ str ".")
   in
   let branch = branches.(0) in
   let ctx, br = decompose_lam_n_assum sigma (Array.length ps) branch in
@@ -410,16 +410,14 @@ let get_arity env ((ind,u),params) =
        parameters *)
     let nparams = List.length params in
     if Int.equal nparams mib.mind_nparams then
-      mib.mind_params_ctxt
+      Inductive.inductive_paramdecls (mib,u)
     else begin
       assert (Int.equal nparams mib.mind_nparams_rec);
-      let nnonrecparamdecls = mib.mind_nparams - mib.mind_nparams_rec in
-      snd (Termops.context_chop nnonrecparamdecls mib.mind_params_ctxt)
+      snd (Inductive.inductive_nonrec_rec_paramdecls (mib,u))
     end in
-  let parsign = Vars.subst_instance_context u parsign in
   let arproperlength = List.length mip.mind_arity_ctxt - List.length parsign in
   let arsign,_ = List.chop arproperlength mip.mind_arity_ctxt in
-  let subst = subst_of_rel_context_instance parsign params in
+  let subst = subst_of_rel_context_instance_list parsign params in
   let arsign = Vars.subst_instance_context u arsign in
   (substl_rel_context subst arsign, Inductive.inductive_sort_family mip)
 
@@ -428,14 +426,14 @@ let build_dependent_constructor cs =
   applist
     (mkConstructU cs.cs_cstr,
      (List.map (lift cs.cs_nargs) cs.cs_params)
-      @(Context.Rel.to_extended_list mkRel 0 cs.cs_args))
+      @(Context.Rel.instance_list mkRel 0 cs.cs_args))
 
 let build_dependent_inductive env ((ind, params) as indf) =
   let arsign,_ = get_arity env indf in
   let nrealargs = List.length arsign in
   applist
     (mkIndU ind,
-     (List.map (lift nrealargs) params)@(Context.Rel.to_extended_list mkRel 0 arsign))
+     (List.map (lift nrealargs) params)@(Context.Rel.instance_list mkRel 0 arsign))
 
 (* builds the arity of an elimination predicate in sort [s] *)
 
@@ -486,16 +484,14 @@ let compute_projections env (kn, i as ind) =
   in
   let pkt = mib.mind_packets.(i) in
   let { mind_nparams = nparamargs; mind_params_ctxt = params } = mib in
-  let subst = List.init mib.mind_ntypes (fun i -> mkIndU ((kn, mib.mind_ntypes - i - 1), u)) in
-  let ctx, cty = pkt.mind_nf_lc.(0) in
-  let rctx, _ = decompose_prod_assum (substl subst (Term.it_mkProd_or_LetIn cty ctx)) in
-  let ctx, paramslet = List.chop pkt.mind_consnrealdecls.(0) rctx in
+  let ctx, _ = pkt.mind_nf_lc.(0) in
+  let ctx, paramslet = List.chop pkt.mind_consnrealdecls.(0) ctx in
   (* We build a substitution smashing the lets in the record parameters so
      that typechecking projections requires just a substitution and not
      matching with a parameter context. *)
   let indty =
     (* [ty] = [Ind inst] is typed in context [params] *)
-    let inst = Context.Rel.to_extended_vect mkRel 0 paramslet in
+    let inst = Context.Rel.instance mkRel 0 paramslet in
     let indu = mkIndU (ind, u) in
     let ty = mkApp (indu, inst) in
     (* [Ind inst] is typed in context [params-wo-let] *)
@@ -519,7 +515,8 @@ let compute_projections env (kn, i as ind) =
       match na.binder_name with
       | Name id ->
         let lab = Label.of_id id in
-        let kn = Projection.Repr.make ind ~proj_npars:mib.mind_nparams ~proj_arg lab in
+        let proj_relevant = match na.binder_relevance with Sorts.Irrelevant -> false | Sorts.Relevant -> true in
+        let kn = Projection.Repr.make ind ~proj_relevant ~proj_npars:mib.mind_nparams ~proj_arg lab in
         (* from [params, field1,..,fieldj |- t(params,field1,..,fieldj)]
            to [params, x:I, field1,..,fieldj |- t(params,field1,..,fieldj] *)
         let t = liftn 1 j t in
@@ -679,6 +676,9 @@ let arity_of_case_predicate env (ind,params) dep k =
 (* Inferring the sort of parameters of a polymorphic inductive type
    knowing the sort of the conclusion *)
 
+let univ_level_mem l s = match s with
+| Prop | Set | SProp -> false
+| Type u -> univ_level_mem l u
 
 (* Compute the inductive argument types: replace the sorts
    that appear in the type of the inductive by the sort of the
