@@ -49,7 +49,7 @@ let reduction_effect_hook env sigma con c =
     effect env sigma (Lazy.force c)
   with Not_found -> ()
 
-let cache_reduction_effect (_,(con,funkey)) =
+let cache_reduction_effect (con,funkey) =
   constant_effect_table := Cmap.add con funkey !constant_effect_table
 
 let subst_reduction_effect (subst,(con,funkey)) =
@@ -67,7 +67,7 @@ let declare_reduction_effect funkey f =
 
 (** A function to set the value of the print function *)
 let set_reduction_effect x funkey =
-  Lib.add_anonymous_leaf (inReductionEffect (x,funkey))
+  Lib.add_leaf (inReductionEffect (x,funkey))
 
 
 (** Machinery to custom the behavior of the reduction *)
@@ -92,45 +92,45 @@ module ReductionBehaviour = struct
   let table =
     Summary.ref (GlobRef.Map.empty : t GlobRef.Map.t) ~name:"reductionbehaviour"
 
-  let load _ (_,(_,(r, b))) =
+  let load _ (_,(r, b)) =
     table := GlobRef.Map.add r b !table
 
   let cache o = load 1 o
 
-  let classify (local,_ as o) = if local then Dispose else Substitute o
+  let classify (local,_) = if local then Dispose else Substitute
 
   let subst (subst, (local, (r,o) as orig)) =
     let r' = subst_global_reference subst r in if r==r' then orig
     else (local,(r',o))
 
   let discharge = function
-    | _,(false, (gr, b)) ->
+    | false, (gr, b) ->
       let b =
         if Lib.is_in_section gr then
-          let vars = Lib.variable_section_segment_of_reference gr in
-          let extra = List.length vars in
+          let vars = Lib.section_instance gr in
+          let extra = Array.length vars in
           more_args extra b
         else b
       in
       Some (false, (gr, b))
-    | _ -> None
+    | true, _ -> None
 
   let rebuild = function
     | req, (GlobRef.ConstRef c, _ as x) -> req, x
     | _ -> assert false
 
   let inRedBehaviour = declare_object {
-                        (default_object "REDUCTIONBEHAVIOUR") with
-                        load_function = load;
-                        cache_function = cache;
-                        classify_function = classify;
-                        subst_function = subst;
-                        discharge_function = discharge;
-                        rebuild_function = rebuild;
-                      }
+      (default_object "REDUCTIONBEHAVIOUR") with
+      load_function = load;
+      cache_function = cache;
+      classify_function = classify;
+      subst_function = subst;
+      discharge_function = discharge;
+      rebuild_function = rebuild;
+    }
 
   let set ~local r b =
-    Lib.add_anonymous_leaf (inRedBehaviour (local, (r, b)))
+    Lib.add_leaf (inRedBehaviour (local, (r, b)))
 
   let get r = GlobRef.Map.find_opt r !table
 
@@ -210,7 +210,7 @@ sig
   val zip : evar_map -> constr * t -> constr
   val check_native_args : CPrimitives.t -> t -> bool
   val get_next_primitive_args : CPrimitives.args_red -> t -> CPrimitives.args_red * (t * EConstr.t * t) option
-  val expand_case : env -> evar_map -> case_stk -> constr * constr array
+  val expand_case : env -> evar_map -> case_stk -> case_info * EInstance.t * constr array * (rel_context * constr) * (rel_context * constr) array
 end =
 struct
   open EConstr
@@ -420,8 +420,8 @@ struct
 
   let expand_case env sigma ((ci, u, pms, t, iv, br) : case_stk) =
     let dummy = mkProp in
-    let (_, t, _, _, br) = EConstr.expand_case env sigma (ci, u, pms, t, iv, dummy, br) in
-    (t, br)
+    let (ci, u, pms, t, _, _, br) = EConstr.annotate_case env sigma (ci, u, pms, t, iv, dummy, br) in
+    (ci, u, pms, t, br)
 
 end
 
@@ -446,8 +446,6 @@ let pr_state env sigma (tm,sk) =
 (*** Reduction Functions Operators ***)
 (*************************************)
 
-let safe_evar_value = Evarutil.safe_evar_value
-
 let safe_meta_value sigma ev =
   try Some (Evd.meta_value sigma ev)
   with Not_found -> None
@@ -456,24 +454,19 @@ let safe_meta_value sigma ev =
 (*** Reduction using bindingss ***)
 (*************************************)
 
-let eta = CClosure.RedFlags.mkflags [CClosure.RedFlags.fETA]
-
 (* Beta Reduction tools *)
 
-let apply_subst recfun env sigma t stack =
+let apply_subst env sigma t stack =
   let rec aux env t stack =
     match (Stack.decomp stack, EConstr.kind sigma t) with
     | Some (h,stacktl), Lambda (_,_,c) ->
        aux (h::env) c stacktl
-    | _ -> recfun sigma (substl env t, stack)
-  in aux env t stack
-
-let stacklam recfun env sigma t stack =
-  apply_subst (fun _ s -> recfun s) env sigma t stack
+    | _ -> (substl env t, stack)
+  in
+  aux env t stack
 
 let beta_applist sigma (c,l) =
-  let zip s = Stack.zip sigma s in
-  stacklam zip [] sigma c (Stack.append_app_list l Stack.empty)
+  Stack.zip sigma (apply_subst [] sigma c (Stack.append_app_list l Stack.empty))
 
 (* Iota reduction tools *)
 
@@ -491,35 +484,29 @@ let contract_cofix sigma (bodynum,(names,types,bodies as typedbodies)) =
   substl closure bodies.(bodynum)
 
 (** Similar to the "fix" case below *)
-let reduce_and_refold_cofix recfun env sigma cofix sk =
-  let raw_answer =
-    contract_cofix sigma cofix in
-  apply_subst
-    (fun _ (t,sk') -> recfun (t,sk'))
-    [] sigma raw_answer sk
+let reduce_and_refold_cofix env sigma cofix sk =
+  let raw_answer = contract_cofix sigma cofix in
+  apply_subst [] sigma raw_answer sk
 
 (* contracts fix==FIX[nl;i](A1...Ak;[F1...Fk]{B1....Bk}) to produce
    Bi[Fj --> FIX[nl;j](A1...Ak;[F1...Fk]{B1...Bk})] *)
 
 let contract_fix sigma ((recindices,bodynum),(names,types,bodies as typedbodies)) =
-    let nbodies = Array.length recindices in
-    let make_Fi j =
-      let ind = nbodies-j-1 in
-      mkFix ((recindices,ind),typedbodies)
-    in
-    let closure = List.init nbodies make_Fi in
-    substl closure bodies.(bodynum)
+  let nbodies = Array.length recindices in
+  let make_Fi j =
+    let ind = nbodies-j-1 in
+    mkFix ((recindices,ind),typedbodies)
+  in
+  let closure = List.init nbodies make_Fi in
+  substl closure bodies.(bodynum)
 
 (** First we substitute the Rel bodynum by the fixpoint and then we try to
     replace the fixpoint by the best constant from [cst_l]
     Other rels are directly substituted by constants "magically found from the
     context" in contract_fix *)
-let reduce_and_refold_fix recfun env sigma fix sk =
-  let raw_answer =
-    contract_fix sigma fix in
-  apply_subst
-    (fun _ (t,sk') -> recfun (t,sk'))
-    [] sigma raw_answer sk
+let reduce_and_refold_fix env sigma fix sk =
+  let raw_answer = contract_fix sigma fix in
+  apply_subst [] sigma raw_answer sk
 
 open Primred
 
@@ -686,7 +673,7 @@ let apply_branch env sigma (ind, i) args (ci, u, pms, iv, r, lf) =
     let ctx = expand_branch env sigma u pms (ind, i) br in
     applist (it_mkLambda_or_LetIn (snd br) ctx, args)
 
-let rec whd_state_gen flags env sigma =
+let whd_state_gen flags env sigma =
   let open Context.Named.Declaration in
   let rec whrec (x, stack) : state =
     let () =
@@ -742,7 +729,7 @@ let rec whd_state_gen flags env sigma =
       whrec stack'
 
     | LetIn (_,b,_,c) when CClosure.RedFlags.red_set flags CClosure.RedFlags.fZETA ->
-      apply_subst (fun _ -> whrec) [b] sigma c stack
+      whrec (apply_subst [b] sigma c stack)
     | Cast (c,_,_) -> whrec (c, stack)
     | App (f,cl)  ->
       whrec
@@ -750,23 +737,7 @@ let rec whd_state_gen flags env sigma =
     | Lambda (na,t,c) ->
       (match Stack.decomp stack with
       | Some _ when CClosure.RedFlags.red_set flags CClosure.RedFlags.fBETA ->
-        apply_subst (fun _ -> whrec) [] sigma x stack
-      | None when CClosure.RedFlags.red_set flags CClosure.RedFlags.fETA ->
-        let env' = push_rel (LocalAssum (na, t)) env in
-        let whrec' = whd_state_gen flags env' sigma in
-        (match EConstr.kind sigma (Stack.zip sigma (whrec' (c, Stack.empty))) with
-        | App (f,cl) ->
-          let napp = Array.length cl in
-          if napp > 0 then
-            let (x', l') = whrec' (Array.last cl, Stack.empty) in
-            match EConstr.kind sigma x', l' with
-            | Rel 1, [] ->
-              let lc = Array.sub cl 0 (napp-1) in
-              let u = if Int.equal napp 1 then f else mkApp (f,lc) in
-              if noccurn sigma 1 u then (pop u,Stack.empty) else fold ()
-            | _ -> fold ()
-          else fold ()
-        | _ -> fold ())
+        whrec (apply_subst [] sigma x stack)
       | _ -> fold ())
 
     | Case (ci,u,pms,p,iv,d,lf) ->
@@ -791,7 +762,7 @@ let rec whd_state_gen flags env sigma =
         |args, (Stack.Fix (f,s')::s'') when use_fix ->
           let x' = Stack.zip sigma (x, args) in
           let out_sk = s' @ (Stack.append_app [|x'|] s'') in
-          reduce_and_refold_fix whrec env sigma f out_sk
+          whrec (reduce_and_refold_fix env sigma f out_sk)
         |_, (Stack.App _)::_ -> assert false
         |_, _ -> fold ()
       else fold ()
@@ -800,7 +771,7 @@ let rec whd_state_gen flags env sigma =
       if CClosure.RedFlags.red_set flags CClosure.RedFlags.fCOFIX then
         match Stack.strip_app stack with
         |args, ((Stack.Case _ |Stack.Proj _)::s') ->
-          reduce_and_refold_cofix whrec env sigma cofix stack
+          whrec (reduce_and_refold_cofix env sigma cofix stack)
         |_ -> fold ()
       else fold ()
 
@@ -841,27 +812,13 @@ let local_whd_state_gen flags env sigma =
     let s = (EConstr.of_kind c0, stack) in
     match c0 with
     | LetIn (_,b,_,c) when CClosure.RedFlags.red_set flags CClosure.RedFlags.fZETA ->
-      stacklam whrec [b] sigma c stack
+      whrec (apply_subst [b] sigma c stack)
     | Cast (c,_,_) -> whrec (c, stack)
     | App (f,cl)  -> whrec (f, Stack.append_app cl stack)
     | Lambda (_,_,c) ->
       (match Stack.decomp stack with
       | Some (a,m) when CClosure.RedFlags.red_set flags CClosure.RedFlags.fBETA ->
-        stacklam whrec [a] sigma c m
-      | None when CClosure.RedFlags.red_set flags CClosure.RedFlags.fETA ->
-        (match EConstr.kind sigma (Stack.zip sigma (whrec (c, Stack.empty))) with
-        | App (f,cl) ->
-          let napp = Array.length cl in
-          if napp > 0 then
-            let x', l' = whrec (Array.last cl, Stack.empty) in
-            match EConstr.kind sigma x', l' with
-            | Rel 1, [] ->
-              let lc = Array.sub cl 0 (napp-1) in
-              let u = if Int.equal napp 1 then f else mkApp (f,lc) in
-              if noccurn sigma 1 u then (pop u,Stack.empty) else s
-            | _ -> s
-          else s
-        | _ -> s)
+        whrec (apply_subst [a] sigma c m)
       | _ -> s)
 
     | Proj (p,c) when CClosure.RedFlags.red_projection flags p ->
@@ -970,11 +927,35 @@ let whd_allnolet_state = raw_whd_state_gen CClosure.allnolet
 let whd_allnolet_stack = stack_red_of_state_red whd_allnolet_state
 let whd_allnolet = red_of_state_red whd_allnolet_state
 
-(* 4. Ad-hoc eta reduction, does not substitute evars *)
+let is_head_evar env sigma c =
+  let head, _ = whd_all_state env sigma (c,Stack.empty) in
+  EConstr.isEvar sigma head
 
-let shrink_eta env c =
-  let evd = Evd.from_env env in
-  Stack.zip evd (local_whd_state_gen eta env evd (c,Stack.empty))
+(* 4. Ad-hoc eta reduction *)
+
+let shrink_eta sigma c =
+  let rec whrec x = match EConstr.kind sigma x with
+    | Cast (c, _, _) -> whrec c
+    | Lambda (_, _, c) ->
+      let (f, cl) = decompose_app_vect sigma (whrec c) in
+      let napp = Array.length cl in
+      if napp > 0 then
+        let x' = whrec (Array.last cl) in
+        match EConstr.kind sigma x' with
+        | Rel 1 ->
+          let lc = Array.sub cl 0 (napp-1) in
+          let u = mkApp (f, lc) in
+          if noccurn sigma 1 u then pop u else x
+        | _ -> x
+      else x
+    | Meta ev ->
+      (match safe_meta_value sigma ev with
+        Some c -> whrec c
+      | None -> x)
+    | App _ | Case _ | Fix _ | Construct _ | CoFix _ | Evar _ | Rel _ | Var _ | Sort _ | Prod _
+    | LetIn _ | Const _  | Ind _ | Proj _ | Int _ | Float _ | Array _ -> x
+  in
+  whrec c
 
 (* 5. Zeta Reduction Functions *)
 
@@ -995,18 +976,16 @@ let nf_evar = Evarutil.nf_evar
    a [nf_evar] here *)
 let clos_norm_flags flgs env sigma t =
   try
-    let evars ev = safe_evar_value sigma ev in
     EConstr.of_constr (CClosure.norm_val
-      (CClosure.create_clos_infos ~univs:(Evd.universes sigma) ~evars flgs env)
+      (Evarutil.create_clos_infos env sigma flgs)
       (CClosure.create_tab ())
       (CClosure.inject (EConstr.Unsafe.to_constr t)))
   with e when is_anomaly e -> user_err Pp.(str "Tried to normalize ill-typed term")
 
 let clos_whd_flags flgs env sigma t =
   try
-    let evars ev = safe_evar_value sigma ev in
     EConstr.of_constr (CClosure.whd_val
-      (CClosure.create_clos_infos ~univs:(Evd.universes sigma) ~evars flgs env)
+      (Evarutil.create_clos_infos env sigma flgs)
       (CClosure.create_tab ())
       (CClosure.inject (EConstr.Unsafe.to_constr t)))
   with e when is_anomaly e -> user_err Pp.(str "Tried to normalize ill-typed term")
@@ -1030,7 +1009,7 @@ let is_transparent e k =
 
 (* Conversion utility functions *)
 
-type conversion_test = Constraint.t -> Constraint.t
+type conversion_test = Constraints.t -> Constraints.t
 
 let pb_is_equal pb = pb == Reduction.CONV
 
@@ -1071,7 +1050,7 @@ let f_conv_leq ?l2r ?reds env ?evars x y =
 
 let test_trans_conversion (f: constr Reduction.extended_conversion_function) reds env sigma x y =
   try
-    let evars ev = safe_evar_value sigma ev in
+    let evars ev = existential_opt_value0 sigma ev in
     let env = Environ.set_universes (Evd.universes sigma) env in
     let _ = f ~reds env ~evars x y in
     true
@@ -1092,9 +1071,9 @@ let check_conv ?(pb=Reduction.CUMUL) ?(ts=TransparentState.full) env sigma x y =
     | Reduction.CUMUL -> f_conv_leq
   in
     let env = Environ.set_universes (Evd.universes sigma) env in
-    try f ~reds:ts env ~evars:(safe_evar_value sigma) x y; true
+    try f ~reds:ts env ~evars:(existential_opt_value0 sigma) x y; true
     with Reduction.NotConvertible -> false
-    | Univ.UniverseInconsistency _ -> false
+    | UGraph.UniverseInconsistency _ -> false
     | e ->
       let e = Exninfo.capture e in
       report_anomaly e
@@ -1107,7 +1086,7 @@ let sigma_compare_sorts env pb s0 s1 sigma =
 let sigma_compare_instances ~flex i0 i1 sigma =
   try Evd.set_eq_instances ~flex sigma i0 i1
   with Evd.UniversesDiffer
-     | Univ.UniverseInconsistency _ ->
+     | UGraph.UniverseInconsistency _ ->
         raise Reduction.NotConvertible
 
 let sigma_check_inductive_instances cv_pb variance u1 u2 sigma =
@@ -1122,9 +1101,35 @@ let sigma_univ_state =
     compare_instances = sigma_compare_instances;
     compare_cumul_instances = sigma_check_inductive_instances; }
 
+let univproblem_compare_sorts env pb s0 s1 uset =
+  let open UnivProblem in
+  match pb with
+  | Reduction.CONV -> UnivProblem.Set.add (UEq (s0, s1)) uset
+  | Reduction.CUMUL -> UnivProblem.Set.add (ULe (s0, s1)) uset
+
+let univproblem_compare_instances ~flex i0 i1 uset =
+  UnivProblem.enforce_eq_instances_univs flex i0 i1 uset
+
+let univproblem_check_inductive_instances cv_pb variances u u' uset =
+  let open UnivProblem in
+  let open Univ.Variance in
+  let mk u = Sorts.sort_of_univ @@ Univ.Universe.make u in
+  let fold cstr v u u' = match v with
+  | Irrelevant -> Set.add (UWeak (u,u')) cstr
+  | Covariant when cv_pb == Reduction.CUMUL -> Set.add (ULe (mk u, mk u')) cstr
+  | Covariant | Invariant -> Set.add (UEq (mk u, mk u')) cstr
+  in
+  Array.fold_left3 fold uset
+    variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
+
+let univproblem_univ_state =
+  let open Reduction in
+  { compare_sorts = univproblem_compare_sorts;
+    compare_instances = univproblem_compare_instances;
+    compare_cumul_instances = univproblem_check_inductive_instances; }
+
 let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Reduction.CUMUL)
     ?(ts=TransparentState.full) env sigma x y =
-  (* FIXME *)
   try
       let ans = match pb with
       | Reduction.CUMUL ->
@@ -1136,7 +1141,7 @@ let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Reduction.CUMUL)
       | None -> None
       | Some cstr ->
         try Some (Evd.add_universe_constraints sigma cstr)
-        with Univ.UniverseInconsistency _ | Evd.UniversesDiffer -> None
+        with UGraph.UniverseInconsistency _ | Evd.UniversesDiffer -> None
       in
       match ans with
       | Some sigma -> ans
@@ -1150,19 +1155,55 @@ let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Reduction.CUMUL)
         Some sigma'
   with
   | Reduction.NotConvertible -> None
-  | Univ.UniverseInconsistency _ when catch_incon -> None
+  | UGraph.UniverseInconsistency _ when catch_incon -> None
   | e ->
     let e = Exninfo.capture e in
     report_anomaly e
 
 let infer_conv = infer_conv_gen (fun pb ~l2r sigma ->
-      Reduction.generic_conv pb ~l2r (safe_evar_value sigma))
+      Reduction.generic_conv pb ~l2r (existential_opt_value0 sigma))
 
-(* This reference avoids always having to link C code with the kernel *)
-let vm_infer_conv = ref (infer_conv ~catch_incon:true ~ts:TransparentState.full)
-let set_vm_infer_conv f = vm_infer_conv := f
-let vm_infer_conv ?(pb=Reduction.CUMUL) env t1 t2 =
-  !vm_infer_conv ~pb env t1 t2
+let infer_conv_ustate ?(catch_incon=true) ?(pb=Reduction.CUMUL)
+    ?(ts=TransparentState.full) env sigma x y =
+  try
+      let ans = match pb with
+      | Reduction.CUMUL ->
+          EConstr.leq_constr_universes env sigma x y
+      | Reduction.CONV ->
+          EConstr.eq_constr_universes env sigma x y
+      in
+      match ans with
+      | Some cstr -> Some cstr
+      | None ->
+        let x = EConstr.Unsafe.to_constr x in
+        let y = EConstr.Unsafe.to_constr y in
+        let env = Environ.set_universes (Evd.universes sigma) env in
+        let cstr =
+          Reduction.generic_conv pb ~l2r:false (existential_opt_value0 sigma) ts
+            env (UnivProblem.Set.empty, univproblem_univ_state) x y in
+        Some cstr
+  with
+  | Reduction.NotConvertible -> None
+  | UGraph.UniverseInconsistency _ when catch_incon -> None
+  | e ->
+    let e = Exninfo.capture e in
+    report_anomaly e
+
+let vm_infer_conv ?(pb=Reduction.CUMUL) env sigma t1 t2 =
+  infer_conv_gen (fun pb ~l2r sigma ts ->
+      Vconv.vm_conv_gen pb (Evd.existential_opt_value0 sigma))
+    ~catch_incon:true ~pb env sigma t1 t2
+
+let native_conv_generic pb sigma t =
+  let evars_of_evar_map sigma =
+    { Nativelambda.evars_val = Evd.existential_opt_value0 sigma;
+      Nativelambda.evars_metas = Evd.meta_type0 sigma }
+  in
+  Nativeconv.native_conv_gen pb (evars_of_evar_map sigma) t
+
+let native_infer_conv ?(pb=Reduction.CUMUL) env sigma t1 t2 =
+  infer_conv_gen (fun pb ~l2r sigma ts -> native_conv_generic pb sigma)
+    ~catch_incon:true ~pb env sigma t1 t2
 
 (********************************************************************)
 (*             Special-Purpose Reduction                            *)
@@ -1353,8 +1394,7 @@ let whd_betaiota_deltazeta_for_iota_state ts env sigma s =
   let env' = Environ.set_typing_flags { (Environ.typing_flags env) with Declarations.share_reduction = false } env in
   let whd_opt c =
     let open CClosure in
-    let evars ev = safe_evar_value sigma ev in
-    let infos = create_clos_infos ~evars all' env' in
+    let infos = Evarutil.create_clos_infos env' sigma all' in
     let tab = create_tab () in
     let c = inject (EConstr.Unsafe.to_constr (Stack.zip sigma c)) in
     let (c, stk) = whd_stack infos tab c [] in
@@ -1444,3 +1484,44 @@ let nf_meta env sigma c =
   let sigma = create_meta_instance_subst sigma in
   let cl = mk_freelisted c in
   meta_instance env sigma { cl with rebus = cl.rebus }
+
+module Infer = struct
+
+open Reduction
+
+let infer_eq (univs, cstrs as cuniv) u u' =
+  if UGraph.check_eq_sort univs u u' then cuniv
+  else
+    univs, (UnivSubst.enforce_eq_sort u u' cstrs)
+
+let infer_leq (univs, cstrs as cuniv) u u' =
+  if UGraph.check_leq_sort univs u u' then cuniv
+  else
+    let cstrs', _ = UnivSubst.enforce_leq_alg_sort u u' univs in
+      univs, Univ.Constraints.union cstrs cstrs'
+
+let infer_cmp_universes _env pb s0 s1 univs =
+  match pb with
+  | CUMUL -> infer_leq univs s0 s1
+  | CONV -> infer_eq univs s0 s1
+
+let infer_convert_instances ~flex u u' (univs,cstrs) =
+  let cstrs' =
+    if flex then
+      if UGraph.check_eq_instances univs u u' then cstrs
+      else raise NotConvertible
+    else Univ.enforce_eq_instances u u' cstrs
+  in (univs, cstrs')
+
+let infer_inductive_instances cv_pb variance u1 u2 (univs,csts') =
+  let csts = get_cumulativity_constraints cv_pb variance u1 u2 in
+  (univs, Univ.Constraints.union csts csts')
+
+let inferred_universes : (UGraph.t * Univ.Constraints.t) universe_compare =
+  { compare_sorts = infer_cmp_universes;
+    compare_instances = infer_convert_instances;
+    compare_cumul_instances = infer_inductive_instances; }
+
+end
+
+let inferred_universes = Infer.inferred_universes
