@@ -28,7 +28,6 @@ open Evardefine
 open Evarsolve
 open Pretype_errors
 open Retyping
-open Coercion
 open Locus
 open Locusops
 open Find_subterm
@@ -570,7 +569,7 @@ let force_eqs c =
   Set.fold
     (fun c acc ->
        let c' = match c with
-         | ULub (l, r) -> UEq (Univ.Universe.make l,Univ.Universe.make r)
+         | ULub (l, r) -> UEq (Sorts.sort_of_univ @@ Univ.Universe.make l, Sorts.sort_of_univ @@ Univ.Universe.make r)
          | ULe _ | UEq _ | UWeak _ -> c
        in
         Set.add c' acc)
@@ -584,11 +583,11 @@ let constr_cmp pb env sigma flags ?nargs t u =
   match cstrs with
   | Some cstrs ->
       begin try Some (Evd.add_universe_constraints sigma cstrs)
-      with Univ.UniverseInconsistency _ -> None
+      with UGraph.UniverseInconsistency _ -> None
       | Evd.UniversesDiffer ->
         if is_rigid_head sigma flags t then
           try Some (Evd.add_universe_constraints sigma (force_eqs cstrs))
-          with Univ.UniverseInconsistency _ -> None
+          with UGraph.UniverseInconsistency _ -> None
         else None
       end
   | None ->
@@ -638,6 +637,18 @@ let check_compatibility env pbty flags (sigma,metasubst,evarsubst : subst0) tyM
       | None -> error_cannot_unify env sigma (m,n)
     else sigma
 
+let check_compatibility_ustate env pbty flags (sigma,metasubst,evarsubst : subst0) tyM tyN =
+  match subst_defined_metas_evars sigma (metasubst,[]) tyM with
+  | None -> UnivProblem.Set.empty
+  | Some m ->
+  match subst_defined_metas_evars sigma (metasubst,[]) tyN with
+  | None -> UnivProblem.Set.empty
+  | Some n ->
+    if is_ground_term sigma m && is_ground_term sigma n then
+      match infer_conv_ustate ~pb:pbty ~ts:flags.modulo_delta_types env sigma m n with
+      | Some uprob -> uprob
+      | None -> error_cannot_unify env sigma (m,n)
+    else UnivProblem.Set.empty
 
 let rec is_neutral env sigma ts t =
   let (f, l) = decompose_app_vect sigma t in
@@ -863,15 +874,18 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
              | _ -> raise ex)
 
         | Case (ci1, u1, pms1, p1, iv1, c1, cl1), Case (ci2, u2, pms2, p2, iv2, c2, cl2) ->
-          let (ci1, p1, iv1, c1, cl1) = EConstr.expand_case env sigma (ci1, u1, pms1, p1, iv1, c1, cl1) in
-          let (ci2, p2, iv2, c2, cl2) = EConstr.expand_case env sigma (ci2, u2, pms2, p2, iv2, c2, cl2) in
             (try
-             if not (Ind.CanOrd.equal ci1.ci_ind ci2.ci_ind) then error_cannot_unify curenv sigma (cM,cN);
+             let () = if not (Ind.CanOrd.equal ci1.ci_ind ci2.ci_ind) then error_cannot_unify curenv sigma (cM,cN) in
              let opt' = {opt with at_top = true; with_types = false} in
-               Array.fold_left2 (unirec_rec curenvnb CONV {opt with at_top = true} ~nargs:0)
-               (unirec_rec curenvnb CONV opt'
-                (unirec_rec curenvnb CONV opt' substn p1 p2) c1 c2)
-                 cl1 cl2
+             let substn = Array.fold_left2 (unirec_rec curenvnb CONV ~nargs:0 opt') substn pms1 pms2 in
+             let (ci1, _, _, p1, _, c1, cl1) = EConstr.annotate_case env sigma (ci1, u1, pms1, p1, iv1, c1, cl1) in
+             let unif opt substn (ctx1, c1) (_, c2) =
+               let curenvnb' = List.fold_right (fun decl (env, n) -> push_rel decl env, n + 1) ctx1 curenvnb in
+               unirec_rec curenvnb' CONV opt' substn c1 c2
+             in
+             let substn = unif opt' substn p1 p2 in
+             let substn = unirec_rec curenvnb CONV opt' substn c1 c2 in
+             Array.fold_left2 (unif {opt with at_top = true}) substn cl1 cl2
              with ex when precatchable_exception ex ->
                reduce curenvnb pb opt substn cM cN)
 
@@ -1031,19 +1045,23 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
       | None -> (* some undefined Metas in cN *) None
       | Some n1 ->
          (* No subterm restriction there, too much incompatibilities *)
-         let sigma =
+         let uprob =
            if opt.with_types then
              try (* Ensure we call conversion on terms of the same type *)
                let tyM = get_type_of curenv ~lax:true sigma m1 in
                let tyN = get_type_of curenv ~lax:true sigma n1 in
-               check_compatibility curenv CUMUL flags substn tyM tyN
+               check_compatibility_ustate curenv CUMUL flags substn tyM tyN
              with RetypeError _ ->
-               (* Renounce, maybe metas/evars prevents typing *) sigma
-           else sigma
+               (* Renounce, maybe metas/evars prevents typing *) UnivProblem.Set.empty
+           else UnivProblem.Set.empty
          in
-        match infer_conv ~pb ~ts:convflags curenv sigma m1 n1 with
-        | Some sigma ->
-          Some (sigma, metasubst, evarsubst)
+        match infer_conv_ustate ~pb ~ts:convflags curenv sigma m1 n1 with
+        | Some uprob' ->
+          let uprob = UnivProblem.Set.union uprob uprob' in
+          begin match Evd.add_universe_constraints sigma uprob with
+          | sigma -> Some (sigma, metasubst, evarsubst)
+          | exception (UGraph.UniverseInconsistency _ | UniversesDiffer) -> None
+          end
         | None ->
           if is_ground_term sigma m1 && is_ground_term sigma n1 then
             error_cannot_unify curenv sigma (cM,cN)
@@ -1139,7 +1157,11 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
     else error_cannot_unify (fst curenvnb) sigma (cM,cN)
   in
 
-  debug_tactic_unification (fun () -> str "Starting unification");
+  debug_tactic_unification (fun () ->
+      str "Starting unification: " ++
+      Termops.Internal.print_constr_env env sigma (fst m) ++ str" ~= " ++
+      Termops.Internal.print_constr_env env sigma (fst n));
+
   let opt = { at_top = conv_at_top; with_types = false; with_cs = true } in
   try
   let res =
@@ -1313,7 +1335,7 @@ let is_mimick_head sigma ts f =
 
 let try_to_coerce env evd c cty tycon =
   let j = make_judge c cty in
-  let (evd',j',_trace) = inh_conv_coerce_rigid_to ~program_mode:false true env evd j tycon in
+  let (evd',j',_trace) = Coercion.inh_conv_coerce_rigid_to ~program_mode:false ~resolve_tc:true env evd j tycon in
   let evd' = Evarconv.solve_unif_constraints_with_heuristics env evd' in
   let evd' = Evd.map_metas_fvalue (fun c -> nf_evar evd' c) evd' in
     (evd',j'.uj_val)
@@ -1574,10 +1596,6 @@ let indirectly_dependent sigma c d decls =
     let open Context.Named.Declaration in
     List.exists (fun d' -> exists (fun c -> Termops.local_occur_var sigma (NamedDecl.get_id d') c) d) decls
 
-let finish_evar_resolution ?(flags=Pretyping.all_and_fail_flags) env current_sigma (pending,c) =
-  let sigma = Pretyping.solve_remaining_evars flags env current_sigma ~initial:pending in
-  (sigma, nf_evar sigma c)
-
 let default_matching_core_flags sigma =
   let ts = TransparentState.full in {
   modulo_conv_on_closed_terms = Some TransparentState.empty;
@@ -1656,7 +1674,7 @@ let make_pattern_test from_prefix_of_ind is_correct_type env sigma (pending,c) =
     | e when CErrors.noncritical e -> raise (NotUnifiable None) in
   let merge_fun c1 c2 =
     match c1, c2 with
-    | Some (evd,c1,x), Some (_,c2,_) ->
+    | Some (_,c1,x), Some (evd,c2,_) ->
       begin match infer_conv ~pb:CONV env evd c1 c2 with
       | Some evd ->
        (let t1 = get_type_of env evd c1 in
@@ -2068,7 +2086,8 @@ let w_unify_to_subterm_all env evd ?(flags=default_unify_flags ()) (op,cl) =
   match res with
   | [] ->
     raise (PretypeError (env,evd,NoOccurrenceFound (op, None)))
-  | _ -> res
+  | _ ->
+    List.map fst res
 
 let w_unify_to_subterm_list env evd flags hdmeta oplist t =
   List.fold_right
