@@ -422,9 +422,9 @@ let clenv_instantiate ?(flags=fchain_flags ()) mv clenv (c, ty) =
   let clenv = clenv_unify ~flags CUMUL ty (clenv_meta_type clenv mv) clenv in
   clenv_assign mv c clenv
 
-let clenv_fchain ?with_univs ?(flags=fchain_flags ()) mv clenv nextclenv =
+let clenv_fchain ?(flags=fchain_flags ()) mv clenv nextclenv =
   (* Add the metavars of [nextclenv] to [clenv], with their name-environment *)
-  let evd = meta_merge ?with_univs nextclenv.evd clenv.evd in
+  let evd = meta_merge nextclenv.evd clenv.evd in
   let clenv' =
     { templval = clenv.templval;
       templtyp = clenv.templtyp;
@@ -764,7 +764,8 @@ let make_evar_clause env sigma ?len t =
       let inst, ctx, args, identity, subst = match inst with
       | None ->
         (* Dummy type *)
-        let ctx, _, args, subst = push_rel_context_to_named_context env sigma mkProp in
+        let hypnaming = RenameExistingBut (VarSet.variables (Global.env ())) in
+        let ctx, _, args, subst = push_rel_context_to_named_context ~hypnaming env sigma mkProp in
         let id = if Int.equal (Environ.nb_rel env) 0 then Identity.make args else Identity.none () in
         Some (ctx, args, id, subst), ctx, args, id, subst
       | Some (ctx, args, id, subst) -> inst, ctx, args, id, subst
@@ -820,7 +821,7 @@ let define_with_type sigma env ev c =
   let t = Retyping.get_type_of env sigma ev in
   let ty = Retyping.get_type_of env sigma c in
   let j = Environ.make_judge c ty in
-  let (sigma, j, _trace) = Coercion.inh_conv_coerce_to ~program_mode:false true env sigma j t in
+  let (sigma, j, _trace) = Coercion.inh_conv_coerce_to ~program_mode:false ~resolve_tc:true env sigma j t in
   let (ev, _) = destEvar sigma ev in
   let sigma = Evd.define ev j.Environ.uj_val sigma in
   sigma
