@@ -23,11 +23,11 @@ type option_command =
 type injection_command =
   | OptionInjection of (Goptions.option_name * option_command)
   (** Set flags or options before the initial state is ready. *)
-  | RequireInjection of (string * string option * bool option)
+  | RequireInjection of (string * string option * Lib.export_flag option)
   (** Require libraries before the initial state is
      ready. Parameters follow [Library], that is to say,
      [lib,prefix,import_export] means require library [lib] from
-     optional [prefix] and [import_export] if [Some false/Some true]
+     optional [prefix] and [import_export] if [Some Lib.Import]/[Some Lib.Export]
       is used.  *)
   | WarnNoNative of string
   (** Used so that "-w -native-compiler-disabled -native-compiler yes"
@@ -72,7 +72,7 @@ type coqargs_pre = {
 type coqargs_query =
   | PrintWhere | PrintConfig
   | PrintVersion | PrintMachineReadableVersion
-  | PrintHelp of Usage.specific_usage
+  | PrintHelp of Boot.Usage.specific_usage
 
 type coqargs_main =
   | Queries of coqargs_query list
@@ -92,10 +92,9 @@ type t = {
 (* Default options *)
 val default : t
 
-val parse_args : usage:Usage.specific_usage -> init:t -> string list -> t * string list
+val parse_args : usage:Boot.Usage.specific_usage -> init:t -> string list -> t * string list
 
 val injection_commands : t -> injection_command list
-val build_load_path : t -> CUnix.physical_path list * Loadpath.vo_path list
 
 val dirpath_of_top : top -> Names.DirPath.t
 
