@@ -21,7 +21,6 @@ open CErrors
 open Util
 open Names
 open Constr
-open Context
 open Declarations
 open Mod_subst
 open Printer
@@ -60,8 +59,8 @@ let rec search_mind_label lab = function
 *)
 
 let rec fields_of_functor f subs mp0 args = function
-  |NoFunctor a -> f subs mp0 args a
-  |MoreFunctor (mbid,_,e) ->
+  | NoFunctor a -> f subs mp0 args a
+  | MoreFunctor (mbid,_,e) ->
     match args with
     | [] -> assert false (* we should only encounter applied functors *)
     | mpa :: args ->
@@ -96,9 +95,9 @@ and fields_of_mp mp =
   Modops.subst_structure subs fields
 
 and fields_of_mb subs mb args = match mb.mod_expr with
-  |Algebraic expr -> fields_of_expression subs mb.mod_mp args expr
-  |Struct sign -> fields_of_signature subs mb.mod_mp args sign
-  |Abstract|FullStruct -> fields_of_signature subs mb.mod_mp args mb.mod_type
+  | Algebraic expr -> fields_of_expression subs mb.mod_mp args expr
+  | Struct sign -> fields_of_signature subs mb.mod_mp args sign
+  | Abstract|FullStruct -> fields_of_signature subs mb.mod_mp args mb.mod_type
 
 (** The Abstract case above corresponds to [Declare Module] *)
 
@@ -109,11 +108,11 @@ and fields_of_signature x =
       (struc, mp0, subs)) x
 
 and fields_of_expr subs mp0 args = function
-  |MEident mp ->
+  | MEident mp ->
     let mb = lookup_module_in_impl (subst_mp subs mp) in
     fields_of_mb subs mb args
-  |MEapply (me1,mp2) -> fields_of_expr subs mp0 (mp2::args) me1
-  |MEwith _ -> assert false (* no 'with' in [mod_expr] *)
+  | MEapply (me1,mp2) -> fields_of_expr subs mp0 (mp2::args) me1
+  | MEwith _ -> assert false (* no 'with' in [mod_expr] *)
 
 and fields_of_expression x = fields_of_functor fields_of_expr x
 
@@ -192,14 +191,12 @@ let fold_with_full_binders g f n acc c =
 
 let get_constant_body kn =
   let cb = lookup_constant kn in
-  let env = Global.env () in
   let access = Library.indirect_accessor in
-  let otab = Environ.opaque_tables env in
   match cb.const_body with
   | Undef _ | Primitive _ -> None
   | Def c -> Some c
   | OpaqueDef o ->
-    match Opaqueproof.force_proof access otab o with
+    match Global.force_proof access o with
     | c, _ -> Some c
     | exception _ -> None (* missing delayed body, e.g. in vok mode *)
 
@@ -224,7 +221,7 @@ let rec traverse current ctx accu t =
       let (curr, data, ax2ty) = accu in
       let obj = ConstRef kn in
       let already_in = GlobRef.Map_env.mem obj data in
-      let data = if not already_in then GlobRef.Map_env.add obj GlobRef.Set_env.empty data else data in
+      let data = if not already_in then GlobRef.Map_env.add obj None data else data in
       let ty = (current, ctx, Vars.subst1 mkProp oty) in
       let ax2ty =
         try let l = GlobRef.Map_env.find obj ax2ty in GlobRef.Map_env.add obj (ty::l) ax2ty
@@ -243,12 +240,12 @@ and traverse_object (curr, data, ax2ty) body obj =
     if already_in then data, ax2ty
     else match body () (* Beware: this can be very costly *) with
     | None ->
-      GlobRef.Map_env.add obj GlobRef.Set_env.empty data, ax2ty
+      GlobRef.Map_env.add obj None data, ax2ty
     | Some body ->
       let contents,data,ax2ty =
         traverse (label_of obj) Context.Rel.empty
                  (GlobRef.Set_env.empty,data,ax2ty) body in
-      GlobRef.Map_env.add obj contents data, ax2ty
+      GlobRef.Map_env.add obj (Some contents) data, ax2ty
   in
   (GlobRef.Set_env.add obj curr, data, ax2ty)
 
@@ -263,33 +260,19 @@ and traverse_inductive (curr, data, ax2ty) mind obj =
    (* Invariant : I_0 \in data iff I_i \in data iff c_ij \in data
       where I_0, I_1, ... are in the same mutual definition and c_ij
       are all their constructors. *)
-   if GlobRef.Map_env.mem firstind_ref data then data, ax2ty else
+   if
+     (* recursive call: *) GlobRef.Set_env.mem firstind_ref curr ||
+     (* already in: *) GlobRef.Map_env.mem firstind_ref data
+   then data, ax2ty
+   else
+     (* Take into account potential recursivity of ind in itself *)
+     let curr = GlobRef.Set_env.add firstind_ref GlobRef.Set_env.empty in
+     let accu = (curr, data, ax2ty) in
      let mib = lookup_mind mind in
      (* Collects references of parameters *)
      let param_ctx = mib.mind_params_ctxt in
      let nparam = List.length param_ctx in
-     let accu =
-       traverse_context label Context.Rel.empty
-                        (GlobRef.Set_env.empty, data, ax2ty) param_ctx
-     in
-     (* Build the context of all arities *)
-     let arities_ctx =
-       let instance =
-         let open Univ in
-         Instance.of_array
-           (Array.init
-              (AUContext.size
-                 (Declareops.inductive_polymorphic_context mib))
-              Level.var)
-       in
-       Array.fold_left (fun accu oib ->
-          let pspecif = ((mib, oib), instance) in
-          let ind_type = Inductive.type_of_inductive pspecif in
-          let indr = oib.mind_relevance in
-          let ind_name = Name oib.mind_typename in
-          Context.Rel.add (Context.Rel.Declaration.LocalAssum (make_annot ind_name indr, ind_type)) accu)
-          Context.Rel.empty mib.mind_packets
-     in
+     let accu = traverse_context label Context.Rel.empty accu param_ctx in
      (* For each inductive, collects references in their arity and in the type
         of constructors*)
      let (contents, data, ax2ty) = Array.fold_left (fun accu oib ->
@@ -302,20 +285,21 @@ and traverse_inductive (curr, data, ax2ty) mind obj =
          in
          Array.fold_left (fun accu cst_typ ->
             let param_ctx, cst_typ_wo_param = Term.decompose_prod_n_assum nparam cst_typ in
-            let ctx = Context.(Rel.fold_outside Context.Rel.add ~init:arities_ctx param_ctx) in
-            traverse label ctx accu cst_typ_wo_param)
+            traverse label param_ctx accu cst_typ_wo_param)
           accu oib.mind_user_lc)
        accu mib.mind_packets
      in
      (* Maps all these dependencies to inductives and constructors*)
-     let data = Array.fold_left_i (fun n data oib ->
+     let data =
+       let contents = GlobRef.Set_env.remove firstind_ref contents in
+       Array.fold_left_i (fun n data oib ->
        let ind = (mind, n) in
-       let data = GlobRef.Map_env.add (GlobRef.IndRef ind) contents data in
+       let data = GlobRef.Map_env.add (GlobRef.IndRef ind) (Some contents) data in
        Array.fold_left_i (fun k data _ ->
-         GlobRef.Map_env.add (GlobRef.ConstructRef (ind, k+1)) contents data
+         GlobRef.Map_env.add (GlobRef.ConstructRef (ind, k+1)) (Some contents) data
        ) data oib.mind_consnames) data mib.mind_packets
      in
-     data, ax2ty
+     (data, ax2ty)
   in
   (GlobRef.Set_env.add obj curr, data, ax2ty)
 
@@ -352,7 +336,7 @@ let assumptions ?(add_opaque=false) ?(add_transparent=false) st gr t =
   (* Only keep the transitive dependencies *)
   let (_, graph, ax2ty) = traverse (label_of gr) t in
   let open GlobRef in
-  let fold obj _ accu = match obj with
+  let fold obj contents accu = match obj with
   | VarRef id ->
     let decl = Global.lookup_named id in
     if is_local_assum decl then
@@ -373,7 +357,7 @@ let assumptions ?(add_opaque=false) ?(add_transparent=false) st gr t =
           let l = try GlobRef.Map_env.find obj ax2ty with Not_found -> [] in
           ContextObjectMap.add (Axiom (TypeInType obj, l)) Constr.mkProp accu
       in
-    if not (Declareops.constant_has_body cb) then
+    if not (Option.has_some contents) then
       let t = type_of_constant cb in
       let l = try GlobRef.Map_env.find obj ax2ty with Not_found -> [] in
       ContextObjectMap.add (Axiom (Constant kn,l)) t accu
@@ -413,3 +397,13 @@ let assumptions ?(add_opaque=false) ?(add_transparent=false) st gr t =
       in
       accu
   in GlobRef.Map_env.fold fold graph ContextObjectMap.empty
+
+(* Reexport a wrapper to preserve the 8.16 API *)
+let traverse current t =
+  let (curr, graph, ax2ty) = traverse current t in
+  let map = function
+  | None -> GlobRef.Set_env.empty
+  | Some s -> s
+  in
+  let graph = GlobRef.Map_env.map map graph in
+  (curr, graph, ax2ty)
