@@ -26,7 +26,7 @@ module Hook = struct
       (** [(n1,t1),...(nm,tm)]: association list between obligation
           name and the corresponding defined term (might be a constant,
           but also an arbitrary term in the Expand case of obligations) *)
-      ; scope : Locality.locality
+      ; scope : Locality.definition_scope
       (**  [locality]: Locality of the original declaration *)
       ; dref : Names.GlobRef.t
       (** [ref]: identifier of the original declaration *)
@@ -81,7 +81,7 @@ module Info = struct
     ; inline : bool
     ; kind : Decls.logical_kind
     ; udecl : UState.universe_decl
-    ; scope : Locality.locality
+    ; scope : Locality.definition_scope
     ; hook : Hook.t option
     ; typing_flags : Declarations.typing_flags option
     }
@@ -89,7 +89,7 @@ module Info = struct
   (** Note that [opaque] doesn't appear here as it is not known at the
      start of the proof in the interactive case. *)
   let make ?(poly=false) ?(inline=false) ?(kind=Decls.(IsDefinition Definition))
-      ?(udecl=UState.default_univ_decl) ?(scope=Locality.Global Locality.ImportDefaultBehavior)
+      ?(udecl=UState.default_univ_decl) ?(scope=Locality.default_scope)
       ?hook ?typing_flags () =
     { poly; inline; kind; udecl; scope; hook; typing_flags }
 
@@ -195,24 +195,17 @@ let check_exists id =
     raise (DeclareUniv.AlreadyDeclared (None, id))
 
 let cache_constant ((sp,kn), obj) =
-  (* Invariant: the constant must exist in the logical environment *)
-  let kn' =
-    if Global.exists_objlabel (Label.of_id (Libnames.basename sp))
-    then Constant.make1 kn
-    else CErrors.anomaly Pp.(str"Missing constant " ++ Id.print(Libnames.basename sp) ++ str".")
-  in
-  assert (Environ.QConstant.equal (Global.env ()) kn' (Constant.make1 kn));
-  Nametab.push (Nametab.Until 1) sp (GlobRef.ConstRef (Constant.make1 kn));
-  Dumpglob.add_constant_kind (Constant.make1 kn) obj.cst_kind
+  let kn = Global.constant_of_delta_kn kn in
+  Nametab.push (Nametab.Until 1) sp (GlobRef.ConstRef kn);
+  Dumpglob.add_constant_kind kn obj.cst_kind
 
-let discharge_constant ((sp, kn), obj) =
-  Some obj
+let discharge_constant obj = Some obj
 
-let classify_constant cst = Libobject.Substitute cst
+let classify_constant cst = Libobject.Substitute
 
-let (objConstant : constant_obj Libobject.Dyn.tag) =
+let (objConstant : (Id.t * constant_obj) Libobject.Dyn.tag) =
   let open Libobject in
-  declare_object_full { (default_object "CONSTANT") with
+  declare_named_object_full { (default_object "CONSTANT") with
     cache_function = cache_constant;
     load_function = load_constant;
     open_function = simple_open open_constant;
@@ -227,12 +220,9 @@ let update_tables c =
   Notation.declare_ref_arguments_scope Evd.empty (GlobRef.ConstRef c)
 
 let register_constant kn kind local =
-  let o = inConstant {
-    cst_kind = kind;
-    cst_locl = local;
-  } in
   let id = Label.to_id (Constant.label kn) in
-  let _ = Lib.add_leaf id o in
+  let o = inConstant (id, { cst_kind = kind; cst_locl = local; }) in
+  let () = Lib.add_leaf o in
   update_tables kn
 
 let register_side_effect (c, body, role) =
@@ -486,10 +476,10 @@ type variable_declaration =
 
 (* This object is only for things which iterate over objects to find
    variables (only Prettyp.print_context AFAICT) *)
-let objVariable : unit Libobject.Dyn.tag =
+let objVariable : Id.t Libobject.Dyn.tag =
   let open Libobject in
   declare_object_full { (default_object "VARIABLE") with
-    classify_function = (fun () -> Dispose)}
+    classify_function = (fun _ -> Dispose)}
 
 let inVariable v = Libobject.Dyn.Easy.inj v objVariable
 
@@ -530,7 +520,7 @@ let declare_variable_core ~name ~kind d =
   in
   Nametab.push (Nametab.Until 1) (Libnames.make_path DirPath.empty name) (GlobRef.VarRef name);
   Decls.(add_variable_data name {opaque;kind});
-  ignore(Lib.add_leaf name (inVariable ()) : Libobject.object_name);
+  Lib.add_leaf (inVariable name);
   Impargs.declare_var_implicits ~impl name;
   Notation.declare_ref_arguments_scope Evd.empty (GlobRef.VarRef name)
 
@@ -651,7 +641,7 @@ let declare_definition_scheme ~internal ~univs ~role ~name c =
   kn, eff
 
 (* Locality stuff *)
-let declare_entry_core ~name ~scope ~kind ~typing_flags ?hook ~obls ~impargs ~uctx entry =
+let declare_entry_core ~name ?(scope=Locality.default_scope) ~kind ~typing_flags ?hook ~obls ~impargs ~uctx entry =
   let should_suggest =
     entry.proof_entry_opaque
     && not (List.is_empty (Global.named_context()))
@@ -1212,7 +1202,7 @@ let declare_mutual_definition ~pm l =
         ( d :: a1
         , r :: a2
         , typ :: a3
-        , CInfo.{name; typ; impargs; args = []; using } :: a4 ))
+        , (CInfo.make ~name ~typ ~impargs ?using ()) :: a4 ))
       defs first.prg_deps ([], [], [], [])
   in
   let fixkind = Option.get first.prg_fixkind in
@@ -1493,6 +1483,7 @@ let start_proof_core ~name ~typ ~pinfo ?(sign=initialize_named_context_for_proof
    can typically contain universe constraints) *)
 let start_core ~info ~cinfo ?proof_ending sigma =
   let { CInfo.name; typ; _ } = cinfo in
+  check_exists name;
   let cinfo = [{ cinfo with CInfo.typ = EConstr.Unsafe.to_constr cinfo.CInfo.typ }] in
   let pinfo = Proof_info.make ~cinfo ~info ?proof_ending () in
   start_proof_core ~name ~typ ~pinfo ?sign:None sigma
@@ -1656,7 +1647,7 @@ let prepare_proof ~unsafe_typ { proof } =
      equations and so far there is no code in the CI that will
      actually call those and do a side-effect, TTBOMK *)
   (* EJGA: likely the right solution is to attach side effects to the first constant only? *)
-  let proofs = List.map (fun (body, typ) -> (to_constr_body body, eff), to_constr_typ typ) initial_goals in
+  let proofs = List.map (fun (_, body, typ) -> (to_constr_body body, eff), to_constr_typ typ) initial_goals in
   proofs, Evd.evar_universe_context evd
 
 let make_univs_deferred ~poly ~initial_euctx ~uctx ~udecl
@@ -1740,7 +1731,7 @@ let close_proof_delayed ~feedback_id ps (fpl : closed_proof_output Future.comput
   (* We only support opaque proofs, this will be enforced by using
      different entries soon *)
   let opaque = true in
-  let make_entry i (_, types) =
+  let make_entry i (_, _, types) =
     (* Already checked the univ_decl for the type universes when starting the proof. *)
     let univs = UState.univ_entry ~poly:false initial_euctx in
     let types = nf (EConstr.Unsafe.to_constr types) in
@@ -1999,18 +1990,20 @@ let get_keep_admitted_vars =
     ~key:["Keep"; "Admitted"; "Variables"]
     ~value:true
 
-let compute_proof_using_for_admitted proof typ pproofs =
-  if not (get_keep_admitted_vars ()) then None
-  else match get_used_variables proof, pproofs with
-    | Some _ as x, _ -> x
-    | None, pproof :: _ ->
-      let env = Global.env () in
-      let ids_typ = Environ.global_vars_set env typ in
-      (* [pproof] is evar-normalized by [partial_proof]. We don't
-         count variables appearing only in the type of evars. *)
-      let ids_def = Environ.global_vars_set env (EConstr.Unsafe.to_constr pproof) in
-      Some (Environ.really_needed env (Id.Set.union ids_typ ids_def))
-    | _ -> None
+let compute_proof_using_for_admitted proof typ iproof =
+  if not (get_keep_admitted_vars ()) || not (Global.sections_are_opened()) then None
+  else match get_used_variables proof with
+    | Some _ as x -> x
+    | None ->
+      match Proof.partial_proof iproof with
+      | pproof :: _ ->
+        let env = Global.env () in
+        let ids_typ = Environ.global_vars_set env typ in
+        (* [pproof] is evar-normalized by [partial_proof]. We don't
+           count variables appearing only in the type of evars. *)
+        let ids_def = Environ.global_vars_set env (EConstr.Unsafe.to_constr pproof) in
+        Some (Environ.really_needed env (Id.Set.union ids_typ ids_def))
+      | [] -> None
 
 let finish_admitted ~pm ~pinfo ~uctx ~sec_vars ~univs =
   let cst = MutualEntry.declare_possibly_mutual_parameters ~pinfo ~uctx ~sec_vars ~univs in
@@ -2024,13 +2017,12 @@ let save_admitted ~pm ~proof =
   let udecl = get_universe_decl proof in
   let Proof.{ poly; entry } = Proof.data (get proof) in
   let typ = match Proofview.initial_goals entry with
-    | [typ] -> snd typ
+    | [_, _, typ] -> typ
     | _ -> CErrors.anomaly ~label:"Lemmas.save_lemma_admitted" (Pp.str "more than one statement.")
   in
   let typ = EConstr.Unsafe.to_constr typ in
   let iproof = get proof in
-  let pproofs = Proof.partial_proof iproof in
-  let sec_vars = compute_proof_using_for_admitted proof typ pproofs in
+  let sec_vars = compute_proof_using_for_admitted proof typ iproof in
   let uctx = get_initial_euctx proof in
   let univs = UState.check_univ_decl ~poly uctx udecl in
   finish_admitted ~pm ~pinfo:proof.pinfo ~uctx ~sec_vars ~univs
@@ -2332,7 +2324,6 @@ let rec solve_obligation prg num tac =
     then Error.depends user_num remaining
   in
   let obl = subst_deps_obl obls obl in
-  let scope = Locality.Global Locality.ImportNeedQualified in
   let kind = kind_of_obligation (snd obl.obl_status) in
   let evd = Evd.from_ctx (Internal.get_uctx prg) in
   let evd = Evd.update_sigma_univs (Global.universes ()) evd in
@@ -2344,7 +2335,7 @@ let rec solve_obligation prg num tac =
   let using = Internal.get_using prg in
   let cinfo = CInfo.make ~name:obl.obl_name ~typ:(EConstr.of_constr obl.obl_type) ?using () in
   let poly = Internal.get_poly prg in
-  let info = Info.make ~scope ~kind ~poly () in
+  let info = Info.make ~kind ~poly () in
   let lemma = Proof.start_core ~cinfo ~info ~proof_ending evd  in
   let lemma = fst @@ Proof.by !default_tactic lemma in
   let lemma = Option.cata (fun tac -> Proof.set_endline_tactic tac lemma) lemma tac in
@@ -2584,5 +2575,5 @@ module OblState = Obls_.State
 let declare_constant ?local ~name ~kind ?typing_flags =
   declare_constant ?local ~name ~kind ~typing_flags
 
-let declare_entry ~name ~scope ~kind =
-  declare_entry ~name ~scope ~kind ~typing_flags:None
+let declare_entry ~name ?scope ~kind =
+  declare_entry ~name ?scope ~kind ~typing_flags:None
