@@ -56,7 +56,7 @@ module Hook : sig
       (** [(n1,t1),...(nm,tm)]: association list between obligation
           name and the corresponding defined term (might be a constant,
           but also an arbitrary term in the Expand case of obligations) *)
-      ; scope : Locality.locality
+      ; scope : Locality.definition_scope
       (** [scope]: Locality of the original declaration *)
       ; dref : GlobRef.t
       (** [dref]: identifier of the original declaration *)
@@ -105,7 +105,7 @@ module Info : sig
     -> ?kind : Decls.logical_kind
     (** Theorem, etc... *)
     -> ?udecl : UState.universe_decl
-    -> ?scope : Locality.locality
+    -> ?scope : Locality.definition_scope
     (** locality  *)
     -> ?hook : Hook.t
     (** Callback to be executed after saving the constant *)
@@ -128,22 +128,13 @@ val declare_definition
   -> Evd.evar_map
   -> GlobRef.t
 
-val declare_assumption
-  :  name:Id.t
-  -> scope:Locality.locality
-  -> hook:Hook.t option
-  -> impargs:Impargs.manual_implicits
-  -> uctx:UState.t
-  -> Entries.parameter_entry
-  -> GlobRef.t
-
 type lemma_possible_guards = int list list
 
 val declare_mutually_recursive
   : info:Info.t
   -> cinfo: Constr.t CInfo.t list
   -> opaque:bool
-  -> ntns:Vernacexpr.decl_notation list
+  -> ntns:Metasyntax.where_decl_notation list
   -> uctx:UState.t
   -> rec_declaration:Constr.rec_declaration
   -> possible_indexes:lemma_possible_guards option
@@ -319,7 +310,7 @@ module Proof : sig
 
 end
 
-(** {2 low-level, internla API, avoid using unless you have special needs } *)
+(** {2 low-level, internal API, avoid using unless you have special needs } *)
 
 (** Proof entries represent a proof that has been finished, but still
    not registered with the kernel.
@@ -327,28 +318,41 @@ end
    XXX: This is an internal, low-level API and could become scheduled
    for removal from the public API, use higher-level declare APIs
    instead *)
-type 'a proof_entry
+type proof_entry
+type parameter_entry
+type primitive_entry
 
 val definition_entry
   :  ?opaque:bool
   -> ?using:Names.Id.Set.t
   -> ?inline:bool
   -> ?types:Constr.types
-  -> ?univs:Entries.universes_entry
+  -> ?univs:UState.named_universes_entry
   -> Constr.constr
-  -> Evd.side_effects proof_entry
+  -> proof_entry
+
+val parameter_entry
+  :  ?inline:int
+  -> ?univs:UState.named_universes_entry
+  -> Constr.constr
+  -> parameter_entry
+
+val primitive_entry
+  :  ?types:(Constr.types * UState.named_universes_entry)
+  -> CPrimitives.op_or_type
+  -> primitive_entry
 
 (** XXX: This is an internal, low-level API and could become scheduled
     for removal from the public API, use higher-level declare APIs
     instead *)
 val declare_entry
   :  name:Id.t
-  -> scope:Locality.locality
+  -> ?scope:Locality.definition_scope
   -> kind:Decls.logical_kind
   -> ?hook:Hook.t
   -> impargs:Impargs.manual_implicits
   -> uctx:UState.t
-  -> Evd.side_effects proof_entry
+  -> proof_entry
   -> GlobRef.t
 
 (** Declaration of local constructions (Variable/Hypothesis/Local) *)
@@ -357,6 +361,7 @@ val declare_variable
   -> kind:Decls.logical_kind
   -> typ:Constr.types
   -> impl:Glob_term.binding_kind
+  -> univs:UState.named_universes_entry
   -> unit
 
 (** Declaration of global constructions
@@ -365,17 +370,17 @@ val declare_variable
    XXX: This is an internal, low-level API and could become scheduled
    for removal from the public API, use higher-level declare APIs
    instead *)
-type 'a constant_entry =
-  | DefinitionEntry of 'a proof_entry
-  | ParameterEntry of Entries.parameter_entry
-  | PrimitiveEntry of Entries.primitive_entry
+type constant_entry =
+  | DefinitionEntry of proof_entry
+  | ParameterEntry of parameter_entry
+  | PrimitiveEntry of primitive_entry
 
 val prepare_parameter
   : poly:bool
   -> udecl:UState.universe_decl
   -> types:EConstr.types
   -> Evd.evar_map
-  -> Evd.evar_map * Entries.parameter_entry
+  -> Evd.evar_map * parameter_entry
 
 (** [declare_constant id cd] declares a global declaration
    (constant/parameter) with name [id] in the current section; it returns
@@ -389,7 +394,7 @@ val declare_constant
   -> name:Id.t
   -> kind:Decls.logical_kind
   -> ?typing_flags:Declarations.typing_flags
-  -> Evd.side_effects constant_entry
+  -> constant_entry
   -> Constant.t
 
 (** Declaration messages, for internal use *)
@@ -409,7 +414,7 @@ val build_by_tactic
   -> poly:bool
   -> typ:EConstr.types
   -> unit Proofview.tactic
-  -> Constr.constr * Constr.types option * Entries.universes_entry * bool * UState.t
+  -> Constr.constr * Constr.types option * (UState.named_universes_entry) * bool * UState.t
 
 (** {2 Program mode API} *)
 
@@ -515,7 +520,7 @@ val add_mutual_definitions :
   -> ?tactic:unit Proofview.tactic
   -> ?reduce:(Constr.t -> Constr.t)
   -> ?opaque:bool
-  -> ntns:Vernacexpr.decl_notation list
+  -> ntns:Metasyntax.where_decl_notation list
   -> fixpoint_kind
   -> OblState.t
 
@@ -551,6 +556,8 @@ val check_program_libraries : unit -> unit
 
 end
 
+val is_local_constant : Constant.t -> bool
+
 (** {6 For internal support, do not use}  *)
 
 module Internal : sig
@@ -558,10 +565,10 @@ module Internal : sig
   (* Liboject exports *)
   module Constant : sig
     type t
-    val tag : t Libobject.Dyn.tag
+    val tag : (Id.t * t) Libobject.Dyn.tag
     val kind : t -> Decls.logical_kind
   end
 
-  val objVariable : unit Libobject.Dyn.tag
+  val objVariable : Id.t Libobject.Dyn.tag
 
 end
