@@ -78,16 +78,19 @@ let create_assoc = function
   | None -> Gramlib.Gramext.RightA
   | Some a -> a
 
+exception NotationLevelMismatch
+  of entry_level * Gramlib.Gramext.g_assoc * Gramlib.Gramext.g_assoc
+
+let () = CErrors.register_handler (function
+  | NotationLevelMismatch (p, current, expected) ->
+      Some Pp.(str "Level " ++ int p ++ str " is already declared to have " ++
+        Gramlib.Gramext.pr_assoc current ++
+        str " while it is now expected to have " ++
+        Gramlib.Gramext.pr_assoc expected ++ str ".")
+  | _ -> None)
+
 let error_level_assoc p current expected =
-  let open Pp in
-  let pr_assoc = function
-    | Gramlib.Gramext.LeftA -> str "left"
-    | Gramlib.Gramext.RightA -> str "right"
-    | Gramlib.Gramext.NonA -> str "non" in
-  user_err
-    (str "Level " ++ int p ++ str " is already declared " ++
-     pr_assoc current ++ str " associative while it is now expected to be " ++
-     pr_assoc expected ++ str " associative.")
+  raise @@ NotationLevelMismatch (p, current, expected)
 
 type position = NewFirst | NewAfter of int | ReuseFirst | ReuseLevel of int
 
@@ -109,7 +112,7 @@ let find_position_gen current ensure assoc lev =
           let a' = create_assoc assoc in
           (init := Some (a', q); (p,a',false)::l)
         else if admissible_assoc (a,assoc) then
-          raise Exit
+          raise_notrace Exit
         else
           error_level_assoc p a (Option.get assoc)
       | l -> after := q; (n,create_assoc assoc,ensure)::l
@@ -250,14 +253,14 @@ type prod_info = production_level * production_position
 type (_, _) entry =
 | TTIdent : ('self, lident) entry
 | TTName : ('self, lname) entry
-| TTReference : ('self, qualid) entry
+| TTGlobal : ('self, qualid) entry
 | TTBigint : ('self, string) entry
 | TTBinder : bool -> ('self, kinded_cases_pattern_expr) entry
 | TTConstr : notation_entry * prod_info * 'r target -> ('r, 'r) entry
-| TTConstrList : notation_entry * prod_info * string Tok.p list * 'r target -> ('r, 'r list) entry
+| TTConstrList : notation_entry * prod_info * (bool * string) list * 'r target -> ('r, 'r list) entry
 | TTPattern : int -> ('self, cases_pattern_expr) entry
 | TTOpenBinderList : ('self, local_binder_expr list) entry
-| TTClosedBinderList : string Tok.p list -> ('self, local_binder_expr list list) entry
+| TTClosedBinderList : (bool * string) list -> ('self, local_binder_expr list list) entry
 
 type _ any_entry = TTAny : ('s, 'r) entry -> 's any_entry
 
@@ -328,12 +331,14 @@ let is_binder_level custom (custom',from) e = match e with
   custom = InConstrEntry && custom' = InConstrEntry && from = 200
 | _ -> false
 
-let make_sep_rules = function
-  | [tk] ->
-    Pcoq.Symbol.token tk
-  | tkl ->
-    let r = Pcoq.mk_rule (List.rev tkl) in
-    Pcoq.Symbol.rules [r]
+let make_pattern (keyword,s) =
+   if keyword then TPattern (Tok.PKEYWORD s) else
+     match NumTok.Unsigned.parse_string s with
+     | Some n -> TPattern (Tok.PNUMBER (Some n))
+     | None -> TPattern (Tok.PIDENT (Some s))
+
+let make_sep_rules tkl =
+  Pcoq.Symbol.tokens (List.map make_pattern tkl)
 
 type ('s, 'a) mayrec_symbol =
 | MayRecNo : ('s, Gramlib.Grammar.norec, 'a) Symbol.t -> ('s, 'a) mayrec_symbol
@@ -373,12 +378,12 @@ let symbol_of_entry : type s r. _ -> _ -> (s, r) entry -> (s, r) mayrec_symbol =
 | TTBinder false -> MayRecNo (Pcoq.Symbol.nterm Constr.one_closed_binder)
 | TTOpenBinderList -> MayRecNo (Pcoq.Symbol.nterm Constr.open_binders)
 | TTBigint -> MayRecNo (Pcoq.Symbol.nterm Prim.bignat)
-| TTReference -> MayRecNo (Pcoq.Symbol.nterm Constr.global)
+| TTGlobal -> MayRecNo (Pcoq.Symbol.nterm Constr.global)
 
 let interp_entry forpat e = match e with
 | ETProdIdent -> TTAny TTIdent
 | ETProdName -> TTAny TTName
-| ETProdReference -> TTAny TTReference
+| ETProdGlobal -> TTAny TTGlobal
 | ETProdBigint -> TTAny TTBigint
 | ETProdOneBinder o -> TTAny (TTBinder o)
 | ETProdConstr (s,p) -> TTAny (TTConstr (s, p, forpat))
@@ -429,7 +434,7 @@ match e with
   | ForConstr ->  push_constr subst (CAst.make @@ CPrim (Number (NumTok.Signed.of_int_string v)))
   | ForPattern -> push_constr subst (CAst.make @@ CPatPrim (Number (NumTok.Signed.of_int_string v)))
   end
-| TTReference ->
+| TTGlobal ->
   begin match forpat with
   | ForConstr  -> push_constr subst (CAst.make @@ CRef (v, None))
   | ForPattern -> push_constr subst (CAst.make @@ CPatAtom (Some v))
@@ -437,7 +442,7 @@ match e with
 | TTConstrList _ -> { subst with constrlists = v :: subst.constrlists }
 
 type (_, _) ty_symbol =
-| TyTerm : string Tok.p -> ('s, string) ty_symbol
+| TyTerm : 'a Tok.p -> ('s, 'a) ty_symbol
 | TyNonTerm : 's target * ('s, 'a) entry * ('s, 'a) mayrec_symbol * bool -> ('s, 'a) ty_symbol
 
 type ('self, _, 'r) ty_rule =
@@ -499,9 +504,10 @@ type ('self, 'r) any_ty_rule =
 let make_ty_rule assoc from forpat prods =
   let rec make_ty_rule = function
   | [] -> AnyTyRule TyStop
-  | GramConstrTerminal tok :: rem ->
+  | GramConstrTerminal (kw,s) :: rem ->
     let AnyTyRule r = make_ty_rule rem in
-    AnyTyRule (TyNext (r, TyTerm tok))
+    let TPattern tk = make_pattern (kw,s) in
+    AnyTyRule (TyNext (r, TyTerm tk))
   | GramConstrNonTerminal (e, var) :: rem ->
     let AnyTyRule r = make_ty_rule rem in
     let TTAny e = interp_entry forpat e in
