@@ -13,7 +13,6 @@ open CErrors
 open Util
 
 open Names
-open Lib
 open Libobject
 
 (************************************************************************)
@@ -214,7 +213,7 @@ let access_opaque_table o =
   match ans with
   | None -> None
   | Some (c, ctx) ->
-    let (c, ctx) = Cooking.cook_constr ci (c, ctx) in
+    let (c, ctx) = Discharge.cook_opaque_proofterm ci (c, ctx) in
     let c = Mod_subst.(List.fold_right subst_mps sub c) in
     Some (c, ctx)
 
@@ -336,34 +335,27 @@ let register_library m =
    - called at module or module type closing when a Require occurs in
      the module or module type
    - not called from a library (i.e. a module identified with a file) *)
-let load_require _ (_,(needed,modl,_)) =
+let load_require _ needed =
   List.iter register_library needed
 
-let open_require i (_,(_,modl,export)) =
-  Option.iter (fun export ->
-      let mpl = List.map (fun m -> unfiltered, MPfile m) modl in
-      (* TODO support filters in Require *)
-      Declaremods.import_modules ~export mpl)
-    export
-
   (* [needed] is the ordered list of libraries not already loaded *)
 let cache_require o =
-  load_require 1 o;
-  open_require 1 o
+  load_require 1 o
 
-let discharge_require (_,o) = Some o
+let discharge_require o = Some o
 
 (* open_function is never called from here because an Anticipate object *)
 
-type require_obj = library_t list * DirPath.t list * bool option
+type require_obj = library_t list
 
 let in_require : require_obj -> obj =
-  declare_object {(default_object "REQUIRE") with
-       cache_function = cache_require;
-       load_function = load_require;
-       open_function = (fun _ _ -> assert false);
-       discharge_function = discharge_require;
-       classify_function = (fun o -> Anticipate o) }
+  declare_object
+    {(default_object "REQUIRE") with
+     cache_function = cache_require;
+     load_function = load_require;
+     open_function = (fun _ _ -> assert false);
+     discharge_function = discharge_require;
+     classify_function = (fun o -> Anticipate) }
 
 (* Require libraries, import them if [export <> None], mark them for export
    if [export = Some true] *)
@@ -373,22 +365,11 @@ let warn_require_in_module =
     (fun () -> strbrk "Use of “Require” inside a module is fragile." ++ spc() ++
                strbrk "It is not recommended to use this functionality in finished proof scripts.")
 
-let require_library_from_dirpath ~lib_resolver modrefl export =
+let require_library_from_dirpath ~lib_resolver modrefl =
   let needed, contents = List.fold_left (rec_intern_library ~lib_resolver) ([], DPmap.empty) modrefl in
   let needed = List.rev_map (fun dir -> DPmap.find dir contents) needed in
-  let modrefl = List.map fst modrefl in
-  if Lib.is_module_or_modtype () then
-    begin
-      warn_require_in_module ();
-      add_anonymous_leaf (in_require (needed,modrefl,None));
-      Option.iter (fun export ->
-          (* TODO import filters *)
-          List.iter (fun m -> Declaremods.import_module unfiltered ~export (MPfile m)) modrefl)
-        export
-    end
-  else
-    add_anonymous_leaf (in_require (needed,modrefl,export));
-  ()
+  if Lib.is_module_or_modtype () then warn_require_in_module ();
+  Lib.add_leaf (in_require needed)
 
 (************************************************************************)
 (*s Initializing the compilation of a library. *)
