@@ -363,7 +363,7 @@ let unparsing_metavar i from typs =
   match x with
   | ETConstr _ | ETGlobal | ETBigint ->
      UnpMetaVar (prec,side)
-  | ETPattern _ | ETName _ | ETIdent ->
+  | ETPattern _ | ETName | ETIdent ->
      UnpBinderMetaVar (prec,NotQuotedPattern)
   | ETBinder isopen ->
      UnpBinderMetaVar (prec,QuotedPattern)
@@ -468,7 +468,7 @@ let rec split_format_at_ldots hd = function
   | u :: fmt ->
       check_no_ldots_in_box u;
       split_format_at_ldots (u::hd) fmt
-  | [] -> raise Exit
+  | [] -> raise_notrace Exit
 
 and check_no_ldots_in_box = function
   | (_,UnpBox (_,fmt)) ->
@@ -628,7 +628,7 @@ let include_possible_similar_trailing_pattern typ etyps sl l =
   | [], NonTerminal x ::l' when is_constr_typ typ x etyps -> try_aux n l'
   | Break _ :: sl, l -> aux n (sl,l)
   | sl, Break _ :: l -> aux n (sl,l)
-  | _ -> raise Exit
+  | _ -> raise_notrace Exit
   and try_aux n l =
     try aux (n+1) (sl,l)
     with Exit -> n,l in
@@ -636,8 +636,8 @@ let include_possible_similar_trailing_pattern typ etyps sl l =
 
 let prod_entry_type = function
   | ETIdent -> ETProdIdent
-  | ETName _ -> ETProdName
-  | ETGlobal -> ETProdReference
+  | ETName -> ETProdName
+  | ETGlobal -> ETProdGlobal
   | ETBigint -> ETProdBigint
   | ETBinder o -> ETProdOneBinder o
   | ETConstr (s,_,p) -> ETProdConstr (s,p)
@@ -837,24 +837,25 @@ let cache_one_syntax_extension (ntn,synext) =
   (* Printing *)
   Option.iter (declare_generic_notation_printing_rules ntn) synext.synext_notprint
 
-let cache_syntax_extension (_, (_, sy)) =
+let cache_syntax_extension (_, sy) =
   cache_one_syntax_extension sy
 
 let subst_syntax_extension (subst, (local, (ntn, synext))) =
   (local, (ntn, synext))
 
-let classify_syntax_definition (local, _ as o) =
-  if local then Dispose else Substitute o
+let classify_syntax_definition (local, _) =
+  if local then Dispose else Substitute
 
 let open_syntax_extension i o =
   if Int.equal i 1 then cache_syntax_extension o
 
 let inSyntaxExtension : syntax_extension_obj -> obj =
-  declare_object {(default_object "SYNTAX-EXTENSION") with
-       open_function = simple_open ~cat:notation_cat open_syntax_extension;
-       cache_function = cache_syntax_extension;
-       subst_function = subst_syntax_extension;
-       classify_function = classify_syntax_definition}
+  declare_object
+    {(default_object "SYNTAX-EXTENSION") with
+     open_function = simple_open ~cat:notation_cat open_syntax_extension;
+     cache_function = cache_syntax_extension;
+     subst_function = subst_syntax_extension;
+     classify_function = classify_syntax_definition}
 
 (**************************************************************************)
 (* Precedences                                                            *)
@@ -885,19 +886,20 @@ let default = {
 
 end
 
+exception UnknownCustomEntry of string
+
+let () = CErrors.register_handler @@ function
+  | UnknownCustomEntry entry -> Some Pp.(str "Unknown custom entry: " ++ str entry ++ str ".")
+  | _ -> None
+
 let check_custom_entry entry =
   if not (Egramcoq.exists_custom_entry entry) then
-    user_err Pp.(str "Unknown custom entry: " ++ str entry ++ str ".")
+    raise @@ UnknownCustomEntry entry
 
 let check_entry_type = function
   | ETConstr (InCustomEntry entry,_,_) -> check_custom_entry entry
   | ETConstr (InConstrEntry,_,_) | ETPattern _
-  | ETIdent | ETGlobal | ETBigint | ETName _ | ETBinder _-> ()
-
-(* To be turned into a fatal warning in 8.14 *)
-let warn_deprecated_ident_entry =
-  CWarnings.create ~name:"deprecated-ident-entry" ~category:"deprecated"
-         (fun () -> strbrk "grammar entry \"ident\" permitted \"_\" in addition to proper identifiers; this use is deprecated and its meaning will change in the future; use \"name\" instead.")
+  | ETIdent | ETGlobal | ETBigint | ETName | ETBinder _-> ()
 
 let interp_modifiers entry modl = let open NotationMods in
   let rec interp subtyps acc = function
@@ -948,12 +950,6 @@ let interp_modifiers entry modl = let open NotationMods in
   (* interpret item levels wrt to main entry *)
   let extra_etyps = List.map (fun (id,bko,n) -> (id,ETConstr (entry,bko,n))) subtyps in
   (* Temporary hack: "ETName false" (i.e. "ident" in deprecation phase) means "ETIdent" for custom entries *)
-  let mods =
-    { mods with etyps = List.map (function
-        | (id,ETName false) ->
-           if entry = InConstrEntry then (warn_deprecated_ident_entry (); (id,ETName true))
-           else (id,ETIdent)
-        | x -> x) mods.etyps } in
   { mods with etyps = extra_etyps@mods.etyps }
 
 let check_useless_entry_types recvars mainvars etyps =
@@ -1024,20 +1020,20 @@ let set_item_scope ?loc main_data ids sc =
   | (id,_)::_ -> user_err ?loc (str "Notation scope for argument " ++ Id.print id ++ str " can be specified only once.")
   | [] -> { main_data with itemscopes }
 
-let interp_non_syntax_modifiers ~reserved ~infix ~syndef deprecation mods =
+let interp_non_syntax_modifiers ~reserved ~infix ~abbrev deprecation mods =
   let set (main_data,rest) = CAst.with_loc_val (fun ?loc -> function
     | SetOnlyParsing ->
        if not (Option.is_empty main_data.format && List.is_empty main_data.extra) then
          (warn_only_parsing_discarded_format ?loc (); (main_data, rest))
        else
          (set_onlyparsing ?loc ~reserved main_data,rest)
-    | SetOnlyPrinting when not syndef -> (set_onlyprinting ?loc main_data,rest)
-    | SetCustomEntry (entry,None) when not syndef -> (set_custom_entry ?loc main_data entry,rest)
-    | SetCustomEntry (entry,Some _) as x when not syndef -> (set_custom_entry main_data entry,CAst.make ?loc x :: rest)
+    | SetOnlyPrinting when not abbrev -> (set_onlyprinting ?loc main_data,rest)
+    | SetCustomEntry (entry,None) when not abbrev -> (set_custom_entry ?loc main_data entry,rest)
+    | SetCustomEntry (entry,Some _) as x when not abbrev -> (set_custom_entry main_data entry,CAst.make ?loc x :: rest)
     | SetEntryType _ when infix -> user_err ?loc Pp.(str "Unexpected entry type in infix notation.")
     | SetItemLevel _ when infix -> user_err ?loc Pp.(str "Unexpected entry level in infix notation.")
-    | SetFormat (TextFormat s) when not syndef -> (set_format ?loc main_data s, rest)
-    | SetFormat (ExtraFormat (k,s)) when not syndef -> (set_extra_format ?loc main_data (k,s), rest)
+    | SetFormat (TextFormat s) when not abbrev -> (set_format ?loc main_data s, rest)
+    | SetFormat (ExtraFormat (k,s)) when not abbrev -> (set_extra_format ?loc main_data (k,s), rest)
     | SetItemScope (ids,sc) -> (set_item_scope ?loc main_data ids sc, rest)
     | modif -> (main_data,(CAst.make ?loc modif)::rest))
   in
@@ -1072,7 +1068,7 @@ let set_entry_type from n etyps (x,typ) =
       | ETConstr (s,bko,n), InternalProd ->
           ETConstr (s,bko,(n,InternalProd))
       | ETPattern (b,n), _ -> ETPattern (b,n)
-      | (ETIdent | ETName _ | ETBigint | ETGlobal | ETBinder _ as x), _ -> x
+      | (ETIdent | ETName | ETBigint | ETGlobal | ETBinder _ as x), _ -> x
     with Not_found ->
       ETConstr (from,None,(make_lev n from,typ))
   in (x,typ)
@@ -1095,7 +1091,7 @@ let join_auxiliary_recursive_types recvars etyps =
 let internalization_type_of_entry_type = function
   | ETBinder _ -> NtnInternTypeOnlyBinder
   | ETConstr _ | ETBigint | ETGlobal
-  | ETIdent | ETName _ | ETPattern _ -> NtnInternTypeAny None
+  | ETIdent | ETName | ETPattern _ -> NtnInternTypeAny None
 
 let make_internalization_vars recvars maintyps =
   let maintyps = List.map (on_snd internalization_type_of_entry_type) maintyps in
@@ -1112,7 +1108,7 @@ let make_interpretation_type isrec isonlybinding default_if_binding = function
   | ETConstr (_,None,_) -> NtnTypeConstr
   (* Others *)
   | ETIdent -> NtnTypeBinder NtnParsedAsIdent
-  | ETName _ -> NtnTypeBinder NtnParsedAsName
+  | ETName -> NtnTypeBinder NtnParsedAsName
   | ETPattern (ppstrict,_) -> NtnTypeBinder (NtnParsedAsPattern ppstrict) (* Parsed as ident/pattern, primarily interpreted as binder; maybe strict at printing *)
   | ETBigint | ETGlobal -> NtnTypeConstr
   | ETBinder _ ->
@@ -1232,7 +1228,7 @@ let find_precedence custom lev etyps symbols onlyprint =
           user_err Pp.(str "The level of the leftmost non-terminal cannot be changed.") in
       (try match List.assoc x etyps, custom with
         | ETConstr (s,_,(NumLevel _ | NextLevel)), s' when s = s' -> test ()
-        | (ETIdent | ETName _ | ETBigint | ETGlobal), _ ->
+        | (ETIdent | ETName | ETBigint | ETGlobal), _ ->
             begin match lev with
             | None ->
               ([fun () -> Flags.if_verbose (Feedback.msg_info ?loc:None) (strbrk "Setting notation at level 0.")],0)
@@ -1403,7 +1399,7 @@ type notation_obj = {
   notobj_also_in_cases_pattern : bool;
 }
 
-let load_notation_common silently_define_scope_if_undefined _ (_, nobj) =
+let load_notation_common silently_define_scope_if_undefined _ nobj =
   (* When the default shall be to require that a scope already exists *)
   (* the call to ensure_scope will have to be removed *)
   if silently_define_scope_if_undefined then
@@ -1416,7 +1412,7 @@ let load_notation_common silently_define_scope_if_undefined _ (_, nobj) =
 let load_notation =
   load_notation_common true
 
-let open_notation i (_, nobj) =
+let open_notation i nobj =
   if Int.equal i 1 then begin
     let scope = nobj.notobj_scope in
     let (ntn, df) = nobj.notobj_notation in
@@ -1444,7 +1440,7 @@ let subst_notation (subst, nobj) =
   { nobj with notobj_interp = subst_interpretation subst nobj.notobj_interp; }
 
 let classify_notation nobj =
-  if nobj.notobj_local then Dispose else Substitute nobj
+  if nobj.notobj_local then Dispose else Substitute
 
 let inNotation : notation_obj -> obj =
   declare_object {(default_object "NOTATION") with
@@ -1661,7 +1657,7 @@ let make_notation_interpretation ~local main_data notation_symbols ntn syntax_ru
 
 let add_reserved_notation ~local ~infix ({CAst.loc;v=df},mods) =
   let open SynData in
-  let (main_data,mods) = interp_non_syntax_modifiers ~reserved:true ~infix ~syndef:false None mods in
+  let (main_data,mods) = interp_non_syntax_modifiers ~reserved:true ~infix ~abbrev:false None mods in
   let mods = interp_modifiers main_data.entry mods in
   let notation_symbols, isnumeral = analyze_notation_tokens ~onlyprinting:main_data.onlyprinting ~infix main_data.entry df in
   let notation_symbols = if infix then adjust_reserved_infix_notation notation_symbols else notation_symbols in
@@ -1670,7 +1666,7 @@ let add_reserved_notation ~local ~infix ({CAst.loc;v=df},mods) =
   let sd = compute_syntax_data ~local main_data notation_symbols ntn mods in
   let synext = make_syntax_rules true main_data ntn sd in
   List.iter (fun f -> f ()) sd.msgs;
-  Lib.add_anonymous_leaf (inSyntaxExtension(local,(ntn,synext)))
+  Lib.add_leaf (inSyntaxExtension(local,(ntn,synext)))
 
 (* Notations associated to a where clause *)
 
@@ -1684,7 +1680,7 @@ let prepare_where_notation decl_ntn =
       decl_ntn_modifiers = modifiers;
       decl_ntn_scope = sc;
     } = decl_ntn in
-  let (main_data,mods) = interp_non_syntax_modifiers ~reserved:false ~infix:false ~syndef:false None modifiers in
+  let (main_data,mods) = interp_non_syntax_modifiers ~reserved:false ~infix:false ~abbrev:false None modifiers in
   match mods with
   | _::_ -> CErrors.user_err (str"Only modifiers not affecting parsing are supported here.")
   | [] ->
@@ -1700,23 +1696,22 @@ let prepare_where_notation decl_ntn =
 let add_notation_interpretation ~local env (decl_ntn, main_data, notation_symbols, ntn, syntax_rules) =
   let { decl_ntn_string = { CAst.loc ; v = df }; decl_ntn_interp = c; decl_ntn_scope = sc } = decl_ntn in
   let notation = make_notation_interpretation ~local main_data notation_symbols ntn syntax_rules df env c sc in
-  syntax_rules_iter (fun sy -> Lib.add_anonymous_leaf (inSyntaxExtension (local,(ntn,sy)))) syntax_rules;
-  Lib.add_anonymous_leaf (inNotation notation);
+  Lib.add_leaf (inNotation notation);
   Dumpglob.dump_notation (CAst.make ?loc ntn) sc true
 
 (* interpreting a where clause *)
 let set_notation_for_interpretation env impls (decl_ntn, main_data, notation_symbols, ntn, syntax_rules) =
   let { decl_ntn_string = { CAst.loc ; v = df }; decl_ntn_interp = c; decl_ntn_scope = sc } = decl_ntn in
   let notation = make_notation_interpretation ~local:true main_data notation_symbols ntn syntax_rules df env ~impls c sc in
-  syntax_rules_iter (fun sy -> Lib.add_anonymous_leaf (inSyntaxExtension (true,(ntn,sy)))) syntax_rules;
-  Lib.add_anonymous_leaf (inNotation notation);
+  syntax_rules_iter (fun sy -> Lib.add_leaf (inSyntaxExtension (true,(ntn,sy)))) syntax_rules;
+  Lib.add_leaf (inNotation notation);
   Option.iter (fun sc -> Notation.open_close_scope (false,true,sc)) sc
 
 (* Main entry point for command Notation *)
 
 let add_notation ~local ~infix deprecation env c ({CAst.loc;v=df},modifiers) sc =
   (* Extract the modifiers not affecting the parsing rule *)
-  let (main_data,syntax_modifiers) = interp_non_syntax_modifiers ~reserved:false ~infix ~syndef:false deprecation modifiers in
+  let (main_data,syntax_modifiers) = interp_non_syntax_modifiers ~reserved:false ~infix ~abbrev:false deprecation modifiers in
   (* Extract the modifiers not affecting the parsing rule *)
   let notation_symbols, isnumeral = analyze_notation_tokens ~onlyprinting:main_data.onlyprinting ~infix main_data.entry df in
   (* Add variables on both sides if an infix notation *)
@@ -1743,8 +1738,8 @@ let add_notation ~local ~infix deprecation env c ({CAst.loc;v=df},modifiers) sc
   (* Build the interpretation *)
   let notation = make_notation_interpretation ~local main_data notation_symbols ntn syntax_rules df env c sc in
   (* Declare both syntax and interpretation *)
-  syntax_rules_iter (fun sy -> Lib.add_anonymous_leaf (inSyntaxExtension (local,(ntn,sy)))) syntax_rules;
-  Lib.add_anonymous_leaf (inNotation notation);
+  syntax_rules_iter (fun sy -> Lib.add_leaf (inSyntaxExtension (local,(ntn,sy)))) syntax_rules;
+  Lib.add_leaf (inNotation notation);
   (* Dump the location of the notation for coqdoc *)
   Dumpglob.dump_notation (CAst.make ?loc ntn) sc true
 
@@ -1766,7 +1761,7 @@ type scope_command =
   | ScopeDelimRemove
   | ScopeClasses of scope_class list
 
-let load_scope_command_common silently_define_scope_if_undefined _ (_,(local,scope,o)) =
+let load_scope_command_common silently_define_scope_if_undefined _ (local,scope,o) =
   let declare_scope_if_needed =
     if silently_define_scope_if_undefined then Notation.declare_scope
     else Notation.ensure_scope in
@@ -1781,7 +1776,7 @@ let load_scope_command_common silently_define_scope_if_undefined _ (_,(local,sco
 let load_scope_command =
   load_scope_command_common true
 
-let open_scope_command i (_,(local,scope,o)) =
+let open_scope_command i (local,scope,o) =
   if Int.equal i 1 then
     match o with
     | ScopeDeclare -> ()
@@ -1803,8 +1798,8 @@ let subst_scope_command (subst,(local,scope,o as x)) = match o with
       local, scope, ScopeClasses cl'
   | _ -> x
 
-let classify_scope_command (local, _, _ as o) =
-  if local then Dispose else Substitute o
+let classify_scope_command (local, _, _) =
+  if local then Dispose else Substitute
 
 let inScopeCommand : locality_flag * scope_name * scope_command -> obj =
   declare_object {(default_object "DELIMITERS") with
@@ -1815,26 +1810,26 @@ let inScopeCommand : locality_flag * scope_name * scope_command -> obj =
       classify_function = classify_scope_command}
 
 let declare_scope local scope =
-  Lib.add_anonymous_leaf (inScopeCommand(local,scope,ScopeDeclare))
+  Lib.add_leaf (inScopeCommand(local,scope,ScopeDeclare))
 
 let add_delimiters local scope key =
-  Lib.add_anonymous_leaf (inScopeCommand(local,scope,ScopeDelimAdd key))
+  Lib.add_leaf (inScopeCommand(local,scope,ScopeDelimAdd key))
 
 let remove_delimiters local scope =
-  Lib.add_anonymous_leaf (inScopeCommand(local,scope,ScopeDelimRemove))
+  Lib.add_leaf (inScopeCommand(local,scope,ScopeDelimRemove))
 
 let add_class_scope local scope cl =
-  Lib.add_anonymous_leaf (inScopeCommand(local,scope,ScopeClasses cl))
+  Lib.add_leaf (inScopeCommand(local,scope,ScopeClasses cl))
 
-let interp_syndef_modifiers deprecation modl =
-  let mods, skipped = interp_non_syntax_modifiers ~reserved:false ~infix:false ~syndef:true deprecation modl in
+let interp_abbreviation_modifiers deprecation modl =
+  let mods, skipped = interp_non_syntax_modifiers ~reserved:false ~infix:false ~abbrev:true deprecation modl in
   if skipped <> [] then
     (let modifier = List.hd skipped in
     user_err ?loc:modifier.CAst.loc (str "Abbreviations don't support " ++ Ppvernac.pr_syntax_modifier modifier));
   (mods.onlyparsing, mods.itemscopes)
 
-let add_syntactic_definition ~local deprecation env ident (vars,c) modl =
-  let (only_parsing, scopes) = interp_syndef_modifiers deprecation modl in
+let add_abbreviation ~local deprecation env ident (vars,c) modl =
+  let (only_parsing, scopes) = interp_abbreviation_modifiers deprecation modl in
   let vars = List.map (fun v -> v, List.assoc_opt v scopes) vars in
   let acvars,pat,reversibility =
     match vars, intern_name_alias c with
@@ -1856,7 +1851,7 @@ let add_syntactic_definition ~local deprecation env ident (vars,c) modl =
   let vars = List.map (fun (x,_) -> (x, Id.Map.find x interp)) vars in
   let also_in_cases_pattern = has_no_binders_type vars in
   let onlyparsing = only_parsing || fst (printability None [] false reversibility pat) in
-  Syntax_def.declare_syntactic_definition ~local ~also_in_cases_pattern deprecation ident ~onlyparsing (vars,pat)
+  Abbreviation.declare_abbreviation ~local ~also_in_cases_pattern deprecation ident ~onlyparsing (vars,pat)
 
 (**********************************************************************)
 (* Declaration of custom entry                                        *)
@@ -1866,7 +1861,7 @@ let warn_custom_entry =
          (fun s ->
           strbrk "Custom entry " ++ str s ++ strbrk " has been overridden.")
 
-let load_custom_entry _ (_,(local,s)) =
+let load_custom_entry _ (local,s) =
   if Egramcoq.exists_custom_entry s then warn_custom_entry s
   else Egramcoq.create_custom_entry ~local s
 
@@ -1874,8 +1869,8 @@ let cache_custom_entry o = load_custom_entry 1 o
 
 let subst_custom_entry (subst,x) = x
 
-let classify_custom_entry (local,s as o) =
-  if local then Dispose else Substitute o
+let classify_custom_entry (local,s) =
+  if local then Dispose else Substitute
 
 let inCustomEntry : locality_flag * string -> obj =
   declare_object {(default_object "CUSTOM-ENTRIES") with
@@ -1888,4 +1883,4 @@ let declare_custom_entry local s =
   if Egramcoq.exists_custom_entry s then
     user_err Pp.(str "Custom entry " ++ str s ++ str " already exists.")
   else
-    Lib.add_anonymous_leaf (inCustomEntry (local,s))
+    Lib.add_leaf (inCustomEntry (local,s))
