@@ -39,7 +39,7 @@ module Vernac_ =
     let command = Entry.create "command"
     let syntax = Entry.create "syntax_command"
     let vernac_control = Entry.create "Vernac.vernac_control"
-    let inductive_definition = Entry.create "Vernac.inductive_definition"
+    let inductive_or_record_definition = Entry.create "Vernac.inductive_or_record_definition"
     let fix_definition = Entry.create "Vernac.fix_definition"
     let red_expr = Entry.create "red_expr"
     let hint_info = Entry.create "hint_info"
